/*
 * translate.c - translate between different pixel formats
 */

/*
 *  Copyright (C) 1999 AT&T Laboratories Cambridge. All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

#include "stdhdrs.h"

#include "translate.h"
#include <stdio.h>
#include "rfb.h"

#define CONCAT2(a,b) a##b
#define CONCAT2E(a,b) CONCAT2(a,b)
#define CONCAT4(a,b,c,d) a##b##c##d
#define CONCAT4E(a,b,c,d) CONCAT4(a,b,c,d)

#define OUT 8
#include "tableinittctemplate.cpp"
#include "tableinitcmtemplate.cpp"
#define IN 8
#include "tabletranstemplate.cpp"
#undef IN
#define __REPLACE_SINGLE
#define IN 16
#include "tabletranstemplate.cpp"
#undef IN
#undef __REPLACE_SINGLE
#define IN 32
#define __REPLACE_RGBTABLE
#include "tabletranstemplate.cpp"
#undef __REPLACE_RGBTABLE
#undef IN
#undef OUT

#define OUT 16
#include "tableinittctemplate.cpp"
#include "tableinitcmtemplate.cpp"
#define IN 8
#include "tabletranstemplate.cpp"
#undef IN
#define IN 16
#include "tabletranstemplate.cpp"
#undef IN
#define IN 32
#define __REPLACE_RGBTABLE
#include "tabletranstemplate.cpp"
#undef __REPLACE_RGBTABLE
#undef IN
#undef OUT

#define OUT 32
#include "tableinittctemplate.cpp"
#include "tableinitcmtemplate.cpp"
#define IN 8
#include "tabletranstemplate.cpp"
#undef IN
#define IN 16
#include "tabletranstemplate.cpp"
#undef IN
#define IN 32
#include "tabletranstemplate.cpp"
#undef IN
#undef OUT


// MaxYin: MMX routines to optimize translation process
#define MMX_TRANS16TO8	\
	{															\
	__asm		movq	mm0, tmp64		/* for R*/						\
	__asm		movq	mm1, mm0		/* for G*/						\
	__asm		movq	mm2, mm0		/* for B*/						\
	__asm		movq	mm3, redMask	/* load red mask*/				\
	__asm		pand	mm0, mm3		/* mask red component*/		\
	__asm		movq	mm3, greenMask	/* load green mask	*/		\
	__asm		pand	mm1, mm3		/* mask green component*/		\
	__asm		movq	mm3, blueMask	/* load blue mask	*/		\
	__asm		pand	mm2, mm3		/* mask blue component*/		\
	__asm		psrlw	mm0, 13			/* R shift right by 13 bits*/	\
	__asm		psrlw	mm1, 5			/* G shift right by 5 bits*/	\
	__asm		psllw	mm2, 3			/* B shift left by 3 bits*/	\
	__asm		por		mm0, mm1		/* merge R, G*/				\
	__asm		por		mm0, mm2		/* merge R, G, B*/				\
	__asm		pxor	mm3, mm3		/* clear mm3*/					\
	__asm		packuswb mm0, mm3		/* pack to lower dword	*/	\
	__asm		movd	tmp32, mm0		/* save to output array	*/	\
	}

static void
rfbTranslateWithSingleTable16to8 (char *table, rfbPixelFormat *in,
				    rfbPixelFormat *out,
				    char *iptr, char *optr,
				    int bytesBetweenInputLines,
				    int width, int height)
{
    unsigned short *ip = (unsigned short *)iptr;
    unsigned char *op = (unsigned char *)optr;
    int ipextra = bytesBetweenInputLines / sizeof(unsigned short) - width;
    unsigned char *opLineEnd;

	__int64 blueMask = 0x0018001800180018 ;
	__int64 greenMask = 0x0700070007000700;
	__int64 redMask = 0xe000e000e000e000;
	int pixelMod = width % 4;
	int i;
	__int32 tmp32;
	__int64 tmp64;

    while (height > 0) {
		opLineEnd = op + width - pixelMod;

		while (op < opLineEnd) {
			//*(op++) = t[*(ip++)];
			/* FIXME:
				Max Yin, Stardate: 20020421.1348
				I didn't find a proper way to access ip array directly from inline
				assembly, so I use C code to retrieve 64-bit data into tmp64 and
				feed into MMX register for processing. Output is in the same manner.
			*/
			tmp64 = *(__int64 *)ip;
			MMX_TRANS16TO8
			*(__int32*)op = tmp32;
			ip+=4;
			op+=4;
		}
		if(pixelMod){
			tmp64 = *(__int64*)ip;
			MMX_TRANS16TO8
			ip += pixelMod ;
		}
		__asm	emms ;

		for(i=0 ; i<pixelMod ; i++)
			*(op++) = ((unsigned char*)(&tmp32))[i];

		ip += ipextra;
		height--;
    }
}

#define MMX_TRANS32TO16	\
	{											\
	__asm	movq	mm0, tmp64	/* for R */				\
	__asm	movq	mm1, mm0		/* for G */				\
	__asm	movq	mm2, mm0		/* for B */				\
	__asm	movq	mm3, redMask	/* load red mask */			\
	__asm	pand	mm0, mm3		/* mask red component*/		\
	__asm	movq	mm3, greenMask	/* load green mask */		\
	__asm	pand	mm1, mm3		/* mask green component*/ 	\
	__asm	movq	mm3, blueMask	/* load blue mask*/			\
	__asm	pand	mm2, mm3		/* mask blue component*/		\
	__asm	psrld	mm0, 8	/* R shift right by 8 bits*/	\
	__asm	psrld	mm1, 5	/* G shift right by 5 bits*/	\
	__asm	psrld	mm2, 3	/* B shift left by 3 bits*/	\
	__asm	por		mm0, mm1	/* merge R, G*/			\
	__asm	por		mm0, mm2	/* merge R, G, B*/			\
	__asm	movq	mm1, mm0		/* copy to mm1*/			\
	__asm	psrlq	mm1, 16	/* mm1 shift right by 16 bits*/\
	__asm	por		mm0, mm1	/* or mm1 to mm0*/			\
	__asm	movd	tmp32, mm0	/* save to output array*/	\
	}

static void
rfbTranslateWithRGBTables32to16 (char *table, rfbPixelFormat *in,
				  rfbPixelFormat *out,
				  char *iptr, char *optr,
				  int bytesBetweenInputLines,
				  int width, int height)
{
    unsigned __int32 *ip = (unsigned __int32 *)iptr;
    unsigned __int16 *op = (unsigned __int16 *)optr;
    int ipextra = bytesBetweenInputLines / sizeof(__int32) - width;
    unsigned __int16 *opLineEnd;
	/* portable code
	__int64 blueMask = (((__int64)in->blueMax)<<in->(blueShift+32))|(in->blueMax<<in->blueShift);
	__int64 greenMask = (((__int64)in->greenMax)<<in->(greenShift+32))|(in->greenMax<<in->greenShift);
	__int64 redMask = (((__int64)in->redMax)<<in->(redShift+32))|(in->redMax<<in->redShift);
	*/
	// RGB565 format
	__int64 blueMask =  0x000000f8000000f8;
	__int64 greenMask = 0x0000fc000000fc00;
	__int64 redMask =   0x00f8000000f80000;

	int pixelMod = width % 2;

	__int32 tmp32;
	__int64 tmp64;

    while (height > 0) {
		opLineEnd = op + width - pixelMod;

		while (op < opLineEnd) {
			/* FIXME:
				Max Yin, Stardate: 20020421.1348
				I didn't find a proper way to access ip array directly from inline
				assembly, so I use C code to retrieve 64-bit data into tmp64 and
				feed into MMX register for processing. Output is in the same manner.
			*/

			tmp64 = *(__int64 *)ip;
			MMX_TRANS32TO16 ;
			*(__int32*)op = tmp32;
			ip+=2;
			op+=2;
		}

		if(pixelMod){
			tmp64 = *(__int64 *)ip;
			MMX_TRANS32TO16 ;
			ip ++;	// pixelMod is either 1 or 0
			*(op++) = ((unsigned __int16*)&tmp32)[0];
		}

		__asm	emms ;
		ip += ipextra;
		height--;
    }

}

#define MMX_TRANS32TO8	\
	{									\
	__asm	movq	mm0, tmp64	/* for R */	\
	__asm	movq	mm1, mm0		/* for G */	\
	__asm	movq	mm2, mm0		/* for B */	\
	__asm	movq	mm3, redMask	/* load red mask */				\
	__asm	pand	mm0, mm3		/* mask red component */			\
	__asm	movq	mm3, greenMask	/* load green mask */			\
	__asm	pand	mm1, mm3		/* mask green component */		\
	__asm	movq	mm3, blueMask	/* load blue mask */			\
	__asm	pand	mm2, mm3		/* mask blue component */		\
	__asm	psrld	mm0, 21	/* R shift right by 20 bits */	\
	__asm	psrld	mm1, 10	/* G shift right by 10 bits */	\
	__asm	por		mm0, mm1	/* merge R, G */				\
	__asm	por		mm0, mm2	/* merge R, G, B */				\
	__asm	pxor	mm3, mm3		/* clear mm3 */				\
	__asm	packuswb mm0, mm3	/* pack to lower word */			\
	__asm	packuswb mm0, mm3	/* pack to lower word */			\
	__asm	movd	tmp32, mm0	/* save to output array */		\
	}


static void
rfbTranslateWithRGBTables32to8 (char *table, rfbPixelFormat *in,
				  rfbPixelFormat *out,
				  char *iptr, char *optr,
				  int bytesBetweenInputLines,
				  int width, int height)
{
    unsigned __int32 *ip = (unsigned __int32 *)iptr;
    unsigned char *op = (unsigned char *)optr;
    int ipextra = bytesBetweenInputLines / sizeof(__int32) - width;
    unsigned char *opLineEnd;
	/* portable code
	__int64 blueMask = (((__int64)in->blueMax)<<in->(blueShift+32))|(in->blueMax<<in->blueShift);
	__int64 greenMask = (((__int64)in->greenMax)<<in->(greenShift+32))|(in->greenMax<<in->greenShift);
	__int64 redMask = (((__int64)in->redMax)<<in->(redShift+32))|(in->redMax<<in->redShift);
	*/
	// BGR233 format
	__int64 blueMask =  0x000000c0000000c0;
	__int64 greenMask = 0x0000e0000000e000;
	__int64 redMask =   0x00e0000000e00000;

	int pixelMod = width % 2;
	__int32 tmp32;
	__int64 tmp64;

    while (height > 0) {
		opLineEnd = op + width - pixelMod;

		while (op < opLineEnd) {
			/* FIXME:
				Max Yin, Stardate: 20020421.1348
				I didn't find a proper way to access ip array directly from inline
				assembly, so I use C code to retrieve 64-bit data into tmp64 and
				feed into MMX register for processing. Output is in the same manner.
			*/

			tmp64 = *(__int64 *)ip;
			MMX_TRANS32TO8 ;
			*(unsigned __int16*)op = (unsigned __int16)tmp32;
			ip+=2;
			op+=2;
		}

		if(pixelMod){
			tmp64 = *(__int64 *)ip;
			MMX_TRANS32TO8 ;
			ip ++;	// pixelMod is either 1 or 0
			*(op++) = ((unsigned char*)&tmp32)[0];
		}

		__asm	emms ;
		ip += ipextra;
		height--;
	}
}

static void
rfbTranslateWithRGBTables24to8 (char *table, rfbPixelFormat *in,
				  rfbPixelFormat *out,
				  char *iptr, char *optr,
				  int bytesBetweenInputLines,
				  int width, int height)
{
    unsigned __int8 *ip = (unsigned __int8 *)iptr;
    unsigned char *op = (unsigned char *)optr;
    int ipextra = bytesBetweenInputLines - 3*width;
    unsigned char *opLineEnd;
	/* portable code
	__int64 blueMask = (((__int64)in->blueMax)<<in->(blueShift+32))|(in->blueMax<<in->blueShift);
	__int64 greenMask = (((__int64)in->greenMax)<<in->(greenShift+32))|(in->greenMax<<in->greenShift);
	__int64 redMask = (((__int64)in->redMax)<<in->(redShift+32))|(in->redMax<<in->redShift);
	*/
	// BGR233 format
	__int64 blueMask =  0x000000c0000000c0;
	__int64 greenMask = 0x0000e0000000e000;
	__int64 redMask =   0x00e0000000e00000;

	int pixelMod = width % 2;
	__int32 tmp32;
	__int64 tmp64;

    while (height > 0) {
		opLineEnd = op + width - pixelMod;

		while (op < opLineEnd) {
			/* FIXME:
				Max Yin, Stardate: 20020421.1348
				I didn't find a proper way to access ip array directly from inline
				assembly, so I use C code to retrieve 64-bit data into tmp64 and
				feed into MMX register for processing. Output is in the same manner.
			*/

			tmp64 = (*((unsigned __int64 *)(ip+3)))<<32 | *((unsigned __int32 *)ip) ;
			MMX_TRANS32TO8 ;
			*(unsigned __int16*)op = (unsigned __int16)tmp32;
			ip+=6;
			op+=2;
		}

		if(pixelMod){
			tmp64 = (*((unsigned __int64 *)(ip+3)))<<32 | *((unsigned __int32 *)ip) ;
			MMX_TRANS32TO8 ;
			ip +=3;	// pixelMod is either 1 or 0
			*(op++) = ((unsigned char*)&tmp32)[0];
		}

		__asm	emms ;
		ip += ipextra;
		height--;
	}
}

static void
rfbTranslateWithRGBTables24to16 (char *table, rfbPixelFormat *in,
				  rfbPixelFormat *out,
				  char *iptr, char *optr,
				  int bytesBetweenInputLines,
				  int width, int height)
{
    unsigned __int8 *ip = (unsigned __int8 *)iptr;
    unsigned __int16 *op = (unsigned __int16 *)optr;
    int ipextra = bytesBetweenInputLines  - 3*width;
    unsigned __int16 *opLineEnd;
	/* portable code
	__int64 blueMask = (((__int64)in->blueMax)<<in->(blueShift+32))|(in->blueMax<<in->blueShift);
	__int64 greenMask = (((__int64)in->greenMax)<<in->(greenShift+32))|(in->greenMax<<in->greenShift);
	__int64 redMask = (((__int64)in->redMax)<<in->(redShift+32))|(in->redMax<<in->redShift);
	*/
	// RGB565 format
	__int64 blueMask =  0x000000f8000000f8;
	__int64 greenMask = 0x0000fc000000fc00;
	__int64 redMask =   0x00f8000000f80000;

	int pixelMod = width % 2;

	__int32 tmp32;
	__int64 tmp64;

    while (height > 0) {
		opLineEnd = op + width - pixelMod;

		while (op < opLineEnd) {
			/* FIXME:
				Max Yin, Stardate: 20020421.1348
				I didn't find a proper way to access ip array directly from inline
				assembly, so I use C code to retrieve 64-bit data into tmp64 and
				feed into MMX register for processing. Output is in the same manner.
			*/
			tmp64 = (*((unsigned __int64 *)(ip+3)))<<32 | *((unsigned __int32 *)ip) ;
			MMX_TRANS32TO16 ;
			*(__int32*)op = tmp32;
			ip+=6;
			op+=2;
		}

		if(pixelMod){
			tmp64 = (*((unsigned __int64 *)(ip+3)))<<32 | *((unsigned __int32 *)ip) ;
			MMX_TRANS32TO16 ;
			ip += 3;	// pixelMod is either 1 or 0 pixel
			*(op++) = ((unsigned __int16*)&tmp32)[0];
		}

		__asm	emms ;
		ip += ipextra;
		height--;
    }	
}

static void
rfbTranslateWithRGBTables24to32 (char *table, rfbPixelFormat *in,
				  rfbPixelFormat *out,
				  char *iptr, char *optr,
				  int bytesBetweenInputLines,
				  int width, int height)
{
 __int32 *o = (__int32*)optr;
 __int8*  i  = (__int8*)iptr;
 __int32 temp;
 int ipextra = bytesBetweenInputLines - width*3 ;
	for(int y=0 ; y<height ; y++){
		for(int x=0 ; x<width ; x++){
			*o++ = *(__int32*)i;
			i+=3;
		}
		i+=ipextra;
	}
}


rfbInitTableFnType rfbInitTrueColourSingleTableFns[3] = {
    rfbInitTrueColourSingleTable8,
    rfbInitTrueColourSingleTable16,
    rfbInitTrueColourSingleTable32
};

rfbInitTableFnType rfbInitColourMapSingleTableFns[3] = {
    rfbInitColourMapSingleTable8,
    rfbInitColourMapSingleTable16,
    rfbInitColourMapSingleTable32
};

rfbInitTableFnType rfbInitTrueColourRGBTablesFns[3] = {
    rfbInitTrueColourRGBTables8,
    rfbInitTrueColourRGBTables16,
    rfbInitTrueColourRGBTables32
};

rfbTranslateFnType rfbTranslateWithSingleTableFns[3][3] = {
    { rfbTranslateWithSingleTable8to8,
      rfbTranslateWithSingleTable8to16,
      rfbTranslateWithSingleTable8to32 },
    { rfbTranslateWithSingleTable16to8,
      rfbTranslateWithSingleTable16to16,
      rfbTranslateWithSingleTable16to32 },
    { rfbTranslateWithSingleTable32to8,
      rfbTranslateWithSingleTable32to16,
      rfbTranslateWithSingleTable32to32 }
};

rfbTranslateFnType rfbTranslateWithRGBTablesFns[4][3] = {
    { rfbTranslateWithRGBTables8to8,
      rfbTranslateWithRGBTables8to16,
      rfbTranslateWithRGBTables8to32 },
    { rfbTranslateWithRGBTables16to8,
      rfbTranslateWithRGBTables16to16,
      rfbTranslateWithRGBTables16to32 },
	{ rfbTranslateWithRGBTables24to8,
	  rfbTranslateWithRGBTables24to16,
	  rfbTranslateWithRGBTables24to32  },
    { rfbTranslateWithRGBTables32to8,
      rfbTranslateWithRGBTables32to16,
      rfbTranslateWithRGBTables32to32 }
};



// rfbTranslateNone is used when no translation is required.

void
rfbTranslateNone(char *table, rfbPixelFormat *in, rfbPixelFormat *out,
		 char *iptr, char *optr, int bytesBetweenInputLines,
		 int width, int height)
{
    int bytesPerOutputLine = width * (out->bitsPerPixel / 8);

    while (height > 0) {
	memcpy(optr, iptr, bytesPerOutputLine);
	iptr += bytesBetweenInputLines;
	optr += bytesPerOutputLine;
	height--;
    }
}

