/*
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2001-2002 Silicon Graphics, Inc. All rights reserved.
 */


#include <linux/types.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <asm/sn/sgi.h>
#include <asm/sn/sn_cpuid.h>
#include <asm/sn/addrs.h>
#include <asm/sn/arch.h>
#include <asm/sn/iograph.h>
#include <asm/sn/invent.h>
#include <asm/sn/hcl.h>
#include <asm/sn/labelcl.h>
#include <asm/sn/xtalk/xwidget.h>
#include <asm/sn/pci/bridge.h>
#include <asm/sn/pci/pciio.h>
#include <asm/sn/pci/pcibr.h>
#include <asm/sn/pci/pcibr_private.h>
#include <asm/sn/pci/pci_defs.h>
#include <asm/sn/prio.h>
#include <asm/sn/xtalk/xbow.h>
#include <asm/sn/ioc3.h>
#include <asm/sn/eeprom.h>
#include <asm/sn/io.h>
#include <asm/sn/sn_private.h>

#ifdef __ia64
#define rmallocmap atemapalloc
#define rmfreemap atemapfree
#define rmfree atefree
#define rmalloc atealloc
#endif

/*
 * Macros related to the Lucent USS 302/312 usb timeout workaround.  It
 * appears that if the lucent part can get into a retry loop if it sees a
 * DAC on the bus during a pio read retry.  The loop is broken after about
 * 1ms, so we need to set up bridges holding this part to allow at least
 * 1ms for pio.
 */

#define USS302_TIMEOUT_WAR

#ifdef USS302_TIMEOUT_WAR
#define LUCENT_USBHC_VENDOR_ID_NUM	0x11c1
#define LUCENT_USBHC302_DEVICE_ID_NUM	0x5801
#define LUCENT_USBHC312_DEVICE_ID_NUM	0x5802
#define USS302_BRIDGE_TIMEOUT_HLD	4
#endif

int                     pcibr_devflag = D_MP;

/*
 * This is the file operation table for the pcibr driver.
 * As each of the functions are implemented, put the
 * appropriate function name below.
 */
struct file_operations pcibr_fops = {
	owner:  THIS_MODULE,
	llseek: NULL,
	read: NULL,
	write: NULL,
	readdir: NULL,
	poll: NULL,
	ioctl: NULL,
	mmap: NULL,
	open: NULL,
	flush: NULL,
	release: NULL,
	fsync: NULL,
	fasync: NULL,
	lock: NULL,
	readv: NULL,
	writev: NULL
};

#ifdef LATER

#if PCIBR_ATE_DEBUG
static struct reg_values ssram_sizes[] =
{
    {BRIDGE_CTRL_SSRAM_512K, "512k"},
    {BRIDGE_CTRL_SSRAM_128K, "128k"},
    {BRIDGE_CTRL_SSRAM_64K, "64k"},
    {BRIDGE_CTRL_SSRAM_1K, "1k"},
    {0}
};

static struct reg_desc   control_bits[] =
{
    {BRIDGE_CTRL_FLASH_WR_EN, 0, "FLASH_WR_EN"},
    {BRIDGE_CTRL_EN_CLK50, 0, "EN_CLK50"},
    {BRIDGE_CTRL_EN_CLK40, 0, "EN_CLK40"},
    {BRIDGE_CTRL_EN_CLK33, 0, "EN_CLK33"},
    {BRIDGE_CTRL_RST_MASK, -24, "RST", "%x"},
    {BRIDGE_CTRL_IO_SWAP, 0, "IO_SWAP"},
    {BRIDGE_CTRL_MEM_SWAP, 0, "MEM_SWAP"},
    {BRIDGE_CTRL_PAGE_SIZE, 0, "PAGE_SIZE"},
    {BRIDGE_CTRL_SS_PAR_BAD, 0, "SS_PAR_BAD"},
    {BRIDGE_CTRL_SS_PAR_EN, 0, "SS_PAR_EN"},
    {BRIDGE_CTRL_SSRAM_SIZE_MASK, 0, "SSRAM_SIZE", 0, ssram_sizes},
    {BRIDGE_CTRL_F_BAD_PKT, 0, "F_BAD_PKT"},
    {BRIDGE_CTRL_LLP_XBAR_CRD_MASK, -12, "LLP_XBAR_CRD", "%d"},
    {BRIDGE_CTRL_CLR_RLLP_CNT, 0, "CLR_RLLP_CNT"},
    {BRIDGE_CTRL_CLR_TLLP_CNT, 0, "CLR_TLLP_CNT"},
    {BRIDGE_CTRL_SYS_END, 0, "SYS_END"},

    {BRIDGE_CTRL_BUS_SPEED_MASK, -4, "BUS_SPEED", "%d"},
    {BRIDGE_CTRL_WIDGET_ID_MASK, 0, "WIDGET_ID", "%x"},
    {0}
};
#endif
#endif	/* LATER */

/* kbrick widgetnum-to-bus layout */
int p_busnum[MAX_PORT_NUM] = {                  /* widget#      */
        0, 0, 0, 0, 0, 0, 0, 0,                 /* 0x0 - 0x7    */
        2,                                      /* 0x8          */
        1,                                      /* 0x9          */
        0, 0,                                   /* 0xa - 0xb    */
        5,                                      /* 0xc          */
        6,                                      /* 0xd          */
        4,                                      /* 0xe          */
        3,                                      /* 0xf          */
};

/*
 * Additional PIO spaces per slot are
 * recorded in this structure.
 */
struct pciio_piospace_s {
    pciio_piospace_t        next;       /* another space for this device */
    char                    free;       /* 1 if free, 0 if in use               */
    pciio_space_t           space;      /* Which space is in use                */
    iopaddr_t               start;      /* Starting address of the PIO space    */
    size_t                  count;      /* size of PIO space                    */
};

#if PCIBR_SOFT_LIST
pcibr_list_p            pcibr_list = 0;
#endif

extern int              hwgraph_vertex_name_get(devfs_handle_t vhdl, char *buf, uint buflen);
extern int              hub_device_flags_set(devfs_handle_t widget_dev, hub_widget_flags_t flags);
extern long             atoi(register char *p);
extern cnodeid_t        nodevertex_to_cnodeid(devfs_handle_t vhdl);
extern void             *swap_ptr(void **loc, void *new);
extern char             *dev_to_name(devfs_handle_t dev, char *buf, uint buflen);
extern struct map       *atemapalloc(uint64_t);
extern void             atefree(struct map *, size_t, uint64_t);
extern void             atemapfree(struct map *);
extern pciio_dmamap_t   get_free_pciio_dmamap(devfs_handle_t);
extern void		free_pciio_dmamap(pcibr_dmamap_t);

#define	ATE_WRITE()    ate_write(ate_ptr, ate_count, ate)
#if PCIBR_FREEZE_TIME
#define	ATE_FREEZE()	s = ate_freeze(pcibr_dmamap, &freeze_time, cmd_regs)
#else
#define	ATE_FREEZE()	s = ate_freeze(pcibr_dmamap, cmd_regs)
#endif /* PCIBR_FREEZE_TIME */

#if PCIBR_FREEZE_TIME
#define	ATE_THAW()	ate_thaw(pcibr_dmamap, ate_index, ate, ate_total, freeze_time, cmd_regs, s)
#else
#define	ATE_THAW()	ate_thaw(pcibr_dmamap, ate_index, cmd_regs, s)
#endif


/* =====================================================================
 *    Function Table of Contents
 *
 *      The order of functions in this file has stopped
 *      making much sense. We might want to take a look
 *      at it some time and bring back some sanity, or
 *      perhaps bust this file into smaller chunks.
 */

extern void              do_pcibr_rrb_clear(bridge_t *, int);
extern void              do_pcibr_rrb_flush(bridge_t *, int);
extern int               do_pcibr_rrb_count_valid(bridge_t *, pciio_slot_t);
extern int               do_pcibr_rrb_count_avail(bridge_t *, pciio_slot_t);
extern int               do_pcibr_rrb_alloc(bridge_t *, pciio_slot_t, int);
extern int               do_pcibr_rrb_free(bridge_t *, pciio_slot_t, int);

extern void              do_pcibr_rrb_autoalloc(pcibr_soft_t, int, int);

extern int  		 pcibr_wrb_flush(devfs_handle_t);
extern int               pcibr_rrb_alloc(devfs_handle_t, int *, int *);
extern int               pcibr_rrb_check(devfs_handle_t, int *, int *, int *, int *);
extern int               pcibr_alloc_all_rrbs(devfs_handle_t, int, int, int, int, int, int, int, int, int);
extern void              pcibr_rrb_flush(devfs_handle_t);

static int                pcibr_try_set_device(pcibr_soft_t, pciio_slot_t, unsigned, bridgereg_t);
void                     pcibr_release_device(pcibr_soft_t, pciio_slot_t, bridgereg_t);

extern void              pcibr_clearwidint(bridge_t *);
extern void              pcibr_setwidint(xtalk_intr_t);

void                     pcibr_init(void);
int                      pcibr_attach(devfs_handle_t);
int			 pcibr_detach(devfs_handle_t);
int                      pcibr_open(devfs_handle_t *, int, int, cred_t *);
int                      pcibr_close(devfs_handle_t, int, int, cred_t *);
int                      pcibr_map(devfs_handle_t, vhandl_t *, off_t, size_t, uint);
int                      pcibr_unmap(devfs_handle_t, vhandl_t *);
int                      pcibr_ioctl(devfs_handle_t, int, void *, int, struct cred *, int *);

void                     pcibr_freeblock_sub(iopaddr_t *, iopaddr_t *, iopaddr_t, size_t);

extern int               pcibr_init_ext_ate_ram(bridge_t *);
extern int               pcibr_ate_alloc(pcibr_soft_t, int);
extern void              pcibr_ate_free(pcibr_soft_t, int, int);

extern unsigned          ate_freeze(pcibr_dmamap_t pcibr_dmamap,
#if PCIBR_FREEZE_TIME
	   		 unsigned *freeze_time_ptr,
#endif
						unsigned *cmd_regs);
extern void              ate_write(bridge_ate_p ate_ptr, int ate_count, bridge_ate_t ate);
extern void              ate_thaw(pcibr_dmamap_t pcibr_dmamap, int ate_index,
#if PCIBR_FREEZE_TIME
	 				bridge_ate_t ate,
	 				int ate_total,
	 				unsigned freeze_time_start,
#endif
	 				unsigned *cmd_regs,
	 				unsigned s);

pcibr_info_t             pcibr_info_get(devfs_handle_t);

static iopaddr_t         pcibr_addr_pci_to_xio(devfs_handle_t, pciio_slot_t, pciio_space_t, iopaddr_t, size_t, unsigned);

pcibr_piomap_t          pcibr_piomap_alloc(devfs_handle_t, device_desc_t, pciio_space_t, iopaddr_t, size_t, size_t, unsigned);
void                    pcibr_piomap_free(pcibr_piomap_t);
caddr_t                 pcibr_piomap_addr(pcibr_piomap_t, iopaddr_t, size_t);
void                    pcibr_piomap_done(pcibr_piomap_t);
caddr_t                 pcibr_piotrans_addr(devfs_handle_t, device_desc_t, pciio_space_t, iopaddr_t, size_t, unsigned);
iopaddr_t               pcibr_piospace_alloc(devfs_handle_t, device_desc_t, pciio_space_t, size_t, size_t);
void                    pcibr_piospace_free(devfs_handle_t, pciio_space_t, iopaddr_t, size_t);

static iopaddr_t         pcibr_flags_to_d64(unsigned, pcibr_soft_t);
extern bridge_ate_t     pcibr_flags_to_ate(unsigned);

pcibr_dmamap_t          pcibr_dmamap_alloc(devfs_handle_t, device_desc_t, size_t, unsigned);
void                    pcibr_dmamap_free(pcibr_dmamap_t);
extern bridge_ate_p     pcibr_ate_addr(pcibr_soft_t, int);
static iopaddr_t         pcibr_addr_xio_to_pci(pcibr_soft_t, iopaddr_t, size_t);
iopaddr_t               pcibr_dmamap_addr(pcibr_dmamap_t, paddr_t, size_t);
alenlist_t              pcibr_dmamap_list(pcibr_dmamap_t, alenlist_t, unsigned);
void                    pcibr_dmamap_done(pcibr_dmamap_t);
cnodeid_t		pcibr_get_dmatrans_node(devfs_handle_t);
iopaddr_t               pcibr_dmatrans_addr(devfs_handle_t, device_desc_t, paddr_t, size_t, unsigned);
alenlist_t              pcibr_dmatrans_list(devfs_handle_t, device_desc_t, alenlist_t, unsigned);
void                    pcibr_dmamap_drain(pcibr_dmamap_t);
void                    pcibr_dmaaddr_drain(devfs_handle_t, paddr_t, size_t);
void                    pcibr_dmalist_drain(devfs_handle_t, alenlist_t);
iopaddr_t               pcibr_dmamap_pciaddr_get(pcibr_dmamap_t);

extern unsigned		pcibr_intr_bits(pciio_info_t info, pciio_intr_line_t lines);
extern pcibr_intr_t     pcibr_intr_alloc(devfs_handle_t, device_desc_t, pciio_intr_line_t, devfs_handle_t);
extern void             pcibr_intr_free(pcibr_intr_t);
extern void             pcibr_setpciint(xtalk_intr_t);
extern int              pcibr_intr_connect(pcibr_intr_t);
extern void             pcibr_intr_disconnect(pcibr_intr_t);

extern devfs_handle_t     pcibr_intr_cpu_get(pcibr_intr_t);
extern void             pcibr_xintr_preset(void *, int, xwidgetnum_t, iopaddr_t, xtalk_intr_vector_t);
extern void             pcibr_intr_func(intr_arg_t);

extern void             print_bridge_errcmd(uint32_t, char *);

extern void             pcibr_error_dump(pcibr_soft_t);
extern uint32_t         pcibr_errintr_group(uint32_t);
extern void	        pcibr_pioerr_check(pcibr_soft_t);
extern void             pcibr_error_intr_handler(intr_arg_t);

extern int              pcibr_addr_toslot(pcibr_soft_t, iopaddr_t, pciio_space_t *, iopaddr_t *, pciio_function_t *);
extern void             pcibr_error_cleanup(pcibr_soft_t, int);
extern void                    pcibr_device_disable(pcibr_soft_t, int);
extern int              pcibr_pioerror(pcibr_soft_t, int, ioerror_mode_t, ioerror_t *);
extern int              pcibr_dmard_error(pcibr_soft_t, int, ioerror_mode_t, ioerror_t *);
extern int              pcibr_dmawr_error(pcibr_soft_t, int, ioerror_mode_t, ioerror_t *);
extern int              pcibr_error_handler(error_handler_arg_t, int, ioerror_mode_t, ioerror_t *);
extern int              pcibr_error_devenable(devfs_handle_t, int);

void                    pcibr_provider_startup(devfs_handle_t);
void                    pcibr_provider_shutdown(devfs_handle_t);

int                     pcibr_reset(devfs_handle_t);
pciio_endian_t          pcibr_endian_set(devfs_handle_t, pciio_endian_t, pciio_endian_t);
int                     pcibr_priority_bits_set(pcibr_soft_t, pciio_slot_t, pciio_priority_t);
pciio_priority_t        pcibr_priority_set(devfs_handle_t, pciio_priority_t);
int                     pcibr_device_flags_set(devfs_handle_t, pcibr_device_flags_t);

extern cfg_p            pcibr_config_addr(devfs_handle_t, unsigned);
extern uint64_t         pcibr_config_get(devfs_handle_t, unsigned, unsigned);
extern void             pcibr_config_set(devfs_handle_t, unsigned, unsigned, uint64_t);
extern void             do_pcibr_config_set(cfg_p, unsigned, unsigned, uint64_t);

extern pcibr_hints_t    pcibr_hints_get(devfs_handle_t, int);
extern void             pcibr_hints_fix_rrbs(devfs_handle_t);
extern void             pcibr_hints_dualslot(devfs_handle_t, pciio_slot_t, pciio_slot_t);
extern void	 	pcibr_hints_intr_bits(devfs_handle_t, pcibr_intr_bits_f *);
extern void             pcibr_set_rrb_callback(devfs_handle_t, rrb_alloc_funct_t);
extern void             pcibr_hints_handsoff(devfs_handle_t);
extern void             pcibr_hints_subdevs(devfs_handle_t, pciio_slot_t, uint64_t);

#ifdef BRIDGE_B_DATACORR_WAR
extern int              ql_bridge_rev_b_war(devfs_handle_t);
extern int              bridge_rev_b_data_check_disable;
char                   *rev_b_datacorr_warning =
"***************************** WARNING! ******************************\n";
char                   *rev_b_datacorr_mesg =
"UNRECOVERABLE IO LINK ERROR. CONTACT SERVICE PROVIDER\n";
#endif

extern int		pcibr_slot_reset(devfs_handle_t,pciio_slot_t);
extern int		pcibr_slot_info_init(devfs_handle_t,pciio_slot_t);
extern int		pcibr_slot_info_free(devfs_handle_t,pciio_slot_t);
extern int		pcibr_slot_addr_space_init(devfs_handle_t,pciio_slot_t);
extern int		pcibr_slot_device_init(devfs_handle_t, pciio_slot_t);
extern int		pcibr_slot_guest_info_init(devfs_handle_t,pciio_slot_t);
extern int		pcibr_slot_call_device_attach(devfs_handle_t, pciio_slot_t, int);
extern int		pcibr_slot_call_device_detach(devfs_handle_t, pciio_slot_t, int);
extern int              pcibr_slot_attach(devfs_handle_t, pciio_slot_t, int, char *, int *);
extern int              pcibr_slot_detach(devfs_handle_t, pciio_slot_t, int);
extern int 		pcibr_is_slot_sys_critical(devfs_handle_t, pciio_slot_t);

#ifdef LATER
extern int		pcibr_slot_startup(devfs_handle_t, pcibr_slot_req_t);
extern int		pcibr_slot_shutdown(devfs_handle_t, pcibr_slot_req_t);
extern int		pcibr_slot_query(devfs_handle_t, pcibr_slot_req_t);
#endif

extern int		pcibr_slot_initial_rrb_alloc(devfs_handle_t, pciio_slot_t);
extern int		pcibr_initial_rrb(devfs_handle_t, pciio_slot_t, pciio_slot_t);



/* =====================================================================
 *    Device(x) register management
 */

/* pcibr_try_set_device: attempt to modify Device(x)
 * for the specified slot on the specified bridge
 * as requested in flags, limited to the specified
 * bits. Returns which BRIDGE bits were in conflict,
 * or ZERO if everything went OK.
 *
 * Caller MUST hold pcibr_lock when calling this function.
 */
static int
pcibr_try_set_device(pcibr_soft_t pcibr_soft,
		     pciio_slot_t slot,
		     unsigned flags,
		     bridgereg_t mask)
{
    bridge_t               *bridge;
    pcibr_soft_slot_t       slotp;
    bridgereg_t             old;
    bridgereg_t             new;
    bridgereg_t             chg;
    bridgereg_t             bad;
    bridgereg_t             badpmu;
    bridgereg_t             badd32;
    bridgereg_t             badd64;
    bridgereg_t             fix;
    unsigned long           s;
    bridgereg_t             xmask;

    xmask = mask;
    if (pcibr_soft->bs_xbridge) {
    	if (mask == BRIDGE_DEV_PMU_BITS)
		xmask = XBRIDGE_DEV_PMU_BITS;
	if (mask == BRIDGE_DEV_D64_BITS)
		xmask = XBRIDGE_DEV_D64_BITS;
    }

    slotp = &pcibr_soft->bs_slot[slot];

    s = pcibr_lock(pcibr_soft);

    bridge = pcibr_soft->bs_base;

    old = slotp->bss_device;

    /* figure out what the desired
     * Device(x) bits are based on
     * the flags specified.
     */

    new = old;

    /* Currently, we inherit anything that
     * the new caller has not specified in
     * one way or another, unless we take
     * action here to not inherit.
     *
     * This is needed for the "swap" stuff,
     * since it could have been set via
     * pcibr_endian_set -- altho note that
     * any explicit PCIBR_BYTE_STREAM or
     * PCIBR_WORD_VALUES will freely override
     * the effect of that call (and vice
     * versa, no protection either way).
     *
     * I want to get rid of pcibr_endian_set
     * in favor of tracking DMA endianness
     * using the flags specified when DMA
     * channels are created.
     */

#define	BRIDGE_DEV_WRGA_BITS	(BRIDGE_DEV_PMU_WRGA_EN | BRIDGE_DEV_DIR_WRGA_EN)
#define	BRIDGE_DEV_SWAP_BITS	(BRIDGE_DEV_SWAP_PMU | BRIDGE_DEV_SWAP_DIR)

    /* Do not use Barrier, Write Gather,
     * or Prefetch unless asked.
     * Leave everything else as it
     * was from the last time.
     */
    new = new
	& ~BRIDGE_DEV_BARRIER
	& ~BRIDGE_DEV_WRGA_BITS
	& ~BRIDGE_DEV_PREF
	;

    /* Generic macro flags
     */
    if (flags & PCIIO_DMA_DATA) {
	new = (new
            & ~BRIDGE_DEV_BARRIER)      /* barrier off */
            | BRIDGE_DEV_PREF;          /* prefetch on */

    }
    if (flags & PCIIO_DMA_CMD) {
        new = ((new
            & ~BRIDGE_DEV_PREF)         /* prefetch off */
            & ~BRIDGE_DEV_WRGA_BITS)    /* write gather off */
            | BRIDGE_DEV_BARRIER;       /* barrier on */
    }
    /* Generic detail flags
     */
    if (flags & PCIIO_WRITE_GATHER)
	new |= BRIDGE_DEV_WRGA_BITS;
    if (flags & PCIIO_NOWRITE_GATHER)
	new &= ~BRIDGE_DEV_WRGA_BITS;

    if (flags & PCIIO_PREFETCH)
	new |= BRIDGE_DEV_PREF;
    if (flags & PCIIO_NOPREFETCH)
	new &= ~BRIDGE_DEV_PREF;

    if (flags & PCIBR_WRITE_GATHER)
	new |= BRIDGE_DEV_WRGA_BITS;
    if (flags & PCIBR_NOWRITE_GATHER)
	new &= ~BRIDGE_DEV_WRGA_BITS;

    if (flags & PCIIO_BYTE_STREAM)
	new |= (pcibr_soft->bs_xbridge) ? 
			BRIDGE_DEV_SWAP_DIR : BRIDGE_DEV_SWAP_BITS;
    if (flags & PCIIO_WORD_VALUES)
	new &= (pcibr_soft->bs_xbridge) ? 
			~BRIDGE_DEV_SWAP_DIR : ~BRIDGE_DEV_SWAP_BITS;

    /* Provider-specific flags
     */
    if (flags & PCIBR_PREFETCH)
	new |= BRIDGE_DEV_PREF;
    if (flags & PCIBR_NOPREFETCH)
	new &= ~BRIDGE_DEV_PREF;

    if (flags & PCIBR_PRECISE)
	new |= BRIDGE_DEV_PRECISE;
    if (flags & PCIBR_NOPRECISE)
	new &= ~BRIDGE_DEV_PRECISE;

    if (flags & PCIBR_BARRIER)
	new |= BRIDGE_DEV_BARRIER;
    if (flags & PCIBR_NOBARRIER)
	new &= ~BRIDGE_DEV_BARRIER;

    if (flags & PCIBR_64BIT)
	new |= BRIDGE_DEV_DEV_SIZE;
    if (flags & PCIBR_NO64BIT)
	new &= ~BRIDGE_DEV_DEV_SIZE;

    chg = old ^ new;				/* what are we changing, */
    chg &= xmask;				/* of the interesting bits */

    if (chg) {

	badd32 = slotp->bss_d32_uctr ? (BRIDGE_DEV_D32_BITS & chg) : 0;
	if (pcibr_soft->bs_xbridge) {
		badpmu = slotp->bss_pmu_uctr ? (XBRIDGE_DEV_PMU_BITS & chg) : 0;
		badd64 = slotp->bss_d64_uctr ? (XBRIDGE_DEV_D64_BITS & chg) : 0;
	} else {
		badpmu = slotp->bss_pmu_uctr ? (BRIDGE_DEV_PMU_BITS & chg) : 0;
		badd64 = slotp->bss_d64_uctr ? (BRIDGE_DEV_D64_BITS & chg) : 0;
	}
	bad = badpmu | badd32 | badd64;

	if (bad) {

	    /* some conflicts can be resolved by
	     * forcing the bit on. this may cause
	     * some performance degredation in
	     * the stream(s) that want the bit off,
	     * but the alternative is not allowing
	     * the new stream at all.
	     */
            if ( (fix = bad & (BRIDGE_DEV_PRECISE |
                             BRIDGE_DEV_BARRIER)) ){
		bad &= ~fix;
		/* don't change these bits if
		 * they are already set in "old"
		 */
		chg &= ~(fix & old);
	    }
	    /* some conflicts can be resolved by
	     * forcing the bit off. this may cause
	     * some performance degredation in
	     * the stream(s) that want the bit on,
	     * but the alternative is not allowing
	     * the new stream at all.
	     */
	    if ( (fix = bad & (BRIDGE_DEV_WRGA_BITS |
			     BRIDGE_DEV_PREF)) ) {
		bad &= ~fix;
		/* don't change these bits if
		 * we wanted to turn them on.
		 */
		chg &= ~(fix & new);
	    }
	    /* conflicts in other bits mean
	     * we can not establish this DMA
	     * channel while the other(s) are
	     * still present.
	     */
	    if (bad) {
		pcibr_unlock(pcibr_soft, s);
#if (DEBUG && PCIBR_DEV_DEBUG)
		printk("pcibr_try_set_device: mod blocked by %R\n", bad, device_bits);
#endif
		return bad;
	    }
	}
    }
    if (mask == BRIDGE_DEV_PMU_BITS)
	slotp->bss_pmu_uctr++;
    if (mask == BRIDGE_DEV_D32_BITS)
	slotp->bss_d32_uctr++;
    if (mask == BRIDGE_DEV_D64_BITS)
	slotp->bss_d64_uctr++;

    /* the value we want to write is the
     * original value, with the bits for
     * our selected changes flipped, and
     * with any disabled features turned off.
     */
    new = old ^ chg;			/* only change what we want to change */

    if (slotp->bss_device == new) {
	pcibr_unlock(pcibr_soft, s);
	return 0;
    }
    bridge->b_device[slot].reg = new;
    slotp->bss_device = new;
    bridge->b_wid_tflush;		/* wait until Bridge PIO complete */
    pcibr_unlock(pcibr_soft, s);
#if DEBUG && PCIBR_DEV_DEBUG
    printk("pcibr Device(%d): 0x%p\n", slot, bridge->b_device[slot].reg);
#endif

    return 0;
}

void
pcibr_release_device(pcibr_soft_t pcibr_soft,
		     pciio_slot_t slot,
		     bridgereg_t mask)
{
    pcibr_soft_slot_t       slotp;
    unsigned long           s;

    slotp = &pcibr_soft->bs_slot[slot];

    s = pcibr_lock(pcibr_soft);

    if (mask == BRIDGE_DEV_PMU_BITS)
	slotp->bss_pmu_uctr--;
    if (mask == BRIDGE_DEV_D32_BITS)
	slotp->bss_d32_uctr--;
    if (mask == BRIDGE_DEV_D64_BITS)
	slotp->bss_d64_uctr--;

    pcibr_unlock(pcibr_soft, s);
}

/*
 * flush write gather buffer for slot
 */
static void
pcibr_device_write_gather_flush(pcibr_soft_t pcibr_soft,
              pciio_slot_t slot)
{
    bridge_t               *bridge;
    unsigned long          s;
    volatile uint32_t     wrf;
    s = pcibr_lock(pcibr_soft);
    bridge = pcibr_soft->bs_base;
    wrf = bridge->b_wr_req_buf[slot].reg;
    pcibr_unlock(pcibr_soft, s);
}

/* =====================================================================
 *    Bridge (pcibr) "Device Driver" entry points
 */


/*
 *    pcibr_init: called once during system startup or
 *      when a loadable driver is loaded.
 *
 *      The driver_register function should normally
 *      be in _reg, not _init.  But the pcibr driver is
 *      required by devinit before the _reg routines
 *      are called, so this is an exception.
 */
void
pcibr_init(void)
{
#if DEBUG && ATTACH_DEBUG
    printk("pcibr_init\n");
#endif

    xwidget_driver_register(XBRIDGE_WIDGET_PART_NUM,
			    XBRIDGE_WIDGET_MFGR_NUM,
			    "pcibr_",
			    0);
    xwidget_driver_register(BRIDGE_WIDGET_PART_NUM,
			    BRIDGE_WIDGET_MFGR_NUM,
			    "pcibr_",
			    0);
}

/*
 * open/close mmap/munmap interface would be used by processes
 * that plan to map the PCI bridge, and muck around with the
 * registers. This is dangerous to do, and will be allowed
 * to a select brand of programs. Typically these are
 * diagnostics programs, or some user level commands we may
 * write to do some weird things.
 * To start with expect them to have root priveleges.
 * We will ask for more later.
 */
/* ARGSUSED */
int
pcibr_open(devfs_handle_t *devp, int oflag, int otyp, cred_t *credp)
{
    return 0;
}

/*ARGSUSED */
int
pcibr_close(devfs_handle_t dev, int oflag, int otyp, cred_t *crp)
{
    return 0;
}

/*ARGSUSED */
int
pcibr_map(devfs_handle_t dev, vhandl_t *vt, off_t off, size_t len, uint prot)
{
    int                     error;
    devfs_handle_t            vhdl = dev_to_vhdl(dev);
    devfs_handle_t            pcibr_vhdl = hwgraph_connectpt_get(vhdl);
    pcibr_soft_t            pcibr_soft = pcibr_soft_get(pcibr_vhdl);
    bridge_t               *bridge = pcibr_soft->bs_base;

    hwgraph_vertex_unref(pcibr_vhdl);

    ASSERT(pcibr_soft);
    len = ctob(btoceg routines
 *      are callf_pthingt sones
pagt brunsigner_softe a wr= v, vhphys(en, m_t, io) ((__pss;
    vt(pc_base;
+prot),       nt to wid of pci*/

  rotg &=bet wanent dr"
 *      nor dn DMA
 forn/ioid of pcex_ue respofthis ar been succeedse respofthis may
id of pchlimited to theenabl    *
  swe haveBRIDG, interrlue, with thA
 forn/iowe havethintento get rid of pcXXX- thare_REQ */
voi endiannessed when Did of pcrealhA
 forred by ..or_code & IOECODE_We a wr&&
	((IDGE
	slotp->bEXTERNAH_WR_EN) || (m cal >slotp->bEXTERNAH_WR_EN)       error;
    devfs_handlesoft	ached*adress in ioewe thingsresprnabling.   *urneir.
	    t(voi	he "se * How o streamev, aninal valore the _    * siDriver"watus ibr_ft);splhi()    mask;t until {BRIDGE;
    if (fLASH_WR_EN"},
      mask;t until {BRIDGE_pthinin(pci     bug"wat ibr_fplx(s)val;
}

/*
 * pcibr__t     ibr_map(devfs_handle_t dev,handl_t *);
int     _t off, size_t len           pcibr_vhdl = pcibr_soft->bs_vhdl;
  hdl);
    pcibr_soft_tn int            t          pcibr_soft = pcibr_soft_get(pcibr_vhdl);
    bridge_t               *bridge = pcibr_soft->bs_base;

    hwgraph_vertex_unref(pcibr_vhdl);

    ASSERT(pcibr_soft);
    wid of pci*/A
 fon/iowninal the interruptd off.
 i_t  s specifiton table fonew =handlor_code & IOECODE_ mask;t until {BRIDGE;&   if (fLASH_WR_EN"},
        error;
    devfs_handlesoft	ached*adress in ioewe thingsresprnabling.   *urneir.
	    t(voi	he "se * How o streamev, aninal valore the _    * siDriver"watus ibr_ft);splhi()    mask;t until {BRIDGE;;

    chg LASH_WR_EN"},
      mask;t until {BRIDGE_pthinin(pci     bug"wat ibr_fplx(s)val;
}

*
 * pcibr_
pcibr_mo, and wi    ipciOR" b{
	deat plan g/xtao find an edge,s and mke
ifiton twhich USED ge.  Seef this fu*/
 , not a
 */tnabled wi be in BRIDe sufficig_p   ibr  bridge_t  her_flush(pcib   pcibr_t *);
int     _t obr_sof        free;       /* 1 if fret o, ch    DEVNAME]         pcibr_vhdl = hwgraph_courn        pciio_info = pciio_info_get(pconn       pciio_slot = pciio_infoget(pci &= (pSLOTNABLEt);
   _handle_t , char *r_code)t o, ch,    DEVNAMEft);
    w run*     alatilhis canongnos
path specifite */
wess(es)around  pcibr    cibr_or_code & IOECurn 
  hdl);
    pcibr_soft_tt obr_sofr_req_b     (urn 
!= GRAPH_Vn =EXNABLE soft, st(pconn_vhdl);
    pcchk(urn )bridge) {
rror_codvice attget(pciio_info);
    pcibr_soft_t       	    *beakdd32 |ibr_vhdl);

    ASSERurn )briurn 
  hdl);
    pcibr_soft_ttt         }ef(pcibr_vhdl);

    ASSERtt     
*
 * pcibr_   p  ibr_map(devfs_handle_t dev,nt, void *, int, str_t ofr_unlocerrcmdfr_unlo_t, ioargfr_unlocerr*crp)ce attgvoid        crfr_unlocerr*r(pcp           pcibr_vhdl = pcibr_soft->bs_vhdl;
  hdl);
    pcibr_soft_tn int           t     		pcibr_slot_soft->bs_v_soft = pcibr_soft_get(pcibr_vhdl);
    bridge_t           ver_regunlocerrrrrrrrrrrrrrrrrrrrrre a wr= 0nref(pcibr_vhdl);

    ASSERT(pcibr_soft);
   (e_sta (cmddvic		pcibr_slot_softOR" bGIOCSETBW:
	ice attg/xt,nt, vo_info = pciioconn  	     pciio_slot = pciio_infoget(pci0nre_unlock(p!c pad);

          CAP_Bfine UMGT      	e a wr= EPERM_d64_beakdd3    }e_unlock(prector(argf & lines{
  of(g/xt,nt, vo_info )      	e a wr= EFAULT_d64_beakdd3    }e		pcibrGRew =init\	     #endif

    ::able
br_so: %sprnqbw: %lldvicek)
{
  _inf.ble
br_so, _inf.rnqbw   ver_reng bitsGRew =initr_dmawr_erck(p(get(pciio_flush(pcib   pcibr__inf.ble
br_so)E_CAk)
 &= (pSLOTNABLE     	e a wr= EIO_d64_beakdd3    }e_unlock(p_inf.rnqbw oft, s);
_soft_t, pciio_slot_t, pciioe[slot]. &=_PRew HIGH   	    *beakdd32 _softOR" bGIOCRELEASEBW:
	ice attg/xt,nt, vo_info = pciioconn  	     pciio_slot = pciio_infoget(pci0nre_unlock(p!c pad);

CAP_Bfine UMGT      	e a wr= EPERM_d64_beakdd3    }e_unlock(prector(argf & lines{
  of(g/xt,nt, vo_info )      	e a wr= EFAULT_d64_beakdd3    }e		pcibrGRew =init\	     #endif

    ::able
br_so: %sprnqbw: %lldvicek)
{
  _inf.ble
br_so, _inf.rnqbw   ver_reng bitsGRew =initr_dmawr_erck(p(get(pciio_flush(pcib   pcibr__inf.ble
br_so)E_CAk)
 &= (pSLOTNABLE     	e a wr= EIO_d64_beakdd3    }e_unlock(p_inf.rnqbw oft, s);
_soft_t, pciio_slot_t, pciioe[slot]. &=_PRew LOW   	    *beakdd32 _softOR" bist_p  LOTNSTARTUP:
	ice att {
    pci
extern int	siio_infointnre_unlock(p!c pad);

CAP_Bfine UMGT      	e a wr= EPERM_d64_beakdd3    }e= bad & (BRIDGE_Drector(argf &intes{
  of(int)     rrrrrrrrrrrrrrrre a wr= EFAULT_drrrrrrrrrrrrrrrr_beakddrrrrrrrrrrrr}re_unloe a wr= s_handle_t, pcibr_sT(pcibr_sof &int   	    *beakdd32 softOR" bist_p  LOTNSHUTDOWN:
	ice att {
    pci
extern int	siio_infointnre_unlock(p!c pad);

CAP_Bfine UMGT      	e a wr= EPERM_d64_beakdd3    }e= bad & (BRIDGE_Drector(argf &intes{
  of(int)     rrrrrrrrrrrrrrrre a wr= EFAULT_drrrrrrrrrrrrrrrr_beakddrrrrrrrrrrrr}re_unloe a wr= s_handdle_t, pcibr_sT(pcibr_sof &int   	    *beakdd32 softOR" bist_p  LOTNQUERY:
	ice att {
    pci
extern int	siio_infointnre_unlock(p!c pad);

CAP_Bfine UMGT      	e a wr= EPERM_d64_beakdd3    }e= bad & (BRIDGE_Drector(argf &intes{
  of(int)     rrrrrrrrrrrrrrrre a wr= EFAULT_drrrrrrrrrrrrrrrr_beakddrrrrrrrrrrrr}rerrrrrrrrrrrre a wr= handle_t, pcibr_sT(pcibr_sof &int   	    *beakdd32 kbrick widgetnum-to    cibault:
4_beakddal;
}

/*
 * pcibr__t     ibr
#if DEBUG  *, iopaddr_t *, iopaddr__dmamex_upek)
{
  _, iopaddr__dmamnew pek)
{
  _, iopaddrex_uek)
{
  {
    i{
           pcibr_dmamap_pciaddr_ge_dmamex_ur= __dmamex_up;      pcibr_dmamap_pciaddr_ge_dmamnew ==r__dmamnew p;      pcibr_dmamap_pciaddr_genew ==rex_ur+      - 1GE_DEV_PMU_(new =<e_dmamex_u) ||(pci_ur>e_dmamnew lf_pdesillociopad    sidd an na_device =->bssPMU_(ci_ur<=e_dmamex_u) && (new =>=e_dmamnew lf
pdesillociopad {BRIaine intird an na_dev	__dmamex_up==r__dmamnew pr= 0nref(pc->bssPMU_ci_ur<=e_dmamex_u)
pdesillociopad  wienablof an na_dev	__dmamex_up==rnew =+ 1GE_DEV_->bssPMU_new =>=e_dmamnew l
pdesillociopad  wiif (fof an na_dev	__dmamnew pr= ci_ur- 1GE_DEV_ wid of pcWd an elefave rootwogerouons:this flloc beaid of pcf this an na_"opera"r seleopade resphis flloid of pc beacf this an na_"above"r seleopad. Keep specifitoeless  can gs biggint
ppecifevice =->bssPMU_(ci_ur-e_dmamex_u) > (_dmamnew =- new lf
p__dmamnew pr= ci_ur- 1G bitskeep wish td    ner_softeate_	__dmamex_up==rnew =+ 1G bitskeep upph td    ner_}    pcibr_info   pcibr_inf, char *buf, uint bufleARGSUSED */
int(  pcibr_infopciio_info);
ft_t       }    pcibr_info   pcibe_t, pciiinfnew(
handle_t);   pciio_slot_t slot,	    bridgereg_t mask)
	    brioid        roid sk)
	    brivbrig_t)0;
}vbrig_sk)
	    brie_t, pcidtr_t o, p_t       slotp_info = pciio_info_geotp_inf       pciiooid        addr_ge_id ;gunlocerrrrrrrrrrrrrrrrrrrrrripciGE_DEV__id ~BRIr_id ~Bci &= (pFUNCNABLE  ? 0 : roid GE_DEV_NEW(  pcibr_in)br_soft, sbrie_t, pciiinfnew(&  pcibr_in->f_ sk)
	     hwgraph_vertr_sofk)
	  slot].roid sk)
	 }vbrig_s_t o, p_;ce: atfgr- tswap" stuw ..o..ner_softthe etround u stumberibr_soft, s);
r_in->f_ u s=     ath, vh
extern) {
		badpmu = sr_soft);
   = new) {
!=  &= (pSLOTNABLE    t	ached*anything tendianed ar been   */
 &=hed*aw) {
tumberirespINTA/B/C/D andDriverhed*aoundIr.
	     Bit Number:hed*hed*aaaaaSLOTaaaaaA B C Dhed*aaaaa 0 aaaaa 0 4 0 4hed*aaaaa 1 aaaaa 1 5 1 5hed*aaaaa 2 aaaaa 2 6 2 6hed*aaaaa 3 aaaaa 3 7 3 7hed*aaaaa 4 aaaaa 4 0 4 0hed*aaaaa 5 aaaaa 5 1 5 1hed*aaaaa 6 aaaaa 6 2 6 2hed*aaaaa 7 aaaaa 7 3 7 3hed*hed*aXXX- or piops_handle_t andect of t cibaulthed*aXXX- or pioADMIN andect of t ps_handle_tus ibr_    (ipcir= 0nripci=<e4; ++ipci) 	     pc);
r_in->f_ipci[ipciFLAS		ew) {
+ 4 *ripci) & 7oft	ached*aRre.
 /

    focf this spars */
st   foc_t,pc.us ibr_IT)
	
st <    hwgraph_vertibr_lock(p.ibr_nr_in) 	     pc);
raph_vertibr_lock(p.ibr_r_ins[	
st]r= handle_inf      

*
 * pcibr_handle_inf  }======FIXME: _    npiotswap" stuff,
 by boroohandl.csabled* handle_t, .c.V_Nolding s(es)arbettI want,le fon/

#defiof in conses
 * thfor tvt viafor, s);
_so#defindevice: 
d*aoun_ADDp  PAe ULIMITSNSTORE
d*	 etble focything {
    {of
d*	t, s   cex_ue 
d*	t, s   cnew ,
d*	t, s  piomemfor ex_uekd*	t, s  piomemfor new ,
d*	t, s highomemfor ex_uekd* 	t, s highomemfor l

#def/TIMEOUT_Houn_ADDp  PAe ULIMITSNSTORE())
	\
     pc);
raph_vertip_inf.bci_   ex_ur= bci_   fb;	\
     pc);
raph_vertip_inf.bci_   new ==rbci_   fl;	\
     pc);
raph_vertip_inf.bci_swin ex_ur= bci_l  fb;	\
     pc);
raph_vertip_inf.bci_swin new ==rbci_l  fl;	\
     pc);
raph_vertip_inf.bci_mem ex_ur= bci_hi fb;	\
     pc);
raph_vertip_inf.bci_mem new ==rbci_hi fl; called o  pcibe_t, pc ASSous to
d*	, and_dmass paurnehardwan ecingurordef (brve _    * andounde_t, pkd* 	resprn tvtsaurneound_inraciio_pioSeef 
	         eoada    is ne*	, andandusuosticat plevice(xe sanof , pchg) {ibr_
    sizeund ard   ibrdle_t dev,e_t, pc ASSous tohar *buf, uint bp  pcbr_sof        pciio_info 	  pciio_inf         pcibr_vhdl ndle_t);r_so       pciio_slot 	_   p       pc);
raphl ndle_t);raph      *bridge 		ong          cerrrrrrrrrrrrrrrrrrr pciio_p->b0,r pciio_p->b1      s;
    vooooooooooooos      cer)
	 t_t, inost      cer)
	 t_t, r= 0nref(pc, st(pconn_vhdl);
    pcern) { pcbr_sofbr_soft, s);
hdl;
  dl);
    pcmas tocibr_soft_t       ooooset(pciio_info);
    pcibr_soft_t             pc);
raph_vhdl);
    bridge_t               *bridg

    hwgraph_vertex_unref(pct an
ext or   sizhardwan externecingurordechar                   xtern externt ob, pcibr_sT(pcibraph_vert{ pce[slotft);
    w Fr foror   sizxteribr_soft, s);
t);

stati { pcbr_sofbr_soft wid of pci*/

  RRB ned, uuibr drichar     set(phlimp->bssion.eg bacid of pcexad r dn DMboot-e sancibaulthd of p& IOECODE_T(pcibraph_vertt);
_slot_dfl_lock(p=>=e0    toooooooost);

    if (mask == BRIDGE_DEV_
    w Flloc sizxtericiio_a bits. Re   set(px = bad & (t_t, inost  de_t *, pciio_slot_t, inte[slot].k)
{
                     T(pcibraph_vertt);
_slotlock(p=+k)
{
                     T(pcibraph_vertt);
_slotlock(
+ drrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrist_p RRBpSLOTNVIRTUAL]DGE_DEV_
   ODE_t_t, inost)errrrrrrrrrrre a wr= ERANGR_NO64errrrrrrrrT(pcibraph_vertt);
cinlock(p=
    hwgraph_vertt);
cinlock(p=+drrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrr T(pcibraph_vertt);
_slotlock(p=+krrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrr T(pcibraph_vertt);
_slotlock(=+krrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrr ist_p RRBpSLOTNVIRTUAL]GE_DEV_
    pciio_p->b0=
    hwgraph_vertt);
_slot_dfl_lock(p;_DEV_
    pciio_p->b1=
    hwgraph_vertt);
_slot_dfl_lock(=+krrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrr rrrrrrrrrrrr ist_p RRBpSLOTNVIRTUAL]GE_DEV_
   p s);
}

/* =================_DEV_
   p s);
t, pciio_s { pcbr_so, & pciio_p->b0,r& pciio_p->b1)ddal;
}

/*
 * w Fr forv, aninal tic vos !!ner_softe a winost  dp s);
w);

stati { pcbr_sofbr_softODE_t_t, inost)errrrrrrre a wr= t_t, inost  ef(pct an
ext 

    fo            if (fs. Returet(px = bade a winost  dp s);
ndle_t,pciio_slT(pcibr_sof slotft);
   ODE_t_t, inost)errrrrrrre a wr= t_t, inost  ef(pc pcibr_t_t,        ap/munmap t dev,eIDGET_PARt, rrb_al this is CDLr.
 */cor   s int
pcibr_ichareacon       fers. f this  &=hinit beforous ry  can    cheble fovbrig_/       ID  support,
 by hinit befttion shou=bet waSSous toed.unction     'sd  pcibr    );

   this is aesphis n shou'sdpciio_nt
pcibr_i pcibr_ tatus interrlue, whis is slot'          tatus e have.egoid)
{
#if DEBUG eIDGET_PARt, rrb_alhar *buf, uint bp  pcbr_sosk)
	 }cerrkey1,}cerrkey2,}cerrt_t,          pciio_info 	  pciio_inf       slotp_info = pciio_ihandle_inf         pcibr_vhdl ndle_t);r_so       pciio_slot 	_   p       pc);
raphl ndle_t);raph  ef(pct ate Gath.eg ret(p tatus charvbrig_/       IDr.
 d ard n shous p& IOECODE_(key1~Bci-1) ||(pkey2~Bci-1))errrrrrrr pcibrnref(pc, st(pconn_vhdl);
    pcern) { pcbr_sofbrio_ihandle_infr= handle_infcern) { pcbr_sofbr_soft, s);
hdl;
  dl);
    pcmas tocibr_soft_t       ooooset(pciio_info);
    pcibr_soft_t             pc);
raph_vhdl);
    bridge_t           ef(pct aTome perfbeoaded.
 *
 *      Ts cnpad    aurne_t ned, uerit.
 s p& IOECmr
/* ===========_verteus_opade MR_UPDADG, P OK.         pc);
r_in->f_pci pci_e a wr= t_t,         pc);
raph_vertibr_lock(p.le_t, pctus ;

  LOTNSTATUS_MASKt);
   ODE_t_t,     rrrrrrrr pc);
raph_vertibr_lock(p.le_t, pctus |=  LOTNSTARTUP_INCMPLT_drrrrtp->bss_prrrrrrrr pc);
raph_vertibr_lock(p.le_t, pctus |=  LOTNSTARTUP_CMPLT_drrrrtprrrrrrrref(pct aRt_t pcdn DMbuscnpad p& IOECmr}

/* ===========_verteus_opad    p/munmap t dev,eIDGET_ ASSot, rrb_al this is CDLr.
 */cor   s int
pcibr_ichareacon       fers. f this  &=hinit beforous ry  can    cheble fovbrig_/       ID  support,
 by hinit befttion shou=bet wa ASSous toed.unction     'sd  pcibr    );

   this is aesphis n shou'sdpciio_nt
pcibr_i pcibr_ tatus interrlue, whis is slot'          tatus e have.egoid)
{
#if DEBUG eIDGET_ ASSot, rrb_alhar *buf, uint bp  pcbr_sos drrrrrrrrrrrrrrrrrrrrrrrrrrrrcerrkey1,}cerrkey2,}cerrt_t,          pciio_info 	  pciio_inf       slotp_info = pciio_ihandle_inf         pcibr_vhdl ndle_t);r_so       pciio_slot 	_   p       pc);
raphl ndle_t);raph  ef(pct ate Gath.eg ret(p tatus charvbrig_/       IDr.
 d ard n shous p& IOECODE_(key1~Bci-1) ||(pkey2~Bci-1))errrrrrrr pcibrnref(pc, st(pconn_vhdl);
    pcern) { pcbr_sofbrio_ihandle_infr= handle_infcern) { pcbr_sofbr_soft, s);
hdl;
  dl);
    pcmas tocibr_soft_t       ooooset(pciio_info);
    pcibr_soft_t             pc);
raph_vhdl);
    bridge_t           ef(pct aTome perfbeoaded.
 *
 *      Ts cnpad    aurne_t ned, uerit.
 s p& IOECmr
/* ===========_verteus_opade MR_UPDADG, P OK.         pc);
r_in->f_pci pci_e a wr= t_t,         pc);
raph_vertibr_lock(p.le_t, pctus ;

  LOTNSTATUS_MASKt);
   ODE_t_t,     rrrrrrrr pc);
raph_vertibr_lock(p.le_t, pctus |=  LOTNSHUTDOWN_INCMPLT_drrrrtp->bss_prrrrrrrr pc);
raph_vertibr_lock(p.le_t, pctus |=  LOTNSHUTDOWN_CMPLT_drrrrtprrrrrrrref(pct aRt_t pcdn DMbuscnpad p& IOECmr}

/* ===========_verteus_opad    p/mun hinibu
 doadnedvbrieup olink
path f thislot on m
   ".../<iobrick>/eus/<eusnum>"lot danger were 1ericsuccess, 0till pwiseregistedepe  no
   br_vhdh.eparat wr== '/'  ibrdle_t dev,eus_cnvlinkhar *buf, uint bf_ s}cerr                 freedst    DEVNAME]  	int   dp==rdst;          free*cp, *xp;         cerr xtalk_in;          freet deus[8]  	ar *buf, uint bnvtxf svtx;   errrvint		p =init\	
#endif

    xeus_cnvlink:r    = %spf_ = %evice[k)
slot].f_    	   	     os;
	   reed, ch 256]  		 os  devffnt);.  S(pciath(f_ s}d, ch, 256);t_device: m%s :
path= %svice[_pFUNCTION__,r&d, ch  os]DGE32 kbricksome coGRAPH_SUCCESS
!= ibr_vhdl);

   , chridgef_ s}dst,    DEVNAMEf)}
    }
  0oft	ac}dst exaock(r== /hw/module/001c02/Pbrick/xtern/8/
  /dirrive_dmawdesiresphis  xtalkstumberibr_	xp==rs rs r(dst, "/"Echg LBL_XTALK"/")bridge)xp=== NULL)}
    }
  0of	 xtalk_in==rat i)xp+7)bridge) xtalk_in=< XBOW_PORT_8 ||( xtalk_in=> XBOW_PORT_F)}
    }
  0oft	ac}rn tvt "/
  /dirriv"   */
path br_	cp==rs rs r(dst, "/" Echg LBL_eund"/" "dirriv")bridge)cp=== NULL)}
    }
  0of	*cp==r(  re)NULLoft	ac}alkse fovb

  _    * si xtalksibr_IT)
GRAPH_SUCCESS
!= ibr_vhdltrahouse(NULLs}dp, &svtx))	}
    }
  0oft	*xp==r(  re)NULLo		ac}rn tvt "/xtern/..."   */
path br_t	ac}dst exaock(rnpio== /hw/module/001c02/Pbrick br_t	ac}alkse fo u stumberibr_softoooostrcat(dst, "/ u ");         seviceERT(pcus, "%dce[p_eusnum[ xtalk_in]DGE_	ac}link
e haus e h xtalksibr_r 
  hdl);
   ath,add(NULLs}dp, &nvtx)bridge)GRAPH_SUCCESS
== rv)}
   
  hdl);
  e_addrdd(nvtxf svtx,et deusDGE_	 */
int(  
 = GRAPH_SUCCESS)  }=====ed once durinpciio_stem star    *xe sane focrosstern whis is _inraciio_piol valrythi e he_t, pc    ah xtalkhinit befttan    cheble fop havtumberiwizhandbits. Rethinit beforous r      lled, s abovet
pcibr_ap(devfs_handle_t dev,pciio_slot_t, int, ch x  pcbr_sof       t aREFERENCfs_han    l);
  e                 rc         pcibr_vhdl = hwgraph_connectpt_ge         pcibr_vhdl = hwgraph_coctltpt_ge      *bridge = pcibr_soft->bs_base;

 NULLof    *bridg;

    xmask = masid;gunlocerrrrrrrrrrrrrrrrrrrrrrrn        pcs_v_soft = pcibr_soft_get(pcibr       slotp_info = pciio_i _ihandle_inf      (BRIDGE__info = pciio_i _inf      (t          = pciio_i _i(t                 line_t, de pciio_i _it ob_t, ==r(   line_t, de)0;gunlocerrrrrrrrrrrrrrrrrrrrrr   p      cerrrrrrrrrrrrrrrrrrrrrripciGEi _it opcibr_vhdl = hwgraph_conaddr_t{ pcGEi _i free;       /* 1 if fret o,m    DEVNAME], *s       slotpget(devfs_hio_i _ihandleget(d       badd64;
    bridgereg_t     _interr      s;
    vooooooooooooooooiio_sixedr= 0nref(pc pcibr_dmamap_pciaddr_get(p_   fb,rbci_   fl;ef(pc pcibr_dmamap_pciaddr_get(p_l  fb,rbci_l  fl;ef(pc pcibr_dmamap_pciaddr_get(p_hi fb,rbci_hi fl; cf(pc errrrrrrrrrrrrrrrrrrrrr pl_* wri  		pcibr_slot_soft fre)
{
  *nic_infr= (       0river_regi		p oun_FBBEcf(pc errrrrrrrrrrrrrrrrrrrrrfaci_b_ale_t b_aleinterr  ver_regunlol1, de)
{
  *scp;     nass_node    nass_; cf(pcasync,pciio_dmamap_pciadaa

 NULLof
ciadaa

 async,pciio_dDGE__inf(x{ pcbr_sofbr_UG
    printk("pcibr_init\n");
#endif

    xpciio_stx{ pcbr_so=  %evice[x{ pcbr_sofbrio_i    err os;
	  reed, ch 256]  	 os  devffnt);.  S(pciath(x  pcbr_sos d, ch, 256);t_evice: m%s :
path= %s vice[_pFUNCTION__,r&d, ch  os]DGEio_i}iver_registerthe etuuck aroRB_    * si_base;
f tCONVEYOR BELT      *bad,e.roRB
#defi eoadainncibault FIRE-AND-FORGET      *bad,eor
 *   

    t, pc  ait(void    *an    hub_handle_t, pcibr_dx  pcbr_sos HUB_PIO_CONVEYORsoft->bs_base;

 (*bridge =*)}
(t    pio_t, device_x  pcbr_sos NULLsen/close es{
  of(*bridge ),e mmaisterth      *bCe	BRIse fovb

  _    * siound u , in consoid of pc.
 */
lsoher,
.
 *ange     requ
raph_abled featurn consselect * sinmas to"ovb

  _    or   sied featusoft_d  pcibr    cibr_ somessele->bsRIDGEis needed aTome tuffss.
 * ppe es
 *   w excep nic_handle_);

   _infneeded aa somedefi osanof * si*_vmcnt
pcibr_e tuff accessfs. Retue_adsto get rid of pcOpe static invb

  _w    if    * accessfs.id of pc* siDriver"erous to elegeselvdsto get ran    rc
  hdl);
   ath,add(x  pcbr_sos Echg LBL_eun,lot];

 r_sofbrio_i len = rc
 = GRAPH_SUCCESS)  _soft tltpt_ge

 NULLof     tltpt_ge

 hdl);
  SSous tohT(pcibr_sof Echg LBL_R\n"ROLLERlot_t slot)
{
  se esDEVFS_FL_AUTw =iV		   t_t slot)
{
  se es0  t_t slot)
{
  seS_IFCHR |eS_IRUSR |eS_IWUSR |eS_IRGRP,e es0  t_t slot)
{
  seot];

 fopss NULLft);
    len =  tltpt_ge
!= NULLft);
   th      *bde{
	de	    ice resp->bsRctedcouldRIDGEou
id of pc  pcibr    cibr_ is rstill p n shous DMA
gracking DMan g_or_code & 		pcibr_slot_softnic_infr=    chg Vn =EXNMFG_INFO(x  pcbr_sos (nic_r    i) &  mask;t unnicdriver_registerth      *bGlkse fohcerr iio_piol; ODE osanNIC , rrb_al t    *bmarthi ec invb

  _a s");
ex-IDG"/tnablsoid of pcjust  */
ints rs,es
 *   dostate new cal->bsto get ran    handleget(d_vhdl);
 t, int);
ex  pcbr_sos  mmaisterODE_T(pcibget(d_ntkT(pcibget(d->
  );
ex_rot)_	 */
int-1 we wang  */
  oper     sed off.
  device ==d[slot].reg;
  s_nid;gunlorn 
  X BRIDGE_WIDGREV			 (idmmaisterhdl);
  _infoadd LBLhT(pcibr_sof INFO LBL_eunBR_ASICGREVs (arpcirarybr_infopcr     
*
 *th      *bciio_a b raph_ pcter iio_piol, f    inn osa      *bfieldse resphook bacuuckolippevb

  to get ran    NEW(  pcibtoceg routiB OK.==============
  of *  pcibtoceg routi pcs_v_softsdge_t        ,   pcibtoceg r      pc);
raph_vert  pc
  x  pcbr_so routi pcs_v_sofu = sr_so_vhdl);
 r_so routi pcs_v_sofu = sex_ur= slot_t       slotp;
  _vertt ob_in==rrn        pcs_v_sof_vertvoid     r= handle_iid     E_DEV_D64_i slotp->b(*bridg)_soft, s);
_sof_vertvoi_S(pc     
    }

  INTERNAH_sloDEV_, s);
_sof_vertx_base;

 1_drrrrtp->bss_pt, s);
_sof_vertvoi_S(pc     
   }

  INTERNAH_sloDEV_, s);
_sof_vertx_base;

 device[sl     nass_

 NASID_IDG(*bridg)  ooooscp==r&NODEPDA( NASID_TO_COMPACTNABDEID(nass_) )->module->->bc       pcs_v_sof_vertl1, ==rscp;      pcs_v_sof_vertmodules_

 iobricktmodulet);
escpg routi pcs_v_sof_veri e  e_iidr= 0nref(pcwanDriverss pathrough REV C      *bc   unterrluoh.eg his nirriv      *bbytece ipto elo cibr_soft->or_code & IOECODE_ slotp;
  _vertt ob_in=<
   }

  _WIDGREV	C_soft, s);
_sof_vertp      _  pci &= (pcibr_soft->EV_, s);
_sof_vertp      _mempci &= (pcibr_soft->EVice[sl		p oun_p  OFT_LISTrio_i   , s);
lus fig_addr(devfsselfGE_	NEW(self);t_self_vel
raph_vhdl);
    b;t_self_vel
r_so_vhdl);
 r_so r_self_vel
nexh_vhdl);
 lus  r_self_vel
nexh_vhce i_ptr(m_t, io*)eot];

 lus , m_t, io)self);tice[sl	er_registerth      *balkse fo, chnof * gs bbase;
vb

  _aespkeep 

    fo. UsQ */
v      *bt to is rstit valorostictuff,
 now: likere a wrir.
	    sto get ran    s  devfe_t , cha_t        , t o,m,    DEVNAMEft)     pcs_v_sof_vert, chn= kmciio_s iilenrese+ 1, GFP_KERNEL)  oooostrcpy_ slotp;
  _vert, ch, sfbr_UG
 SHOWGREVS ||(_init\UG
 !_init\n");ODE_kdebug)    }
	}
evice: m%sDriver"ASIC:orn 
%s ({
	d=0x%x)Man %svice  	  slotp->b(*bridg) ? "X" :
"ce  	(rn 
 
   }

  _WIDGREV	A) ? "A" :  	(rn 
 
   }

  _WIDGREV	B) ? "B" :  	(rn 
 
   }

  _WIDGREV	C) ? "C" :  	(rn 
 
   }

  _WIDGREV	D) ? "D" :  	(rn 
 
 X  }

  _WIDGREV	A) ? "A" :  	(rn 
 
 X  }

  _WIDGREV	B) ? "B" :  	"unknowice  	r off slotp;
  _vert, chdriver_register_infr= (BRIDGE__info);
ex  pcbr_soft)     pcs_v_sof_vertxs_

 (BRIDGE__infos_nibr__infft)     pcs_v_sof_vertmas to

 (BRIDGE__infomas tocibr__infft)     pcs_v_sof_vertmxs_

 (BRIDGE__infomas tos_nibr__infft)isterth      *bItinesriver"opad.o get ran    spin npadd&& ATot];

    s = pcopad    sterth      *bIfiwizhaveless, to map se fohcersr iio_piolor_code & IOECODE_ slotphcers_softiio_sixedr= T(pcibget(d->
  iio_sixedGE_	 slotp;
  _verttio_sixedr= iio_sixedGE_	ODE_ slotphcers->
  _iid     ) 	     pc);
raph_vertvoid     r= handlehcers->
  _iid     GE_	    (get(pci0noset(p< 8; ++     vice att errrrrrrrrrrrrrrrrrrrrrhget(pciio_fluget(d->
  )osttibr_lock(pr- 1GE__unlock(phset(p< 0r_soft, s);
raph_vertibr_lock(p.)osttibr_ =r   p  	rrrrtp->bss_ptt, s);
raph_vertibr_lock(p.)as )ost

 1_dft, s);
raph_vertibr_lock(p.)osttibr_ =rh   p  	rrrrt== BRIDGE_DEV_th      *b.eg adaint, pc {
    {d
pcipcterfieldsr_code & IOEC    (get(pci0noset(p< 8; ++     vice, s);
raph_vertibr_lock(p.ridge->bo.ridde_t,pcpci &= (p PAe UABLEt)e, s);
raph_vertibr_lock(p.ridge64sex_ur= ntk("pc4_ucASE_UNSETt)e, s);
raph_vertibr_lock(p.ridge32sex_ur= ntk("pc2_ucASE_UNSETt)e, s);
raph_vertibr_lock(p.ridgexh_S(ps_rit.vur= ATOMICGINIT(0)evice[sl         (ipcir= 0nripci=<e8; ++ipci)s_pt, s);
_sof_vertvoir[ipciF.eri (t        r= 0nrt, s);
_sof_vertvoir[ipciF.eri handle_iid w);
.iw
raph_vhdl);
    b;t_, s);
_sof_vertvoir[ipciF.eri handle_iid w);
.iw
lus 

 NULLoft, s);
_sof_vertvoir[ipciF.eri handle_iid w);
.iw
rpct

 en/ccccc&_ mask;t unvoi_ pctus)oft, s);
_sof_vertvoir[ipciF.eri handle_iid w);
.iw
_sorcntr= 0nrt, s);
_sof_vertvoir[ipciF.eri handle_iid w);
.iw
rharedr= 0nrt, s);
_sof_vertvoir[ipciF.eri handle_iid w);
.iw
  pcibredr= 0nrice[sl     th      *bItin pc    varialloDriver"erous to t
ppecifevice =th      *bOnable-Rev.Desrivers,h.eg his oun_RETRY_CNT      *blo zerorogra_t, idro been stores. (#475347)r_code & IOECODE_rn 
<   }

  _WIDGREV	D)   mask;t uneus_e sa   a;

    chg BUS_oun_RETRY_MASKt);
   th      *bC
ext or  pe  een ir.
	    sto get ran     mask;t unvoi_rst
rpct

 ,
			   IRR_ALL_CLR    sterth      *bUe */
ill pwiseb.eg adless askeassumt bra ir.
	    sgrams,s aske  */
set(p7to get ran     mask;t unvoi_e->b_wid_( = pcibr) 0xffffffff   ster{   mask;4;
    bridgereg_tnirmapnrt,ibr_dmamap_pciaddr_ge   ibr_bridpcibr_dmamap_pciaddr_gexex_unr	x xtalk_in  = pciio_i _i(portbridpcibr_dmamap_pciaddr_gerotg &;   error;
    devfs_handle_in *   iesr= 0nrtcerrrrrrrrrrrrrrrrrrrrrre    ;
	 n
	ds_node n
	ds_;
	nass_node	nass_; c	the etr* siDriver's 32-b
     rogrXTern ed*aDirriveMap"erous tofs. Retumost
useful ed*ais the
  DMA
de tomine.V_Nlicit PClsoied*amust
useramseenrrlxs_
    or  of:hed*aaaaa nirriv-nent dr32-b
  singaccesseshed*aaaaa nirriv-nent dr64-b
  singaccesseshed*aaaaa singaccessesathrough his oMUhed*aaaaa ir.
	    si	he "sn table fot to iaydian_uar		 *cit PCI	 pc  ock(pam(s) .
	    sgw    reacona CPUhed*aaf tofor  singr   chlimitacoed aemforoi	he (Of  pcr_ue ts rstperfbeoadf */
   ipci	he n shous/{BRIDGEto eleatREAM or
 to manaerhed*a* gs orde *   to blem.)us ibr
	 n
	ds_r= 0nrcwancibault n
	d==d[ibr_nass_

 COMPACTNTO_NASID_ABDEID( n
	ds_)nrt,ibr_

 NBDE_OFFSET(nass_) + 0oft	ac}cything that
 just assumt  can gfiwizasn ed*a    o singar been ian"zero"r selXIO ed*a)ost
w    _t, dmuEQ */
voi ogramited to ed*a    e fonow to     nof aemforoi	her_	xex_ur= (t    dma_t, device_x  pcbr_sos 0sen/c	ge   ibr_,  _WGESZs  mmaiidge)xex_ur!=lXIO= ~BHERE vice att ge)XIO=PACKED)xex_u      	(port =lXIO=PORT)xex_u _dftxex_ur= XIO=ADDp)xex_u _df att}teate_		(port =l pcs_v_sof_vertmxs_GE__unlorotg &==exex_uS |
(1u   <<   }

  DIRMAP_OFF=ADDpSHFT)r- 1u   _df attxex_uS>>=   }

  DIRMAP_OFF=ADDpSHFTGE__unlonirmap==export <<   }

  DIRMAP_W_ID_SHFTGE__unlodge)xex_u)}
 nirmap=|=   }

  DIRMAP_OFFS |xex_unr	rrrre>bssPMU_rotg &=>= (512 << 20))}
 nirmap=|=   }

  DIRMAP_ADD512GE__unloreg);
#endiiv, vh  deirmapnrt}t	ached*a etrbriver's idea thesagt      accordeen ianRetur  whe'si	he idea the"IOesagt     ".uncBD:nctioidea theIOesagt     hed*awbe in orosticggraiayoi	her_	ached*adress in ioewe thingsresprnabling.   *urneir.
	    t(voi	he "se * How o streamev, aninal valore the _    * siDriver"watus ibr_fpl_* wri );splhi()  UG
 IOPGnew;
 
 4096   mask;t until {BRIDGE;;

    chg LASH__WGE new;		#elG
 IOPGnew;
 
 16384   mask;t until {BRIDGE;
    if (fLASH__WGE new;		#elte_	<<<Unterrluohdealhe rooIOPGnew;
>>>	    }
	}
 mask;t until {BRIDGE_pthinin(pci     bug"wat ibr_fplx(spl_* wriDGE_	ac}Itin pc    ir.
	s foar been *   iesribr_    (*    p= 0nr*    p<    hwgraph_vertvoi_S(pc    nr*    ++ vice att mask;t unvoi_S(pcram[*    ].w r= 0nrt}ft	ached*aDe tominepofthisrr's exh
	s foar been SSRAMo
  */
v ed*a mask;.unSeg adaDriver"{BRIDGE;erous tofr bro riS(p thi	he itin tc    SSRAMe resp.eg raphwan euuckolmanaer RAMhed*adr  iesrad
pciciio_a terrlcinguror.hed*hed*aCything that
 just upcdn DMrm*alled, sockolmanaer ATEhed*aaiio_a t(vo cWd wbe in to bterylcipl,pcp*/
vores tuhed*aB to Fitaaiio_a or.hed*hed*aF   npi, gfiwizhavelexh
	s foSSRAMe r_t, ied whhed*a*   ir.
	s fossram:e
  DMA'tchg &=GE_DEV_Po
 hed*awnablso upcdn DMir.
	s foSSRAM; respbesidds,hed*a* gs 
lsoh_uar		 *c eleatRnoaaiio_a t(v
w   hed*aw_t,ddlcdn DMir.
	s f/exh
	s foliss, sohsoied*aDMA
increming ATE thingsr   essesar*/
stalea hed*arri oca *   againet   }

  INTERNAH_sloDr    *hed*a* sa.us ibr_IT)
i slotp->b(*bridg)_}
 _in *   iesr= 0nrteate_		_in *   iesr= handle_iitgexh_S(pcram(*bridg)  _	ac}wizaliayszhavel128 ATEs (512     X*bridg) inei	de	   chiphed*advablIT)d off.
      debugsk;	.us ibr_   hwgraph_vertvoi_S(pc vh  drmciio_ndl_   hwgraph_vertvoi_S(pc    )oft, s);
S(pcio_slT(pcib======0ff slotp;
  _vertvoi_S(pc    )of		p oun_p ATE_=init\	
#endif

    xpciio_st%spINTERNAH ATEsvice[ slotp;
  _vertvoi_S(pc    )of	bricksome co_in *   iesr>[ slotp;
  _vertvoi_S(pc    )vic		p oun_p ATE_NOTBOTHwe wan    debug --his mrss ockolupcdexh
	s foS(pscibr_unlo
#endif

    xpciio_std off.een ir.
	s foATEs.et_dri	     pc);
S(pcciio_s {pcib====== slotp;
  _vertvoi_S(pc    )of	bricks	     pc);
raph_vertexh_S(pc vh  drmciio_ndl__in *   iesdri	     pc);
S(pcio_slT(pcib====== slotp;
  _vertvoi_S(pc    sen/clos_in *   iesr-f slotp;
  _vertvoi_S(pc    )of		p oun_p ATE_=init\	unlo
#endif

    xpciio_st%spEXTERNAH ATEsvicek)
{
  _in *   iesr-f slotp;
  _vertvoi_S(pc    )of	bricks	}rice[sl     {   mask;4;
    bridgereg_tnirmapnrtdpcibr_dmamap_pciaddr_gexex_unrt	ached*anpio, uuide	   *oros*externeex_uSr   esshed*a* atRnirmap=se  no ocko.us ibr_nirmap==ereg);
#endiiv, vhbridge)nirmap=&   }

  DIRMAP_OFF)df attxex_uS= (dpcibr_dm))nirmap=&   }

  DIRMAP_OFF)df		<<   }

  DIRMAP_OFF=ADDpSHFTnrteate dge)nirmap=&   }

  DIRMAP_ADD512)df attxex_uS= 512 << 20nrteate_	 attxex_uS= 0GE_	 slotp;
  _vertiiv,xex_ur= (ex_unrt	actit valintirdl pcissierrlueioewe mnt,levice
v ed*acibr_,zhaveleurRnirmap=cibr_een someis rs ed*ao/
stalea leurRnmas to"oport.us ibr_   hwgraph_vertiiv,xport =_	 att)nirmap=&   }

  DIRMAP_W_ID)
>>   }

  DIRMAP_W_ID_SHFTGEice[sl     /p t dev ngurordeanre a wrir.
	    ;      *bfiuuide   *is rstuoh.es. ftto get rid of pcIf*urneir.
	    
#defiinterru innt, inteid of pc* s this n/iow.eg ass paurdleurRir.
	    id of pchlimal* Horfbeenarri pcibredrinnmlf (backing DMabovet
pget rid of pcNolding .eg his D INTR_ISERRr*crpid of pcf this t ob_t, =at pl    or o_a twhen Did of pce a wrir.
	    , soheurRir.
	    
w   hd of pcbs n/iperrylc   ff ann tooft_t   dto get rid of pcIf*eurRcrosstern if    *r if (sfs.id of pc /*  xtalkse a wrir.
	    sing .   if (      *bdes tw     s, D INTR_ISERRr wiepioiacking DMknowlowed
 ice
vt
ppecifevice =(t        r= (t         ciio_sx  pcbr_sos d ob_t, ,hdl);
 r_sofbrio_i len = (t        r!= NULLft);
    pcs_v_sof_veri e  e_iidr= (t          ice =th      *bOnaIP35ores tXB, inte[we * To staexhra checkvoi e pcs_v_et xtdle_d of pcf torde  e h orgisters. o star   esstwhelimit     so cI torde id of pc    * evifird wor    h orgin/iperrye[we nolding  mke ess isoid of pcto hav  */
aMknown clear_ tatsto get ran    handleclear xtdle(*bridg)  _ce =(t        
  pcibr (t        ,  (t        v_ett
pcdm) pcs_v_et xtdle, m_t, io)*bridg)  _ce =th      *bnpio
  DMA
to havbr_vheen * a wrir.
	    s;      *binterrlor  ofelege needed aNOTE: o staound_ite perfal* Horfbeiinterruto get ran         _interr==ereg);
#end   _interr=|   }

  ISR_ERRORS; ca     eg);
#end   _interr==t     _interr       eg);
#end   _ad,eo= 0Ge wande Gath.es. "clearrir.
	    "opacketblfevice =/* wait until Bridge PIO complete */
    pcibr_unlock(pcibr__ce =th      *bDepe  een o this rn 
ofet, inte[d off.
 cerIain feapiolvt
ppecif Easi to iaydseems e haveng ss mr his oun_p NOw ev    id of pc crp e haver_icharor  sing, rr , in conect of tsaurnid of pcoun_p w ev    c crp harevs this _et twheof inev    id of pc /ioweis ounew =MAeclassc crps (harevs t /ioweis o/
stid of pcoun_pc crps===
up oNOnect of tsaYES)to get ran    handleraph_vertima_t, pcr= 0nref(pcwanGE_DEV_P:id of pcAliaysznlock(pcryld off.
      REV.A;      *bevi

    xprefetch_interrtt o"e reyonoid of pc skeen     ounew GE_DEV_Pogetblis needed aBet es thispcdnwo cibr_ , youzhavelograskid of pc    oun_p GE_DEV_P, in conn/ioisc eleatid of pcyeurRn     Tknowloabo  *known     pciWARsor_code & IOECODE_ slotp;
  _vertt ob_in=<   }

  _WIDGREV	B)r_   hwgraph_vertima_t, pcr|= oun_p NOGE_DEV_P;     eate dge) slotp;
  _vertt ob_in=<   	(
			    BRIDGE_WIDGET_ << 4=| 
    xprefetch_interrtt o))r_   hwgraph_vertima_t, pcr|= ounIO= ~GE_DEV_P; ef(pcwanBRIDG_GATHER:id of pcD off.
  uuckolnot Gath
uplud when Did of pcrevvtumberii e pcs_vwg_interrtt oao find_d of pcfsRnoa"WAR r>bss"aa soes tprefetchor_code & IOECODE_ slotp;
  _vertt ob_in=<   	(
			    BRIDGE_WIDGET_ << 4=| 
    xwg_interrtt o))r_   hwgraph_vertima_t, pcr|= oun_p NOBRIDG_GATHERnref(pc, st(pif    *r SSous tohT(pcibr_sof &
    xpr    *rft)     pct(pif    *r  pcibr_sT(pcibr_soft);
    pc_   fbr= 0x00000004 PIO cI/O FlloBnpad Bx_urran    han_   flr= 0xFFFFFFFF PIO cI/O FlloBnpad Lew =br__ce =t(p_l  fbr= 0x00000010Ge wanLpioMemfor FlloBnpad Bx_urran    han_l  flr= 0x001FFFFF PIO cLpioMemfor FlloBnpad Lew =br__ce =t(p_hi fbr= 0x00200000 PIO cHighoMemfor FlloBnpad Bx_urran    han_hi flr= 0x3FFFFFFF PIO cHighoMemfor FlloBnpad Lew =br__n    oun_ADDp  PAe ULIMITSNSTORE(); ef(pcwanbu
 do"no-set("c  pcibr    cibr_o get ran    handle_infr= handlee_t, pciiinfnew
	lT(pcib====== &= (pSLOTNABLE== &= (pFUNCNABLE,hed &= (pVENDOR_ID_ABLE== &= (pBfine UID_ABLEft)    naddr_t{ pcpciio_infe_t, pciiinfSSous to
	hT(pcibr_sof &
    xr_in->f_ ); ef(pcwanRememberie fo,o
set(p  pcibr    cibr_ _infr    *ea *   iacking DMibr_
r
 *   pciio_to get ran    handleraph_vertnaddr_t{ pcpcinaddr_t{ pcGEi _ihandleraph_vertnaddr_t_infr= handle_inf;i		p oun_FBBEcf(pcfaci_b_ale_t b_aleinterr

 1_dver_regi		p oun_FBBEcf(pc T)
	aci_b_ale_t b_aleinterr viceached*aAr  e_t, ps o this aus deficapterr
of 	acicexad r dexad, sohed*awe nolding .eg his 	acicexad r dexadhich icicii e_t, ps o hed*a*   aus e evideficapterr
of dostatsuconaccesses.us ibrf(pc}dver_regi		pcibr_slot_soft/pci*/

  sriver"hlimbeenarr.eg hisn ts rstfsRnoanolding f (backing DMn DMirdi   ualaoundddr_sor_code & IOEC    (get(pci0noset(p< 8; ++     v t	actRr.eg or   sizddr_s ibr_m_t, ) pcs_v_ern insdge_t        , slotft)ver_register    (get(pci0noset(p< 8; ++     t	actFirdleut inevtfsReut ts rstibr_m_t, ) pcs_v_ern _infos& AT_t        ,slotft);
       (get(pci0noset(p< 8; ++     v t	actSeg adahis a   essc_t,pcichar     set(pf this p s  ann ibr_m_t, ) pcs_v_ern ibr_d_t,pcos& AT_t        ,slotft);
       (get(pci0noset(p< 8; ++     v t	actSegadahis e->b_wierous tofibr_m_t, ) pcs_v_ern e_t, pciiige_t        , slotft);
       (get(pci0noset(p< 8; ++     v t	actSegada)ost/gd toierlat.
 s p& _m_t, ) pcs_v_ern gd to _infos& AT_t        ,slotft);
       (get(pci0noset(p< 8; ++     v t	actItin pc RRB manaerming ibr_m_t, ) pcs_v_ern _iin pc
t, pciio_s {        ,slotft);
   /e n shoudpciio_nlled, socwbe in betem stareut  /iowg  */
  heeuxb{
	de & IOEC    (get(pci0noset(p< 8; ++     v t	actCor   size->b_wipciio_nibr_m_t, ) pcs_v_ern em s e_t, pcpciio_s_t        , slot,e mmaisterth      *bEio_nPbrick ound u st to hlimddr_s 1 ann 2.   Similarryl         *b xtalks0xver_iIbrickso cAiio_a b RRB'sdpccordeenlyor_code & IOECODE_ slotp;
  _vertmodules_
> 0r_sof(e_sta (MODULE_IDG_BV_PAR_ slotp;
  _vertmodules_      OR" b'p':PIO cPbrick br_		e_t *, pciio_autociio_s {pcib======1, 8 _dfte_t *, pciio_autociio_s {pcib======2, 8 _dft*beakdd3OR" b'i':PIO cIbrick br_	  IO cport 0xver_i sizIbrick t to hlimddr_s 1 ann 2 br_		ODE_ slotp;
  _vertxs_

= 0xr vice		e_t *, pciio_autociio_s {pcib======1, 8 _dftte_t *, pciio_autociio_s {pcib======2, 8 _dft}  	e>bss_pttIOECIO cciio_a b ess RRB_    * sis*/
os
port br_		te_t *, pciio_autociio_s {pcib======0==1 _dft}  	*beakdd32 /e (e_sta ibrf(pc}di		pcibr_slot_soft= new rs r(nic_inf, XTALK_oun__WIDGET_      e_t *, pciio_autociio_s {pcib======1, 8 _d		p oun_p RRBp=init\	
#endf("\n\nFers. XTALK_oun (030-1275)Man %vvice[x{ pcbr_sofbr\	
#endf("
    xpciio_st%v Shoebox RRB_MANAGEMENTst%s+%spfllovicek)
_t        ,dft, s);
raph_vertiio_avail[0],dft, s);
raph_vertiio_avail[1]DGE_	    (get(pci0noset(p< 8; ++     t	i _ih#endf("\t%s+%s+%dcet	i _i0xFFF &ihandleraph_vertt);
_slotlock(pet	i _i0xFFF &ihandleraph_vertt);
_slotlock( + ist_p RRBpSLOTNVIRTUAL],s	     pc);
raph_vertt);
cinlock(pfbr\	
#endf("et_driver_regunlo}	#elte_	FIXME("
    xpciio_stCor  e_t *, pciio_autociio_tnic_infet_driver_reg_soft= neaa t	i _iasync,pciio_dadd _inf(naddr_t{ pc, aa br_soft, sbrie_t, pcpciio_snaddr_t{ pc,  mmai;
   /e       *bTearribr_
cibr_e  e hasyncipciio_n_infr--hasyncith* Hosl         *bbriver's _t, er_aite perfbetrun statbut/

  sriver's  orgi    essor_code & IOECODE_aa t	i _iasync,pciio_ddel__inf(x{ pcbr_sofbr_i _i   }
  0of}alled o  pcibe_iio_sne*	Dciio_n

  sriver"       f/ioweis  br_vhdhaf tofcleartwheoug or   sizne*	rs.errytwhevb

, ps   ibrdle_t dev,e_iio_slot_t, int, ch x  pc         pciio_slot 	   p         pcibr_vhdl n_t               pc);
raphl ndl);
    b;t    *bridge 		ong      ;
   /e Glkse fobbase;
vb

  _f/iow   rxterne  pcibr    cibr_  & IOECODE_ibr_vhdltrahouse(x{ pc, Echg LBL_eun,lot];

 r_sofr!= GRAPH_SUCCESS)_	 */
in(1ft);
    pcs_v_sof_vhdl);
    bridge_t               *bridg =l pcs_v_sof_vertex_unref(pct aD off.
dn DMir.
	    sif/ioweis *bridg ran     mask;t unvoi_interr

 0nref(pcwanDciio_nor   sizounde_t, pwiiflkeen ianRegs bbase;
 & IOEC   (get(pci0noset(p< 8; set(++ vic		pcibr=init\	
#endif

    x   line_tiio_nem star    %p/%dvicek)
_t        ,slotft)ver_regndl);
  ern e_iio_s_t        , slot,e mmaf(pc}dif(pcwanUnerous tofs fo,o-set(p  pcibr    cibr_ ran    haninfe_t, pciiinf ASSous toh_t        ,dftttI&_ slotp;
  _vert,addr_t_inf->f_ )ft);
   spin npadddes royTot];

    s = pcopad   
   kio_slT(pcib====_vert, chdri
   if(pcwanE a wrbr_vhdrogetbl ASSous toedawnabl* si xtalks  focfs       *bclearedar_code & IOEC w Flloc siz_sof_ATE mapse & IOECODE_ slotp;
  _vertvoi_S(pc vh)_	 mio_sndl_   hwgraph_vertvoi_S(pc vh); IOECODE_ slotp;
  _vertexh_S(pc vh)_	 mio_sndl_   hwgraph_vertexh_S(pc vh)nref(pct aD o  pcibr Retue a wrir.
	     ann flloc sizxternecingurordeid of pc ssocia bitoes tg_or_code & code(t        vd o  pcibr_   hwgraph_veri e  e_iid)      (t         io_slT(pcib====_veri e  e_iid)  ef(pct an
ext 

  raphwan eipctermaiRIain,
 by 

  sriver"  shoudchar          *bbriveror_code & codeDELlT(pcib====)      wanRemtvt * siDriver"ervis    labelrru infoc & codem_t, )hdl);
  _inforn tvt LBLhT(pcibr_sof INFO LBL_eunBR_ASICGREVs NULLft)    wanRemtvt * si  reac tofe->b_wipssocia bitoes tRegs bbase;
 & IOECm_t, )hdl);
  e_addrn tvthT(pcibr_sof Echg LBL_R\n"ROLLERl NULLft)    wanRemtvt * siound base;
vb

  _ & IOECm_t, )hdl);
  e_addrn tvthx{ pc, Echg LBL_eun,lNULLft);
    */
in(0)  }==dle_t dev,asictt ohar *buf, uint bp  pcbr_sof          pcibr_vhdl = hwgraph_connectpt_ge      arpcirarybr_infograph_coa_inf;i IOECODE_GRAPH_SUCCESS
!=
	ibr_vhdltrahouse(p  pcbr_sos Echg LBL_MASTERl ot];

 r_sof)_	 */
int-1 i IOECODE_GRAPH_SUCCESS
!=
	ibr_vhdl_info);
 LBLhT(pcibr_sof INFO LBL_eunBR_ASICGREVs &a_inff)_	 */
int-1 i IOEC */
int(dle)oa_inf;i}==dle_t dev,thing_g*/
st

statiar *buf, uint bp  pcbr_sof        pciio_info pc, st(pconn_vhdl);
    pcern) { pcbr_sofbrio_ihandle_soft =  pcs_v_sof_vh(handle_soft pciio_info);
m	aci_ibr_soft_t       oooo pciio_slot rr   p      set(pciio_info);
    pcibr_soft_t       io_ihandlee_t, pcthing_g*/
st

stati {pcib======slotft)i _i   }
  0of}amun =====================================================================hinit bbr_uMANAGEMENTdeviceipcticc pcibr_dm_t dev,abr_dhan_totxsohar *buf, uint bp  pcbr_sosk)
aph_connebridgereg_t mask)
aph_connebridt,pcotc_t,pcsk)
aph_codpcibr_dmahan_ibr_,k)
aph_co       reqc    sen/aph_cos;
    vot, pc_t       slotp_info = pciio_info_geotp_inf = handle_infcern) { pcbr_sofbr     pciio_info pcpciio_info_get(pconn_vh&
    xr_in->f_        pcs_v_soft = pcibr_soft_get(pcibr_vh(handle_soft pciio_info);
m	aci_ibr_soft_t       oooo*bridge = pcibr_soft->bs_base;

  pcs_v_sof_vertex_unref(pcs;
    voooooooooooooooobar;IO con concASEieroer_ie->b_wi     cod whe & IOECOpcibr_dmamap_pciaddr_gexifoaddr =lXIO= ~BHEREbr_soft, sbridt,pcotcccccccccccwdt,pc;IO con con_t,pcie->b_wi     cod whe & IOECOpcibr_dmamap_pciaddr_gewex_unIO cex_urof d->b_wide{
	der_ioundran    s                       w    nIO c     of d->b_wide{
	der_ioundran IOECOerrrrrrrrrrrrrrrrrrrrrr   ;	wanDcvIO(x)toenibr scan statorde  {BRIDGE; & IOECOerrrrrrrrrrrrrrrrrrrrrroen;IO con conDcvIO(x)toenibr gs bet wa s
  devsoft, sbridt,pcotcccccccccccmdt,pc;IO ctaralks_t,pcichare->bo(x)terous tofibrIOECOpcibr_dmamap_pciaddr_gemex_unIO cex_urof d->bo(x)tnent dran aer_ioundran    s                       m    nIO c     of d->bo(x)tnent dran aer_ioundran    s                       mmasknIO c     bctedcoaned innDcvb_w(x)tran IOECs;
    volo waaaaaaaaaaast);
   s_vhdl);
 lf (mask == BRIDGE_DEV_ODE_ slotp;
  _vertibr_lock(p.)as )ostr_sof(et(pciio_fluraph_vertibr_lock(p.)osttibr_oft, s);
_inf = handleraph_vertibr_lock(p.ridgr_ins[0]_drrrrtprrrr= newt,pcpcci &= (p PAe UABLE)_	gowed
 neGE_DEV_ODE_wt,pcpcci &= (p PAe UCFG viceached*aUsuosti,le fofirstoar beenhed*ads terishlding adounde_t, pk	 pcfsRe hets ned, uc_t,pc.us ik	 pcIn*urnecx_ue wancib_iinel*hed*adoaNOTanolding worryoabo  hed*aoundcASEierous to e reshed*aMUSTaNOTaattempiafor,ibr_oed*a*   DcvIO(x)toenibr a_oed*a* is accessf...us ibr_IT)
((t, pcr&i &= (pcibr_soft->)

= 0) &&_	 att)(han_ibr_ + reqc    )=<
   }

  TYPE0UCFGpFUNCNOFF))df attxifoaddr =lhan_ibr_ +   }

  TYPE0UCFGpDEV(slotft);	gowed
 neGErrrrtprrrr= newt,pcpcci &= (p PAe UROM viceacbbr_uianRetuExpans    Romoi	he D shoudvalorsponsierrl    	 *binterstate dtd off.eenhed*ade{
	dsin/iperry.us ibr_wex_ur= bci  xr_in->f_rex_unr	w     
 bci  xr_in->f_r    nrt	ached*aWhi.
dn DM      Tsbe in knowrbettI hed*a* aning attempiaformap=m*   wt,pched*a* aninhcie->b_wi     cod wh, etumighthed*adoait;rbettI wr dex*/
iutts rs.us ibr_IT)
(han_ibr_ + reqc    )=> w    )df attgowed
 neGE_	han_ibr_ +=ewex_unr_fp,pcpci &= (p PAe UMEMGErrrrtprrrrth      *bneducetoenibr ar beensRe hraw      *bfp,pcpar beensR(maybe or o_a twh      *b xnibrs)e resp   pcharDcvIO(x)      *busaer (_et twhein gfiit valavailterr to get ran     ar );sp,pcp-i &= (p PAe UWIN0; IOECODE_ ar < 6 vicewfp,pcpcibci  xr_in->f_ xnibr[ ar].widt,pcbridge) wt,pcpcci &= (p PAe UABLE)_	 attgowed
 neGE_	ac}alksound x_uSrs. o    ibr_wex_ur= bci  xr_in->f_ xnibr[ ar].wiex_unr	w     
 bci  xr_in->f_ xnibr[ ar].wid   nrt	ached*aWhi.
dn DM      Tsbe in knowrbettI hed*a* aning attempiaformap=m*   wt,pched*a* aninhcie->b_wi     cod wh, etumighthed*adoait;rbettI wr dex*/
iutts rs.us ibr_IT)
(han_ibr_ + reqc    )=> w    )df attgowed
 neGE_	O c hift  /iowoenibr erlat.veloghed*ade{
	dd;sp,pcperlat.ve.us ibr_   _ibr_ +=ewex_unr_fp,pcpciwdt,pc;Errrrtteate_	 ar );-1 i IOECactScpciciia*   DcvIO(x)toenibrsRewb_wilookeen     onoid of pce eviDMA
tat.sfy*eurRited toao fiofirstoe sane rougheid of pct to lookMan as
    vooenibrs; * sis*neddoe sa, 
lsoid of pclookMan  &= (p PAe UABLEooenibrs. Arr>bssweis orde id of pcsohsozaliayszlookMan eurRbr_
oenibr firstto get rid of pcWmesseleGathattempiafortat.sfy*amseenrrlited to d of pcby ned_a twa twhemultick(roenibrs.r_code & IOEC    (   p= 0nr   p< 16; ++   ) {   mask;4;
    bridgereg_tnev4;
nr_s;
    voooooooooooooooorotg &; f	 xcpci(   p+=slotf % 8GE_	ac}If * gs DcvIO(x)tar been an aeDMA
if    *hed*aagar been ian* is a   ess,lupcdit.us ibr_m     
 ( xcp< 2) ? 0x200000 : 0x100000 r_mmask );-m    nr_IT)
fp,pcp!ci &= (p PAe UIO)df attmmask &= 0x3FFFFFFF E_	rotg &==e   _ibr_ &R(m     -=1 _d_	ac}If * gs oenibr DMA'tccissieryrbr_vhdce eviited to,hed*agoer_i anRetunexh_oenibr.us ibr_IT)
((   _ibr_ &R(m     -=1 ) + reqc    )=> m    )df att{BRIinueGE_	nev4;
 = handleraph_vertibr_loenp.ridge->bceGE_	O cIs * gs oenibr "nx*/ voibr_"?k	 pcIfeGat, maybe 
  DMA
upcdit.us i (t to check * gs * sis*neddoe sane rough)us ibr_m p,pcpcibci  xraph_vertibr_loenp.ridge->bo.ridde_t,pcnr_IT)
(   p> 7) &&R(m t,pcpcci &= (p PAe UABLE)) { df attac}If * gs gs * siprima  pDcvIO(x)toenibrdf att pc     osanoll p n->bce, skipdit.us code & _unlodge)( xcp!==slotf &&_		( &= (pVENDOR_ID_ABLE
!=
		ibci  xraph_vertibr_loenp.ridgvbrig__s_  
	  BRIinueGE_	 attac}It'  a fllocoenibre respwiofi(pf tit.us code tSeg adaDcvb_w(oen)ckolippetas t.us code & _unlomex_ur= bci_ibr_ &RmmasknE_	 attac}check * ioewe te in orosticget  /ious code ts rstuohts rs.us code & _unlodge)(mex_ur|orotg &fr!= bci_ibr_ 
	  BRIinueGE_	 attnev4;
 ;

    chg DEV_OFF=MASKt)_unlodge)fp,pcp!ci &= (p PAe UIO)df	nev4;
 |=   }

  DEV	DEV	 (pMEMGE	rrrre>bsdf	nev4;
 ;

    chg DEV_DEV	 (pMEMGE	rrrrnev4;
 |= (mex_ur>> 20)=&   }

  DEV_OFF=MASKt)_	 attac}cibault gs cibr_soft->.us code tdgeyouz.   ify boll,us code toper      gs rs.efin,
.us code & _unlodge)t, pcr&i &= (pcibr_soft->)df	nev4;
 |=   }

  DEV	DEV	SWAPGE	rrrre>bsdf	nev4;
 ;

    chg DEV_DEV	SWAPGE _unlodge)handleraph_vertibr_loenp.ridge->bcep!cinev4;
 vice	reg);
#endi->bceloenp.4;
 = nev4;
nr_e, s);
raph_vertibr_loenp.ridge->bcep= nev4;
nr_e/* wait until Bridge PO complete */
    pcibr_unlock(pcibr__UG
    printkoun_=init\		
#endif

    aDcvb_w(%d): 0x%lxvice[oen, reg);
#endi->bceloenp.4;
)of	bricks	    }s	     pc);
raph_vertibr_loenp.ridge->bo.ridde_t,pc );sp,pc;s	     pc);
raph_vertibr_loenp.ridge->bo.riddeex_ur= mex_unr	rrrrxifoaddr =l   chg DEVIO(oen)c+ (   _ibr_ - mex_ufbr_UG
    printkoun_=init\	     vice: m%s LILE
%drmap=ng . ,pc %dn_t,pcie-sc 0x%x[%lx..%lx]c     et(p%dnciio_a bs DcvIO(%d)rnev4;
 0x%xvice[k)
    _pFUNCTION__,r__LILE__,r_t,pcsr_t,pcb_t, ,k)
    han_ibr_, han_ibr_ + reqc     -=1,k)
    slot,eoen, nev4;
 of	bricksom    gowed
 neGE	}	we wanbrickpDcvIO(x)tGathcibr_en ibr_mex_ur= bci  xraph_vertibr_loenp.ridge->bo.riddeex_uGE_	O cNbr Dheck     ited toh
up ocattoes tDcvIO(x) s ibr_IT)
(mfp,pcp!cifp,pc) ||_	 att)han_ibr_ < mex_uf ||_	 att)(han_ibr_ + reqc    )=> (mex_ur+ m    )f ||_	 att)(t, pcr&i &= (pcibr_soft->)
ntk!(nev4;
 ;   }

  DEV	DEV	SWAP)f ||_	 att)!(t, pcr&i &= (pcibr_soft->)
ntk(nev4;
 ;   }

  DEV	DEV	SWAP)f)df att{BRIinueGE_	wanDcvIO(x)toenibr gs cibr_en an  &= wt,pched*a* ath
upludps oppetaralk.tCorculBRIse fhed*afinallXIO a   ess,lrt_t pcdn DMnpad reshed*a   }
 oi	her_	xifoaddr =l   chg DEVIO(oen)c+ (   _ibr_ - mex_ufbr_UG
    printkoun_=init\	 vice: m%s LILE
%drmap=ng . ,pc %dn[0x%p..0x%p]c     et(p%dnusbs DcvIO(%d)vicek)
_pFUNCTION__,r__LILE__,r_t,pcsr han_ibr_, han_ibr_ + reqc     -=1, slot,eoen)of	bricks	gowed
 neGErrrrtp;
   se_sta (fp,pc) iceached*aAccessesato d->b_wide{
	dhed*aaoros e evidg adGathfi(hed*awiew ca*   DcvIO(x)t. ,pc ars ed*amod filding begaccessesaviaoed*a*   dirrivear been an as.us ik	 pcIfeGecessa  , n shous DMA 	 *biAM or
 to ask     ar beensk	 pcfnuohts ss a   essc_t,pcs,hed*abut/

iocwbe in nehou=bectuff,
.us ibrf(pcOR" b &= (p PAe UMEM:PIO c"memp_t,pc" ibrf(pcOR" b &= (p PAe UMEM32:PIO c"mem,lupcd32-b
 -wi	de u " ibr_IT)
(han_ibr_ +   }

  _un_MEM32ucASE + reqc     -=1)=<
df att  }

  _un_MEM32uLIMIT)df attxifoaddr =lhan_ibr_ +   }

  _un_MEM32ucASE;
	*beakddrf(pcOR" b &= (p PAe UMEM64:PIO c"mem,lupcd64-b
 -wi	de u " ibr_IT)
(han_ibr_ +   }

  _un_MEM4_ucASE + reqc     -=1)=<
df att  }

  _un_MEM4_uLIMIT)df attxifoaddr =lhan_ibr_ +   }

  _un_MEM4_ucASE;
	*beakddrf(pcOR" b &= (p PAe UIO:PIO c"i/op_t,pc" ibr	wanDriver Hardwan eBugiWAR #482741:hed*a fio4G an aettan   ps dirrivryl /ious *lXIO . ,pc uoh &= I/O . ,pc gs bus tshed*ate */
    pciRev D.us ibr_IT)
(hanotp;
  _vertt ob_in=>   }

  _WIDGREV	C) &&_	 att)(han_ibr_ +   }

  _un_ (pcASE + reqc     -=1)=<
df att   }

  _un_ (pLIMIT))df attxifoaddr =lhan_ibr_ +   }

  _un_ (pcASE;
	*beakddrrrrtp;
   t anheck * ioe"Dirrivebr_"bbytece iptwhem  chebeid of pc*rydianch>bsswin gfiit doesRno_or_code & codedge)xifoaddr !=lXIO= ~BHERE vices;
    voooooooooooooooobst PO cnonzerorogrg &=bytectbeam ibr_s;
    vooooooooooooooo*bfpnIO c     ofarri rd ofaepioce ipto gs slksibr_s;
    vooooooooooooooooswb;IO con con{BRIDGE;ich formuwhe & es;
    voooooooooooooooobfo;	ac}cythingoce ipto _et twhe & es;
    voooooooooooooooobfn;IO cdpsthe _ce ipto _et twhe & 
	*fp==r(ewt,pcpcci &= (p PAe UIO)df attttt? Tot];

    s = pcp      _  )df attttt: Tot];

    s = pcp      _mem)ft);	*ff = *bfpn);	*s 

 t, pcr&i &= (pcibr_soft->t);	*fcpci*s 
?i &= (pcibr_soft->t:  &= (pcibr_soft->EVr_IT)
*fcpc=obfo vi		ac}wizal* Horfm  ch.e & _unlodd32 eate dge)*ff != 0) i		ac}wizhavela ned, ort.e & 		p    printkoun_=init\	     vice: mt dev,abr_dhan_totxso:_ce i ned, ort inn  ,pc %dn,eoas%s%s, if (%s%svicek)
{
  _t,pcsrk)
{
  *ff &i &= (pcibr_soft-> ? " cibr_soft->" :
"ce  	{
  *ff &i &= (pcibr_soft-> ? " cibr_soft->" :
"ce  	{
  *fn &i &= (pcibr_soft-> ? " cibr_soft->" :
"ce  	{
  *fn &i &= (pcibr_soft-> ? " cibr_soft->" :
"c)of	bricks	    xifoaddr =lXIO= ~BHEREbr32 eate {we wanOKing  mke * si  rnk;.u & _unlo mask;4;
    bridgereg_toivr, nivrGE _unloswb==r(wt,pcpcci &= (p PAe UIO) ?   if (fLASH_ (p WAP :
  if (fLASH_MEM	SWAPGE	rrrroivr slot].reg;
  s_n{BRIDGE_E	rrrrnivr slos 
?ioivr |oswb=:ioivr & ~swb;E _unlodge)oivr != nivr)e wan mke * si  rnk;odgeurnebr_		 mask;t until {BRIDGE;= nivrGE _unlo**fp==rbfn;I		ac}rni rd his as
   ming ibr 		p    printkoun_=init\	     vice: mt dev,abr_dhan_totxso:_ce i       ,pc %dn .eg ho%s%svicek)
{
  _t,pcsrk)
{
  *fn &i &= (pcibr_soft-> ? " cibr_soft->" :
"ce  	{
  *fn &i &= (pcibr_soft-> ? " cibr_soft->" :
"c)of	bricks	 BRIDGE_DE
 ne:      pcs_vunlf (mask == BRI, sfbr       }
  xifoaddrof}amunap(devfs6 ibrask ==p    pdm_t dev,p    pdciio_sar *buf, uint bp  pcbr_sosk)
aph   line_t, de d ob_t, ,k)
aphnnebridt,pcotc_t,pcsk)
aphdpcibr_dmahan_ibr_,k)
aph       reqc    sen/aph       reqc    _maxsen/aphs;
    vot, pc_t       slotp_info 	     slotp_inf = handle_infcern) { pcbr_sofbr     pciio_info pcpciio_info_get(pconn_vh&
    xr_in->f_        pciio_slot rrpciio_info_get(pset(pciio_info);
    pcibr_soft_t       io_ihandle_soft = pcibr_soft_get(pcibr_vh(handle_soft pciio_info);
m	aci_ibr_soft_t       oooo   pcibr_vhdl = hwgraph_cox{ pcbr_sor= bci  xraph_vert{ pcGE io_ihandlep    pdmwgraph_co*ar b        handlep    pdmwgraph_co   plus  r    handlep    pdmwgraph_co handlep    p;ef(pc pcibr_dmamap_pciaddr_gexifoaddrofr_gext    pio  pdmwgraph_co xt    pio  p      s;
    volo waaaaaaaaaaast);
   ac}Mmke ess i* ioehis rnqh    s definon-zeror & codedge)(reqc     <=1)=|| (reqc    _max <=1))_	 */
intNULLof
ciadth      *bC
	de	o _t, dlBRIs   p/_t,pc/addr      *bfnuohxifoaddr isznlomo es
t es id of pce valoled, s ann tandlep  _t, deviceor_code & code(ifoaddr =lhanev,abr_dhan_totxsohp  pcbr_sos _get(pset(,r_t,pcsrhan_ibr_, reqc    sot, pc_GE_DEV_ODE_(ifoaddr ==lXIO= ~BHERE _	 */
intNULLof
ciadthanheck * e pio  p lus 
ogrg epofthisrr valal* Horfpciciio_a eled featupio  p *    pnot Gath
ulupc.cIfesoher,
. ioeesso Oll pwiseid of pc iio_a b actuwupio  p *    pann add ch for* e pio  p lus r_code & codear b  _vh&(
    xr_in->f_pio  pft);
   s_vhdl);
 lf (mask == BRIDGE
       (handlep    p = *ar b    	 handlep    pr!= NULL  	 handlep    pr= handlep    p_vep
nexh vicedge)handlep    p_vep
  psz

= 0) _unlo meakddrrrrtp;
   dge)handlep    p) _ar b  _vhNULLof    ->bss_pt, s);
unlf (mask == BRI, sfbr	NEW(  pcibpio  pft)rrrrtp;
   handlep    p_vep
dn 
  p  pcbr_so r    handlep    p_vep
set(pciio_inf   p      handlep    p_vep
t, pcr= t, pc r    handlep    p_vep
st,pc );sp,pc;s    handlep    p_vep
hanaddr =lhan_ibr_;s    handlep    p_vep
  psz

 reqc     r    handlep    p_vep
saph_vhdl);
    b;t    handlep    p_vep
toc[0] = ATOMICGINIT(0)ev;
   dge)ar b  r_sof(_vhdl);
 lf (mask == BRIDGE	  plus  = *ar b    	handlep    p_vep
nexhr= m plus  r	*ar b  _vhhandlep    p;ef(pc}      pcs_vunlf (mask == BRI, sfbrp;
   dge)handlep    p)_sofxt    pio  p =_	 attxt    pio  pdciio_sx  pcbr_sos 0sen/closcode(ifoaddrsen/closcodereqc    soreqc    _maxsen/closcodet, pcr&i IOMAP_FLAGS)  idge)xt    pio  p vice atthandlep    p_vep
xt    addr =lxifoaddrofe atthandlep    p_vep
xt    p  pcixt    pio  p  	tp->bss_pt    handlep    p_vep
  psz

 0ofe atthandlep    pr= 0nrt}ff(pc}      */
inthandlep    p;e}amunap(devfse & _t, _t dev,p    pdio_slT(pcibpio  pdmwhandlep    p) {fr_gext    pio  pdio_slT(pcibpio  p_vep
xt    p  );t    handlep    p_vep
xt    p  pci0;s    handlep    p_vep
  psz

 0;e}amunap(devfse & cibr_dm_t dev,pio  pdcice_T(pcibpio  pdmwhandlep    psen/apdpcibr_dmahan_ibr_,k)
ap       reqc    ) {fr_ge   }
  xt    pio  pdcice_T(pcibpio  p_vep
xt    p  sen/closcohandlep    p_vep
xt    addr +en/closcohan_ibr_ - handlep    p_vep
hanaddrsen/closcoreqc    );e}amunap(devfse & _t, _t dev,p    pd
 nelT(pcibpio  pdmwhandlep    p) {fr_gext    pio  pd
 nelT(pcibpio  p_vep
xt    p  );t}amunap(devfse & cibr_dm_t dev,pio_t, device_ar *buf, uint bp  pcbr_sosk)
aph_   line_t, de d ob_t, ,k)
aphonnebridt,pcotc_t,pcsk)
aph_dpcibr_dmahan_ibr_,k)
aph_       reqc    sen/aph_s;
    vot, pc_t       sliio_info pcpciio_info_get(pconn_vhpl);
    pcern) { pcbr_sofbrio_ihaniio_slot rrpciio_info_get(pset(pciio_info);
    pcibr_soft_t       io_ihandle_soft = pcibr_soft_get(pcibr_vh(handle_soft pciio_info);
m	aci_ibr_soft_t       oooo   pcibr_vhdl = hwgraph_cox{ pcbr_sor= bci  xraph_vert{ pcGE io_i pcibr_dmamap_pciaddr_gexifoaddrof code(ifoaddr =lhanev,abr_dhan_totxsohp  pcbr_sos _get(pset(,r_t,pcsrhan_ibr_, reqc    sot, pc_GE_DEV_ODE_(ifoaddr ==lXIO= ~BHERE _	 */
intNULLof
ciad   }
  xt    pio_t, device_x  pcbr_sos 0se(ifoaddrs reqc    sot, pcr&i IOMAP_FLAGS)  }amunhinibr_uS ,pc aiio_a t(v
ann manaerming.hinit b cAiio_a b ann Manaer * sioundPIO . ,pc (mempann iop_t,pc)hinit b cT valoled, s gs c   ty_  mplus iclevice
voe sa, 
ndhinit b cdoesRc   ty_trivi foarnaerming of aiio_a t(v
ann io_sk;	..hinit b cT e}cythingocchrmi gs c ess     fragminga t(vo.hinit b cCh>bssweis cchrmi kolupcdbit  ps   ibrmunap(devfse &  pcibr_dm_t dev,piodt,pcociio_sar *buf, uint bp  pcbr_sosk)
aphh_   line_t, de d ob_t, ,k)
aphoonnebridt,pcotc_t,pcsk)
aph_c       reqc    sen/aph_c       al   ming_t       slotp_info = pciio_info_geotp_inf = handle_infcern) { pcbr_sofbr     pciio_info pcpciio_info_get(pconn_vh&
    xr_in->f_        pcs_v_soft = pcibr_soft_get(pcibr_vh(handle_soft pciio_info);
m	aci_ibr_soft_t       r     pciiopiodt,pco = pcibr_piodt      s;
    volo waaaaaaaaaaast);
    pcibr_dmamap_pciaddr_g*hanaddrsg*hanlas  r     pcibr_dmamap_pciaddr_ge pcib_ibr_;s    s                       al   _masknE_    th      *bCheck     n/iper al   mingr_code & code len = al   ming=>= NBPPfbrio_i len =  al   ming=&  al   ming=-=1 ) 
= 0)nE_    al   _mask_vhal   ming=-=1;;
   s_vhdl);
 lf (mask == BRIDGE_DEV_th      *bFirstolookMdgeuRc  viallto aiio_a el c    nexus sor_code & IOECODE_(piodt 
 bci  xr_in->f_piodt,pc)) iceached*aLookMthrough his lus      o right s   n fllocc    .us ibr_novice att ge)piodt->flloc&&_		(piodt->wt,pcpccifp,pc) &&_		(piodt->coung=>= reqc    )=&&_		!(piodt->wo hav& al   _mask      	piodt->flloc= 0nrtt, s);
unlf (mask == BRI, sfbr		 */
intpiodt->wo ha  	rrrrt==ibr_piodt 
 biodt->nexh  	tpwhi.
d(piodt);ef(pc}      len = !piodt);e;
   se_sta (fp,pc) ic
   OR" b &= (p PAe UIO: 	hanaddr =lot];

    s = pcspinfo.pan_   ex_unr	hanlas  =lot];

    s = pcspinfo.pan_   las  r	*beakddrrrrOR" b &= (p PAe UMEM:rf(pcOR" b &= (p PAe UMEM32: 	hanaddr =lot];

    s = pcspinfo.pan_mem ex_unr	hanlas  =lot];

    s = pcspinfo.pan_mem las  r	*beakddrrrrcibault: 	 len = 0)oft, s);
unlf (mask == BRI, sfbr	   }
  0off(pc} 
r_ge pcib_ibr_ = *hanaddrGE_DEV_th      *bAl   e pcib_ibr_or_code & IOECODE_ pcib_ibr_ & al   _mask of(pcib_ibr_ = _ pcib_ibr_ + al   _mask  & ~al   _masknE_    ODE_( pcib_ibr_ + reqc    )=> *hanlas ) iceached*aIf *oodbigramited tos rejriveit.us ibr_   hwgunlf (mask == BRI, sfbr	   }
  0off(pc} dr_g*hanaddr = _ pcib_ibr_ + reqc    );e dr_gNEW( iodt);ef(pcpiodt->flloc= 0nrf(pcpiodt->st,pc );sp,pc;s    hiodt->wo hav=e pcib_ibr_;s    piodt->coung=
 reqc     r    hiodt->nexh 
 bci  xr_in->f_piodt,pc       pcs_vr_in->f_piodt,pc 
 biodtGE io_ihandleunlf (mask == BRI, sfbr       }
   pcib_ibr_;s}amunap(devfse & _t, _t dev,p  dt,pcoio_slar *buf, uint bp  pcbr_sosk)
aphhnnebridt,pcotc_t,pcsk)
aph_dpcibr_dmahanibr_,k)
aph_       reqc    _t       slotp_info = pciio_info_geotp_inf = handle_infcern) { pcbr_sofbr     pcs_v_soft = pcibr_soft_get(pcibr_vh(handle_soft pciios_vr_in->f_m	aci  r     pciiopiodt,pco = pcibr_piodt      s;
    volo waaaaaaaaaaast)f(pcOhar                    , ch[1024]GE_DEV_th      *bLookMthrough his sriver"    cstrucpiolv_    * si pciiopiodt,pco       *bstrucpiol cororspondeen ian 'hanibr_'r_code & IOECs_vhdl);
 lf (mask == BRIDGE
   piodt 
 bci  xr_in->f_piodt,pcGE
   whi.
d(piodt) iceached*aPiodt,pc fllocca le to be_    * sinlock(pched*ac    nann Gathccibs of it..us ibr_IT)
hiodt->wo hav==ahanibr_)vice att ge)piodt->coung=
= reqc    )  	*beakdd3DEV_th s code tImn/iper      past pl    io_sk;	..hs code tPviceaagaessagb ann *beakdd3DEV_u & _unloibr_vhdlvb

  t, chcern) { pcbr_so, , ch, 1024dri	     vice: KERN_WARNING i

    xp  dt,pcoio_s:ue a w"dri	     vice: KERN_WARNING i
Dcvb_w %s io_sk;	      (0x%lx) diffehingo* aniaiio_a el (0x%lx)",dfttt	, ch, reqc    sopiodt->coungdri	     vice: KERN_WARNING i
Fo_sk;	 0x%lx inetead"sopiodt->coungdri	    *beakdd32
	piodt 
 biodt->nexh      tp;
   dge)!piodt) ice vice: KERN_WARNING idft

    xp  dt,pcoio_s:uA   essc0x%lx      0x%lx - Nofm  chvicek)
_t addrs reqc    )oft, s);
unlf (mask == BRI, sfbr	   }
 ;ef(pc}      iodt->flloc= 1; io_ihandleunlf (mask == BRI, sfbr       }
 of}amun =====================================================================hinit bsingMANAGEMENTdevhinit b cT e}    pciASIC
if    *sathrlocmethods
of dostahinit b csin:aviaaag"dirrivear ";erous tofrvailterr inhinit b c32-b
     r. ,pc (on con_elrivsla nedtiguall 2Ghinit b ca   essc_t,pcion  osanoll p  xtalk),aviaoinit b c"dirriv"ar   esstwheviaa64-b
     r. ,pc (a  hdnit b cdes tw      inform      nloesc /ioweis ounhinit b ca   ess,h
uplud whent, dfoudpciributes)e respviaoinit b caRnmant d";erour_i sataaiiowloa bunconefhdnit b cdiffehingosmcii ar beensRe hbeiis terishldiwiewhdnit b chis oMU.devhinit b cF   efficiencye[we most
prefer kolupcdhis 32-b
 hdnit b cdirrivear been faanlity===
up oitlore thesRnohdnit b ccinguror aiio_a t(vsao fioadvantagb of ed when Didnit b coMUnect dhis 64-b
  dirriveis e eviseenrr-cyclDidnit b co  rr   esstwheca lbo upcd;ahis a vantagb ofidnit b ced whe64-b
  dirriveect doMUnr   esstwhei eleatidnit b cwe * TGathhavelograiio_a b *   iesr ca*   oMU.dev/amunhiniConvb

co  -g  */
  raphwan et, pcrresp    pc-.   if ( raphwan et, pc
 *bfnuoh    pc-.   if ( DirriveMapdpciribute bcte.dev/aipcticc pcibr_dm_t dev,t, pc_totd64(s;
    vot, pc,  pcs_v_soft =ask == BRIDt       pcibr_dmamap_pciaddr_gepciributes

 0nref(pcwanSanity Dheck:}    pcie to aiiowlou_urof VCHAN1eviaa64-b
  ibr_se & 		pcibr_slot_soft len =_ALWAYS)!(t, pcr&i &=BR_VCHAN1)=|| (t, pcr&i &= (p=MAeA64)driver_reg_soft/e Gl */
  macroet, pc
 EV_u & 
   dge)t, pcr&i &= (p=MAeDATA) i	/pcto ndard"    c  rnneE; & 	pciributes
;

  &=64_ATTR_BAR PO cno  ar ie  bct; & 	pciributes
|= oun64_ATTR_GE_D PO cprefetchion ibrf(pc}d
   dge)t, pcr&i &= (p=MAeCMD) i	/pcto ndard"nlomresp  rnneE; & 	pciributes
|= oun64_ATTR_BAR PO c ar ie  bct;on ibr	pciributes
;

  &=64_ATTR_GE_D PO cd off.
 prefetchiibrf(pc}d
   /e Gl */
  e_iiilet, pc
 EV_u & 
   dge)t, pcr&i &= (pGE_DEV_P) 	pciributes
|= oun64_ATTR_GE_D  
   dge)t, pcr&i &= (p ~GE_DEV_P)r	pciributes
;

  &=64_ATTR_GE_D g_soft/e eis ce i bit val ca*   a   esscpciributes
    xbbase;
 & IOECODE_ slotp;
  _vertx*bridg)       	dge)t, pcr&i &= (pcibr_soft->)diaddr_ge	pciributes
|= oun64_ATTR_SWAPGE    	dge)t, pcr&i &= (pcibr_soft->)diaddr_ge	pciributes
;

  &=64_ATTR_SWAPGE    tp;
   t aPf    *r-.   if ( t, pc
 EV_u & 
   dge)t, pcr&i &=BR_BARRIER) 	pciributes
|= oun64_ATTR_BAR  
   dge)t, pcr&i &=BR_NOBARRIER) 	pciributes
;

  &=64_ATTR_BAR 
 
   dge)t, pcr&i &=BR_GE_DEV_P) 	pciributes
|= oun64_ATTR_GE_D  
   dge)t, pcr&i &=_p NOGE_DEV_P)r	pciributes
;

  &=64_ATTR_GE_D g_softdge)t, pcr&i &=BR_GE_CISE) 	pciributes
|= oun64_ATTR_GE_C  
   dge)t, pcr&i &=_p NOGE_CISE) 	pciributes
;

  &=64_ATTR_GE_C g_softdge)t, pcr&i &=BR_VCHAN1) 	pciributes
|= oun64_ATTR_VIRTUAL  
   dge)t, pcr&i &=_p VCHAN0) 	pciributes
;

  &=64_ATTR_VIRTUAL  r       }
  (pciributes);s}amunap(devfse &  slotpdma  pdm_t dev,dma  pdciio_sar *buf, uint bp  pcbr_sosk)
aph   line_t, de d ob_t, ,k)
aph       reqc    _maxsen/aphs;
    vot, pc_t       sliio_info pcpciio_info_get(pconn_vhpl);
    pcern) { pcbr_sofbrio_ihandle_soft = pcibr_soft_get(pcibr_vh(handle_soft pciio_info);
m	aci_ibr_soft_t       oooo   pcibr_vhdl = hwgraph_cox{ pcbr_sor= bci  xraph_vert{ pcGEinfo_get(p_slot rrpciio_info   p      x xtalk_in  = hwgraph_coxiiopori  r    xt    dma  pdmhwgraph_coxt    dma  p  io_ihandleema  pdmhwgraph_cohandleema  p  
   derrrrrrrrrrrrrrrrrrrrrrS(pccoung  
   derrrrrrrrrrrrrrrrrrrrrrS(pcindex g_soft/e merk;odn ss mrvot, pce & IOEC , pcr|= bci  xraph_vertima_t, pc  ice =th      *bOnaSNIA64,hts ss   ps an epre-aiio_a el becau_urt dev,dma  pdciio_s)      *bca lbo em starwiew caanrir.
	     th* Hoto get ran    handledma  p_vh(handleema  pdm));
 fllo__get(pema  p_ slotp;
  _vertr_sofbr_i _idge)!pandleema  p)r	   }
  0ofr    xt    dma  ppcixt    dma  pdciio_sx  pcbr_sos d ob_t, ,hreqc    _maxsen/ccloscodt, pcr&i=MAMAP_FLAGS)  i _idge)!xt    dma  p vic		p oun_p ATE_=init\	
#endif

    xpciio_stxt    dma  pdciio_ fa*/ vet_driver_reg	fllo__get(pema  p_ slotpdma  p br	   }
  0off(pc} dr_gxiioporir= bci  xraph_vertmxid      set(pciio_info);
    pcibr_soft_t       n    handledma  p_ved
dn 
  p  pcbr_so r    handledma  p_ved
set(pci   p      handledma  p_ved
saph_vhdl);
    b;t    handledma  p_ved
xterne=oxt    dma  p  io_ihandleema  p_ved
max_     
 reqc    _max;t    handledma  p_ved
xiioporir= xiiopori  r    dge)t, pcr&i &= (p=MAeA64)vicedge)!pandletryv_etdi->bcei {pcib======slot,ot, pc,   }

  DEV	D4_ucITS))vice att pcibr_dmamap_pciaddr_gepan_ibr_;s	
   derrrrrrrrrrrrrrrrrrrrrrhaveciios;s	
   derrrrrrrrrrrrrrrrrrrrrrminciios;s_	 attac}D->b_wi   capterr
of A64toper     s,us code tresp hecpciributes
ofelegbsingars ed    *bconsito nttoes turnec  viallbsin ed    *bar beensRed wheshahe _cingurord.us code & fe atthanoaddr =lhanev,t, pc_totd64(t, pc,  pcs_v_sof);
fe atthandledma  p_ved
t, pcr= t, pc re atthandledma  p_ved
(ifoaddr =l0 re atthandledma  p_ved
hanoaddr =lhanoaddrof 	 attac}Mmke ess isozhavelan RRB_(hartwo).us code & _unlodge)!(hanotp;
  _verttrb_fixe _& (1 << slotf      	dge)t, pcr&i &=BR_VCHAN1) 	
    slot +=eist_p RRBpSLOTNVIRTUALbr		haveciiosr= bci  xraph_vertt);
_slotlock(p;  	dge)haveciiosr< 2) _pttIOECODE_ sl_ibr_ &  &=64_ATTR_GE_D) 	
	minciiosr= 2;pttIOECe>bsdf		minciiosr= 1;pttIOECODE_haveciiosr< minciios)_		te_t *, pciio_autociio_s {pcib======slot,ominciiosr-rhaveciios _dft}  f(pc} 		p oun_p ATE_=init\	 atth#endif

    x ma  pdciio_:Red whedirriv64et_driver_reg	 att */
inthandledma  p  	} 		p oun_p ATE_=init\	h#endif

    x ma  pdciio_:Renterr
kolupcddirriv64et_driver_reg	t, pcr&

  &= (p=MAeA64GErrrrtprrrr= net, pcr&i &= (pFIXED viceacbwar sta:bar beensRperffa*/ lBRIr,hed*a= ndirriv32 DMA'tcgeh for* e a   ess.us ibr_IT)
!pandletryv_etdi->bcei {pcib======slot,ot, pc,   }

  DEV	D32ucITS))vice attwanUs p n-sthesRDIE_CT A32 oper     s,us code tresp hecpciributes
ofelegbsingars ed    *bconsito nttoes turnec  viallbsin ed    *bar beensRed wheshahe _cingurord.us code eMapptwhecallsRperffa*/ oftharalkus code tdsReutsi	de*   dirriv32 r>bss.us code & 		p oun_p ATE_=init\	 atth#endif

    x ma  pdciio_:Red whedirriv32et_driver_reg	 atthandledma  p_ved
t, pcr= t, pc re atthandledma  p_ved
(ifoaddr =lbci  xraph_vertiir_xex_unr	rrrrhandledma  p_ved
hanoaddr =loun32uDIE_CTpcASE;
	 att */
inthandledma  p  	} 		p oun_p ATE_=init\	h#endif

    x ma  pdciio_:Renterr
kolupcddirriv32et_driver_reg	/pci*/

  us p n-mress FIXED respwioDMA'thed*ag.velch forhim,ffa*/oi	her_	xt    dma  pdio_slxt    dma  p ;g	fllo__get(pema  p_ slotpdma  p br	   }
  0off(pc} dr_gth      *bAlio_a b A   esscTt, dlBRur_iE   iesr /ioweis ar been RAM needed aUnlessceis oun_p NO ATE_ROUNDUPc crp gs s   if edeid of pc*is arximumvtumberiof ATEs gs ex_uder_i sizworst-OR"       *bscenario, in rstuhrlited toldinaralksval ca*        *blas  byte of an ATE;ahius, ar been IOPGSIZE+2      *bdoesRer_ adaore th when rlocATEsor_code & IOECODE_!(t, pcr&i &=BR_NO ATE_ROUNDUP))viceS(pccoung =lIOPG((IOPGSIZE=-=1 eacbworstcOR" bwo havrotg &=br_		_code+reqc    _max wan mx ar been bytec=br_		_code-=1)=+ 1;		ac}rers. UPcibrf(pc} eate {w/pc ssumrlited toldinaralksvalpagb al    voibr	pcpccoung =lIOPG(reqc    _max ft/e mmx ar been bytec=br_		_code-=1)=+ 1;		ac}rers. UPcibrf(pc}E_    a(pcindex =lhanev,ate_ciio_s {pcib======pcpccoungfbr_i _idge)a(pcindex !);-1)vicedge)!pandletryv_etdi->bcei {pcib======slot,ot, pc,   }

  DEV	PMUucITS))vice att*bridgea(pcrrrrrrrrrrrrra(pcproto;s	
   derrrrrrrrrrrrrrrrrrrrrrhaveciios;s	
   derrrrrrrrrrrrrrrrrrrrrrminciios;s_		p oun_p ATE_=init\	 atth#endif

    x ma  pdciio_:Red whePMUet_driver_reg_	rrrra(pcproto =lhanev,t, pc_tota(p(t, pc_GE_	 atthandledma  p_ved
t, pcr= t, pc re atthandledma  p_ved
hanoaddr =_		oun32uMAPPEDpcASE + IOPGSIZE=*rS(pcindex g3DEV_th s code t    xbbase;
his syte-ce i bit c=obit 29 of o  rr   essus code & _unlodge) slotp;
  _vertx*bridg)   ttIOECODE_t, pcr&i &= (pcibr_soft->)df	tIOECATE_SWAP_ON(handledma  p_ved
hanoaddr);pttIOECth ss code tIf ce i was slks caridge->bcepinthandleer_ranv_et() ss code twe nolding ch>bssweis a   esscbit.uss code & __unlodge) slotp;
  _vertibr_lock(p.ridge->bcep& dfttt			  }

  DEV	SWAP_PMU)df	tIOECATE_SWAP_ON(handledma  p_ved
hanoaddr);pttIOECdge)t, pcr&i &= (pcibr_soft->)df	tIOECATE_SWAP_OFF(handledma  p_ved
hanoaddr);ptf(pc}Ee atthandledma  p_ved
(ifoaddr =l0 re atthandledma  p_ved
a(pcp  _vhhandleate_cice_T(pcib======pcpcindex) re atthandledma  p_ved
a(pcindex =lS(pcindex g3DEV_handledma  p_ved
a(pccoung =lS(pccoung  e atthandledma  p_ved
a(pcproto =la(pcproto;s 	 attac}Mmke ess isozhavelan RRB_(hartwo).us code & _unlodge)!(hanotp;
  _verttrb_fixe _& (1 << slotf      	haveciiosr= bci  xraph_vertt);
_slotlock(p;  	dge)haveciiosr< 2) _pttIOECODE_a(pcproto &CATE_GE_D) 	
	minciiosr= 2;pttIOECe>bsdf		minciiosr= 1;pttIOECODE_haveciiosr< minciios)_		te_t *, pciio_autociio_s {pcib======slot,ominciiosr-rhaveciios _dft}  f(pc} tIOECODE_a(pcindex >= bci  xraph_vertvoi_S(pc     && dfttt! slotp;
  _vertx*bridg)   tt*bridge = pcibr_soft->bs_base;

  pcs_v_sof_vertex_unr		volBRurr
s;
    voooooo*cmd_regp  	es;
    voooooooooooooooocmd_reg  	es;
    volo waaaaaaaaaaast);tt, s);
dma  p_ved
t, pcr|=eist_p =MAMAP_SSR->t);	f(_vhdl);
 lf (mask == BRIDGE		cmd_regp_vh&( mask;t en/closcob_type0_cfgge->lock(p.en/closcol[oun_CFGpCOMMAND / 4]DGE		cmd_reg = *cmd_regp  	e slotp;
  _vertibr_lock(p.ridgcmd_cibr_e  = cmd_regp  	e slotp;
  _vertibr_lock(p.ridgcmd_shaibr =ocmd_reg  	e, s);
unlf (mask == BRI, sfbr	f(pc} tIOEC */
inthandledma  p  	} 		p oun_p ATE_=init\	h#endif

    x ma  pdciio_:Renterr
kolupcdPMUet_driver_reg	handleate_io_slT(pcib======pcpcindex==pcpccoungfbrf(pc}  attac}tot  ffa*/ss : sorry,eyouzjustcORA'th code tget  /iots rstuohts rs * ioewayor_code & 		p oun_p ATE_=init\ atth#endif

    x ma  pdciio_:Rnlock(pcifa*/ss .et_driver_regunloxt    dma  pdio_slxt    dma  p ;gunlofllo__get(pema  p_ slotpdma  p brIOEC */
int0;e}amunap(devfse & _t, _t dev,dma  pdio_slhandleema  pdmhpandleema  p)r       slotp_soft = pcibr_soft_get(pcibr_vhhandledma  p_ved
saphGEinfo_get(p_slot rrpciio_info   p_vhhandledma  p_ved
s  p       s;
    voooooooooooooooot, pcr= , s);
dma  p_ved
t, pct);
   ac}Mmke ess i* ioeridgexh_S(ps_act.veh code tisin/iperry kepg adato datror_code & _i _idge)ist_p =MAMAP_BUSY &Ct, pc_t	dge)ist_p =MAMAP_SSR-> &Ct, pc_t	i _iatomicge-c(&_ slotp;
  _vertibr_lock(p.eridgexh_S(ps_act.ve)ft);
   xt    dma  pdio_slhandledma  p_ved
xternDGE_DEV_ODE_ slotpdma  p_ved
t, pcr&i &= (p=MAeA64)vice *, pcit_t pcdi->bcei {pcib======slot,o  }

  DEV	D4_ucITS)GErrrrtprrrr= nehandledma  p_ved
a(pccoung)vice *, pcate_io_slT(pcibdma  p_ved
saphsen/aph_cothandledma  p_ved
a(pcindexsen/aph_cothandledma  p_ved
a(pccoungdri	 *, pcit_t pcdi->bcei {pcib======slot,o  }

  DEV	PMUucITS)GE    tp;
   fllo__get(pema  p_ slotpdma  p br}amunhinicothandleibr_d(ifot_t *,:ag.ven adoIO r>bss,rbr_vidnit b cback * e cororspondeen bR" b &= MEM a   ess;hdnit b chigs gs  s
  ogrghort-circuit singited to eleatidnit b cloopcback onian* is ound ue.dev/aipcticc pcibr_dm_t dev,ibr_d(ifot_t *,(handle_soft  saphsen/aph_co pcibr_dma(ifoaddrsen/o_info       reqc    _t       pcibr_dmamap_pciaddr_gexifolim =lxifoaddr + reqc     -=1 r     pcibr_dmamap_pciaddr_gepan_ibr_;s    hant(p_slot rrpciio_info   p  prrrr= ne_(ifoaddr >=   }

  _un_MEM32ucASE)=&&_	(xifolim <=   }

  _un_MEM32uLIMIT))vice *, addr =lxifoaddr -=  }

  _un_MEM32ucASE;
	 */
inthan_ibr_;s    }prrrr= ne_(ifoaddr >=   }

  _un_MEM4_ucASE)=&&_	(xifolim <=   }

  _un_MEM4_uLIMIT))vice *, addr =lxifoaddr -=  }

  _un_MEM4_ucASE;
	 */
inthan_ibr_;s    }prrrr    (get(pci0noset(p< 8; ++slotfr_IT)
((ifoaddr >=   }

  DEVIO(slotf  &&_	 att)xifolim <   }

  DEVIO(slot=+ 1)))vice att*bridg4;
    bridgereg_tnevGE_	 attnevpci 
  _vertibr_lock(p.ridge->bce  e atthan addr =lnevp&   }

  DEV_OFF=MASKt)e atthan addr <<=   }

  DEV_OFF=ADDR_SHFTt)e atthan addr +=lxifoaddr -=  }

  DEVIO(slotf;
	 att */
int(nevp&   }

  DEV_DEV	 (pMEM) ? han addr :  &== ~BHEREbr32rIOEC */
int0;e}amunap(devfse &  pcibr_dm_t dev, ma  pdcice_T(pcibema  pdmhpandleema  psen/o_cibr_dmahibr_,k)
ap       reqc    ) {fr_ge slotp_soft = pcibr_soft_get(pcibr;ef(pc pcibr_dmamap_pciaddr_gexifoaddrofr_gex xtalk_in  = hwgraph_coxiiopori       pcibr_dmamap_pciaddr_gepan_ibr_;s    s;
    voooooooooooooooot, pc  prrrr len = pandleema  pr!= NULLfbrio_i len = reqc     >e mmaf(pc len = reqc     <= , s);
dma  p_ved
max_       n    handlecibr_vhhandledma  p_ved
saphGE IOEC , pcr= , s);
dma  p_ved
t, pct);
   (ifoaddr =lxt    dma  pdcice_T(pcibema  p_ved
xtern,ahibr_, reqc    )ofrrrr= neXIO=PACKED((ifoaddr))vicexiioporir= XIO=PO = (ifoaddr);_	xifoaddr =lXIO=ADDR (ifoaddr);_    }Ce>bsdfxiioporir= bci  xdma  p_ved
xiioporit);
   ac}If * gs DingfsRe hania   esscleatidddddc}rnfous back *an* is     pcichip,idddddc}rnducetit back *an* e cororctidddddc} &= MEM a   essor_code & codedge)xifoporir=
  pcs_v_sof_vertxid)vice *, addr =lt dev,ibr_d(ifot_t *,(handle_sofse(ifoaddrs reqc    );_    }Ce>bs dge)t, pcr&i &= (p=MAeA64)viceac}A64tsin:hed*aaliayszupcd64-b
 cdirrivear been,hed*aon conaliayszworksoi	he Dcvb_w(x)twas slksadadureenhed*adma  praiio_a t(voi	her_ceac}pciributes
an eal* Horfbu_vhd_ adafnuohed
hanoaddr.us ibr_   _ibr_ = bci  xdma  p_ved
   _ibr_
	 att|)
(ufnu4_ug)vxifoporir<<  &=64_ATTR_Tap(_SHFT)
	 att|)xifoaddrof 	wanDriver Hardwan eWAR #482836:hed*aIf * e nt, dfoudisRno_cORchb al    vhed*aaesp hec    pciRev isR<=  , ss mrhed*aprefetchie hbeioff.us ibr_IT)
t, pcr&i &=_p NOGE_DEV_P)r	 atthan addr ;

  &=64_ATTR_GE_D g_		p    printkoun_p =MA_=init\	h#endif

    x ma  pdcddr (dirriv64):et_dft
\tif (d_ hibr_n[0x%x..0x%x]et_dft
\tXIO porir0x%xvrotg &=0x%xvicdft
\t */
in wheP&= 0x%xvice 	e,addrs paddr + reqc     -=1e 	exifoporise(ifoaddrs hanoaddr);pver_regunlo}Ce>bs dge)t, pcr&i &= (pFIXED viceacbA32 dirrivesin:hed*aaliayszupcd32-b
  dirrivear been,hed*aon conperffa*/oi	he Dcvb_w(x)twas slksadadureenhed*adma  praiio_a t(voi	her_cedge)xifoporir!=lbci  xraph_vertiir_xpori)r	 atthan addr ci0n		ac}wro waDIDN ibr_e>bs dge)xifoaddr <thandledma  p_ved
(ifoaddr)r	 atthan addr ci0n		ac}iuttofar>bsswibr_e>bs dge)(xifoaddr + reqc    )  en/nehandledma  p_ved
xifoaddr +   }

  DMA_=IE_CTpSIZE))r	 atthan addr ci0n		ac}iuttofar>bsswibr_e>bsfe atthanoaddr =lhanev,dma  p_ved
hanoaddr + 	exifoaddr -=handledma  p_ved
xifoaddr g_		p    printkoun_p =MA_=init\	h#endif

    x ma  pdcddr (dirriv32):et_dft
\tif (d_ hibr_n[0x%x..0x%x]et_dft
\tXIO porir0x%xvrotg &=0x%xvicdft
\t */
in wheP&= 0x%xvice 	e,addrs paddr + reqc     -=1e 	exifoporise(ifoaddrs hanoaddr);pver_regunlo}Ce>bs {   mask;e = pcibr_soft->bs_base;

  pcs_v_sof_vertex_unr	 pcibr_dmamap_pciaddr_gerotg &==eIOPGOFF((ifoaddr);_	*bridgea(pcrrrrrrrrrrrrra(pcproto =lhanev,dma  p_ved
a(pcproto;s	derrrrrrrrrrrrrrrrrrrrrrS(pccoung =lIOPG(rotg &=+ reqc     -=1)=+ 1;
s	derrrrrrrrrrrrrrrrrrrrrrS(pcindex =lhanev,dma  p_ved
a(pcindex; es;
    voooooooooooooooocmd_regs[8]; es;
    voooooooooooooooos;s_		p oun_p FREEZE_TIMEs	derrrrrrrrrrrrrrrrrrrrrrS(pctot  f=lS(pccoung  es;
    vooooooooooooooootree    imeriver_reg_		p oun_p ATE_=init\	*bridgea(pcrrrrrrrrrrrrra(pccmp;_	*bridgea(pcprrrrrrrrrrrra(pccb    	s;
    vooooooooooooooooa(pclo==pcpchi;s	derrrrrrrrrrrrrrrrrrrrrrS(pcbad =l0 rederrrrrrrrrrrrrrrrrrrrrrS(pcrbc =l0 rver_reg	*bridgea(pcprrrrrrrrrrrra(pcp  _vhhandledma  p_ved
a(pcp  ;_	*bridgea(pcrrrrrrrrrrrrra(pof 	wanDriver Hardwan eWAR #482836:hed*aIf * e nt, dfoudisRno_cORchb al    vhed*aaesp hec    pciRev isR<=  , ss mrhed*aprefetchie hbeioff.us ibr_IT)
t, pcr&i &=_p NOGE_DEV_P)r	 atta(pcproto &

 ATE_GE_Dof 	a b =la(pcproto
	 att|)
xifoporir<< ATE_TIDSHIFT)
	 att|)(xifoaddr -=rotg &fGE_	han_ibr_ =lhanev,dma  p_ved
hanoaddr +orotg &; f	/*bFillpintoppear been erous to 
ed*awiewweis apn/ipria b xterne    ,hed*aaesphann *ack * e o  rr   essoi	her_ce len = a(pccoung >e mma	ODE_a(pccoung <= , s);
dma  p_ved
a(pccoung)vice	ATE_FREEZE(DGE		ATE_WRITE(DGE		ATE_THAW(DGE		/* wait until Bridge PO complete */
    pcibr_unlock(pcibr_	tp->bss_ptPO c fiotumberiof ATE'slore thed isRg* HtI wr aninhcitumberen/n*iaiio_a el     * isRpep. Oneeway * isRca lhant   gs reg	ed*ap    x ma  pdciio_()twas em starwiewceis oun_p NO ATE_ROUNDUPg	ed*at, p,aaesp hen in nettan   p gs  s
  (right now),se fheed*a*aralksa   essclellsRusisozorosticdid nolding rers.up.heed*a fiooll p cissienlityeis e evieis ar eis justcpla ca*ooheed*asmcii forhr_vhdce rlited toldinaralksan a.heed*& 		p oun_p ATE_=init\	e vice: KERN_WARNING 

    x ma  pdcddr :et_dftt
\tif (d_ hibr_n[0x%x..0x%x]et_dftt
\ta(pccoung 0x%xved
a(pccoung 0x%xvicdftt
\tATE'slore thed >otumberiaiio_a elvice 	ee,addrs paddr + reqc     -=1e 	e	pcpccoung, , s);
dma  p_ved
a(pccoung) rver_reg		han addr ci0nr32rs    }prrrr */
inthan_ibr_;s}amunap(devfse & alenlus dm_t dev, ma  pdlus _T(pcibema  pdmhpandleema  psen/o_alenlus dm palenlus sen/o_s;
    vot, pc_t       slotp_soft = pcibr_soft_get(pcibr;ef(pc mask;e = pcibr_soft->bs_base;=NULLof
ciads;
    vooooooooooooooooal
t, pcr= )t, pcr&i &= (p ~SLEEP) ? ALp ~SLEEP : 0  
   derrrrrrrrrrrrrrrrrrrrrrinplapc 
 t, pcr&i &= (pINPLACEnE_    alenlus dm              hant(palenlus c= 0nrf(pcalenlus dm              xt    alenlus ;s    s                       length       pcibr_dmamap_pciaddr_gerotg &; ciads;
    voooooooooooooooodirriv64  
   derrrrrrrrrrrrrrrrrrrrrrS(pcindexc= 0nrf(pcderrrrrrrrrrrrrrrrrrrrrrS(pccoung =l0nrf(pcderrrrrrrrrrrrrrrrrrrrrrS(pctot  f=l0;ef(pc mask;ea(pcprrrrrrrrrrrra(pcp  _vh( mask;ea(pcp)0;ef(pc mask;ea(pcrrrrrrrrrrrrra(pcproto =l( mask;ea(pct)0;ef(pc mask;ea(pcrrrrrrrrrrrrra(pcprevGEf(pc mask;ea(pcrrrrrrrrrrrrra(pnrf(pcalenibr_dmamap_pciaddr_gxifoaddrofr_gex xtalk_in  = hwgraph_coxiiopori       pcibr_dmamap_pciaddr_gepan_ibr_;s    alenibr_dmamap_pciaddr_gnew_ibr_;s    s;
    voooooooooooooooocmd_regs[8];     s;
    voooooooooooooooos

 0nre		p oun_p FREEZE_TIMEs    s;
    vooooooooooooooootree    imeriver_reg     ntn/closcate_io_s   
 ne ci0n	O c o pair ATE_THAWdfttt			d*awiewwan ATE FREEZEdfttt			d*/ n    handlecibr_vhhandledma  p_ved
saphGE IOECxt    alenlus  =lxt    dma  pdlus _T(pcibema  p_ved
xtern,ahilenlus sen//closcodet, pcr&i=MAMAP_FLAGS)  i _idge)!xt    ilenlus )s	gowedfa*/nE_    alenlus dcursg__snit(xt    ilenlus s 0seNULLfbr i _idge)inplapc)vice *,t(palenlus c= xt    alenlus ;s    } ->bss_pt, st(palenlus c= alenlus dc* HtI(al
t, pcmma	ODE_!, st(palenlus )
	 attgowedfa*/nE    }r i _idirriv64r= , s);
dma  p_ved
t, pcr&i &= (p=MAeA64  i _idge)!dirriv64) {   mask;

  pcs_v_sof_vertex_unr	a(pcp  _vhhandledma  p_ved
a(pcp  ;_	S(pcindex =lhanev,dma  p_ved
a(pcindex; ea(pcproto =lhanev,dma  p_ved
a(pcproto;s	ATE_FREEZE(DGE	ate_io_s   
 ne ci1n	O cRememberi* ioewe nolding dowan ATE THAWiibrf(pc}d
   han_ibr_ =lhanev,dma  p_ved
hanoaddr;E_    a(pcprev ci0n		 wan m chebcno _slot ATEs ibrf(pcwhi.
d(ALENLIST_SUCCESS ==_	 atalenlus dibr_xt    ilenlus s NULLs 0sen/c&(ifoaddrs &length,oal
t, pc))vicedge)XIO=PACKED((ifoaddr))viceh_coxiioporir= XIO=PO = (ifoaddr);_	h_coxiioaddr =lXIO=ADDR (ifoaddr);_	}Ce>bsdfh_coxiioporir= bci  xdma  p_ved
xiioporit);	dge)xifoporir=
  pcs_v_sof_vertxid)vicedr_gnew_ibr_ =lt dev,ibr_d(ifot_t *,(handle_sofse(ifoaddrs length);_	h_codge)new_ibr_ ==  &== ~BHERE)_		gowedfa*/nE32 eate dge)dirriv64) {  dr_gnew_ibr_ =lt d_ibr_ |)xifoaddr_		|)
(ufnu4_ug)vxifoporir<<  &=64_ATTR_Tap(_SHFT);s 	 attac}Driver Hardwan eWAR #482836:heddddd*aIf * e nt, dfoudisRno_cORchb al    vheddddd*aaesp hec    pciRev isR<=  , ss mrheddddd*aprefetchie hbeioff.us dddd*& _unlodge)t, pcr&i &=_p NOGE_DEV_P)r		new_ibr_ ;

  &=64_ATTR_GE_D g_	tp->bss_pt    ac}corculBRIse f a(p _slue_   heddddd*ae f firstor   essoaIf ikus code t m chebc* si   viallus code tATE writt ne(ie.isozhavheddddd*amultick( blf (al ca*   eddddd*asamelIOPG),se fn *ack upheddddd*aaespreer,
. ioeATE.us dddd*heddddd*aWc ars NOTtgoeen iaheddddd*aagg esstvelyc*rydiaheddddd*areer,
urneoll p ATEsors dddd*& _unlorotg &==eIOPGOFF((ifoaddr);_	    a(p =la(pcproto
		|)
xifoporir<< ATE_TIDSHIFT)
		|)(xifoaddr -=rotg &fGEtIOECODE_a(p == a(pcprev vic		p oun_p ATE_=init\		h#endif

    x ma  pdlus :tATE shaheet_driver_reg		a(pcp  --;g		a(pcindex--;g		t d_ibr_ -= IOPGSIZEbr	f(pc} tIOECnew_ibr_ =lt d_ibr_ +orotg &; f	    ac}Fillpint* sihardwan eATEsheddddd*ae ioenedta ca* isRblf (ors dddd*& _unloS(pccoung =lIOPG(rotg &=+ length -=1)=+ 1;
_unloS(pctot  f+=lS(pccoung  f	    ac}Eness i* ioehiisRpepenedta csRerough ATE'sl*& _unlodge)S(pctot  f> , s);
dma  p_ved
a(pccoung)vic		p oun_p ATE_=init\	e vice: KERN_WARNING 

    x ma  pdlus c:et_dftt
\tif (d_ xifoaddr [0x%x..0x%x]et_dftt
\ta(pctot  f0x%xved
a(pccoung 0x%xvicdftt
\tATE'slore thed >otumberiaiio_a elvice 	ee(ifoaddrs xifoaddr + length -=1e 	e	pcpctot  , , s);
dma  p_ved
a(pccoung) rver_reg		gowedfa*/nE3    }r tIOECATE_WRITE(DGE
_unloS(pcindex +=lS(pccoung  e atta(pcp  _+=lS(pccoung  f	    a(pccoung <<=lIOPFNSHIFT;_	    a(p +=lS(pccoung  e atthan addr +=lS(pccoung  e} f	/*bwritr * sioundsinga   essus c}iutt*an* e s_a ter-gall p lus .us ibr_IT)
inplapc)viceunlodge)ALENLIST_SUCCESS !=_		alenlus dreplapc(, st(palenlus s NULLsen//cl&new_ibr_s &length,oal
t, pc))_		gowedfa*/nE32 eate iceunlodge)ALENLIST_SUCCESS !=_		alenlus dant  d(, st(palenlus sen//cnew_ibr_s length,oal
t, pc))_		gowedfa*/nE32s    }prrrr= ne!inplapc)
	alenlus d
 nelxt    ilenlus )t);
   ac}Re.eg hherir.
	n  fcursg_
ofelegbalenlus ce hbei */
inen *ackid of pc*an* e cm st_or_code & IOECalenlus dcursg__snit(, st(palenlus s 0seNULLfbr ;
   ac}IncOR" ban ATE FREEZEtwas 
 ne dan* e ATE THAWikolunrollpcii f        *bch>bsss
. ioeATE FREEZEthas 
 ne kol mplrming * e ex.
	n  fSSR->      *bbugzworkarers.or_code & codedge)ate_io_s   
 ne)viceATE_THAW(DGE	/* wait until Bridge PPO complete */
    pcibr_unlock(pcibr_    }prrrr */
inthant(palenlus t);
 fa*/:;
   ac}Ts rs an evarious cibr_s
ofefa*/ss  aftI wdostawan ATE FREEZEdddddd*aWc nolding dowan ATE THAWo Oll pwisen* e ATEs
an elf (el    evt_or_code a fiodecisur_i g dowan ATE THAWinoldsRe hbeiex_uder_iwhell p ar_code aan ATE FREEZEtwas 
 ne be   eor_code & codedge)ate_io_s   
 ne)viceATE_THAW(DGE	/* wait until Bridge Errrrtprrrr= nehant(palenlus c&& !inplapc)
	alenlus d
ectboyehant(palenlus  brIOEC */
int0;e}amunap(devfse & _t, _t dev,dma  pd
 nelT(pcibema  pdmhpandleema  p)r      th      *bWe couldtgoMthrough ann in_slota b ATEs
s rs;      *b    nerform up o* Hs  s,cwe * n'_or_code aWc also * n'_Rerss mrn* e strort al.
	n  t(v      *bb
t es  _ibr_/dlus cann d
 ne,pnot Hubbdoesor_code & _i _idge), s);
dma  p_ved
t, pcr&i &=_p =MAMAP_BUSY)vice *, pcdma  p_ved
t, pcr&

  &=_p =MAMAP_BUSYt);	dge), s);
dma  p_ved
t, pcr&i &=_p =MAMAP_SSR->_t	i _iatomicge-c(&_ slotpdma  p_ved
saph_vertibr_lhandledma  p_ved
s  pp.eridgexh_S(ps_act.ve)ft)rrrrtprrrrxt    dma  pd
 nelT(pcibema  p_ved
xternDGE}ammunhiniF   eRchc mask;,elegbsIR_OFF _slue_int* siDirriveMapbeen Rrous tohinid(pcrminesceis ounRe hCrossternememoryear been e hbeiut pl    a  hdni32-b
  DirriveMapbeen memoryeaccessesao fisRpepptwheca lbo e hanyhdnin
	deint* sisys tmao fisRfuncRur_iwillp */
inte ioenempactin
	deid   ibrmunap(devfse & cn
	dil B_t dev,);
 ema_t, den
	dsar *buf, uint bp  pcbr_so)r  pt, st(p_info 	pget(pconn_vhpl);
    pcern) { pcbr_sofbre slotp;
  o 	pgedlecibr_vh(handle_soft pciio_info);
m	aci_ibr_soft_t       r	 */
in(NASID_TOpCOMPACT NODEID(NASID_GET_ slotp;
  _vertiir_xex_u)));s}amunap(devfse &  pcibr_dm_t dev, ma_t, device_ar *buf, uint bp  pcbr_sosk)
aph_   line_t, de d ob_t, ,k)
aphonibr_dmahibr_,k)
aph_       reqc    sen/aph_s;
    vot, pc_t       sliio_info pcpciio_info_get(pconn_vhpl);
    pcern) { pcbr_sofbrio_ihanotp_soft = pcibr_soft_get(pcibr_vh(handle_soft pciio_info);
m	aci_ibr_soft_t       oooo   pcibr_vhdl = hwgraph_cox{ pcbr_sor= bci  xraph_vert{ pcGEinfo_get(p_slot rrpciio_info_get(p_slopciio_info);
    pcibr_soft_t       io_ihanotp_soft_slot rrpciio_slop_vh&bci  xraph_vertibr_lhant(p_slo]GE_DEV_x xtalk_in  = hwgraph_coxiiopori       pcibr_dmamap_pciaddr_gexifoaddrofr_ge pcibr_dmamap_pciaddr_gepan_ibr_;srf(pcderrrrrrrrrrrrrrrrrrrrrrhaveciios;sf(pcderrrrrrrrrrrrrrrrrrrrrrminciios;s_rrrr/e merk;odn ss mrvot, pce & IOEC , pcr|= bci  xraph_vertima_t, pc  ice =(ifoaddr =lxt    dma_t, device_x  pcbr_sos 0sehibr_, reqc    sen//clost, pcr&i=MAMAP_FLAGS)  i _idge)!xifoaddr)vic		p oun_p =MA_=init\	h#endif

    x ma_t, device:et_dft
\thant(enednecRur_icibr_ %vet_dft
\txternenednecRur_icibr_ %vet_dft
\tif (d_ hibr_n[0x%x..0x%x]et_dft
\txt    dma_t, devicei */
inen 0x%xvice 	e,  pcbr_sos x  pcbr_sosk)
,addrs paddr + reqc     -=1e 	exifoaddr);pver_reg	   }
  0off(pc} dr_gth      *bfiespwn conXIO porirhiisRgoesRtoor_code & codedge)XIO=PACKED((ifoaddr))viceODE_(ifoaddr ==lXIO= ~BHERE vic		p oun_p =MA_=init\	r_gep#endif

    x ma_t, device:et_dftr_ge
\thant(enednecRur_icibr_ %vet_dftr_ge
\txternenednecRur_icibr_ %vet_dftr_ge
\tif (d_ hibr_n[0x%x..0x%x]et_dftr_ge
\txtern dma_t, devicei */
inen 0x%xvice 	er_gepa pcbr_sos x  pcbr_sosk)
r_gepaddrs paddr + reqc     -=1e 	ece =(ifoaddrdriver_reg	 att */
int0nr32rexiioporir= XIO=PO = (ifoaddr);_	xifoaddr =lXIO=ADDR (ifoaddr);__    }Ce>bsdfxiioporir= bci  xraph_vertmxid   dr_gth      *bIf * gs Dingnloescback *anus,idddddc}rn/
inte  b &= MEM a   ess (v      *bwn con
  wouldtlr_v,    NULL      *bif * e naralksval osaew cg      *bo es   pciell p r ani &= MEMor_code & codedge)xifoporir=
  pcs_v_sof_vertxid)vice *, addr =lt dev,ibr_d(ifot_t *,(handle_sofse(ifoaddrs reqc    );_	 */
inthan_ibr_;s    }prrrr/pci*/

  cm st_eca ler,
A64,htrydiah     *bs  tsfyce rlited torwiewceis 64-b
 h     *bdirrivear ao fisRca lfa*/ ofth        *bconfigur      bcteodn Dcvb_w(x)      *bconflort wiewcoppet, pcor_code & _i _idge)t, pcr&i &= (p=MAeA64)vice *, addr =l_slop_verstd64tex_unr	 DE_!(t, pcr&i &=BR_VCHAN1)_t	i _it, pcr|=eist_p VCHAN0nr	 DE__ sl_ibr_ !=eist_p =4_ucASE_UNSET  &&_	 att)t, pcr==l_slop_verstd64tt, pc))vic 		pcibrCONFIG_IA64_SGI_SN2 e atthan addr |=e(PHYS_TOp=MA((ifoaddr))_		|)
(ufnu4_ug)vxifoporir<<  &=64_ATTR_Tap(_SHFT);s#e>bsfe atthanoaddr |=e)xifoaddr_		|)
(ufnu4_ug)vxifoporir<<  &=64_ATTR_Tap(_SHFT);sver_reg_		p    printkoun_p =MA_=init\		p HWG_PERF_CHECKceunlodge)xifoaddr !=e0x20000000)rver_reg		h#endif

    x ma_t, device: [reer,]et_dftt
\thant(enednecRur_icibr_ %vet_dftt
\txternenednecRur_icibr_ %vet_dftt
\tif (d_ hibr_n[0x%x..0x%x]et_dftt
\txtern dma_t, devicei */
inen 0x%xvicdftt
\tdirrive64b
  ibr_ess is 0x%xvice 	ee,  pcbr_sos x  pcbr_sosk)
e,addrs paddr + reqc     -=1e 	e	(ifoaddrs hanoaddr);pver_reg	 att */
int(hanoaddr);p32redge)!pandletryv_etdi->bcei {pcib======hant(p_slo,ot, pc,   }

  DEV	D4_ucITS))vice att *, addr =lt dev,t, pc_totd64(t, pc,  pcs_v_sof);

aph_ slop_verstd64tt, pcr= t, pc re att_slop_verstd64tex_u =lhanoaddrof		pcibrCONFIG_IA64_SGI_SN2 e atthan addr |=e(PHYS_TOp=MA((ifoaddr))_		|)
(ufnu4_ug)vxifoporir<<  &=64_ATTR_Tap(_SHFT);s#e>bsfrrrrrrrrrrrrhan addr |=exifoaddr_		|)
(ufnu4_ug)vxifoporir<<  &=64_ATTR_Tap(_SHFT);sver_reg_	 attac}Mmke ess isozhavelan RRB_(hartwo).us code & _unlodge)!(hanotp;
  _verttrb_fixe _& (1 << hant(p_slof      	dge)t, pcr&i &=BR_VCHAN1) 	
    _get(p_slop+=eist_p RRBpSLOTNVIRTUALbr		haveciiosr= bci  xraph_vertt);
_slotlhant(p_slo]GE 	dge)haveciiosr< 2) _pttIOECODE_ sl_ibr_ &  &=64_ATTR_GE_D) 	
	minciiosr= 2;pttIOECe>bsdf		minciiosr= 1;pttIOECODE_haveciiosr< minciios)_		te_t *, pciio_autociio_s {pcib======hant(p_slo,ominciiosr-rhaveciios _dft}  f(pc} 		p oun_p =MA_=init\		p HWG_PERF_CHECKceunlodge)xifoaddr !=e0x20000000)rver_reg		h#endif

    x ma_t, device:et_dftt
\thant(enednecRur_icibr_ %vet_dftt
\txternenednecRur_icibr_ %vet_dftt
\tif (d_ hibr_n[0x%x..0x%x]et_dftt
\txtern dma_t, devicei */
inen 0x%xvicdftt
\tdirrive64b
  ibr_ess is 0x%xvicdftt
\tnew t, pc: 0x%xvice 	ee,  pcbr_sos x  pcbr_sosk)
e,addrs paddr + reqc     -=1e 	e	(ifoaddrs hanoaddr, (ufnu4_ug)vt, pcmmaver_reg	 att */
int(hanoaddr);p32reac}iupet, pcbconflort wiewcDcvb_w(x).us ibr_t, pcr= t, pcg	 att&
  &= (p=MAeA64g	 att&
  &=_p VCHAN0g	 attnre		p oun_p =MA_=init\	h#endif

    x ma_t, device:et_dft
\thant(enednecRur_icibr_ %vet_dft
\txternenednecRur_icibr_ %vet_dft
\tif (d_ hibr_n[0x%x..0x%x]et_dft
\txt    dma_t, devicei */
inen 0x%xvicdft
\tUnterr
kol.eg Dcvb_w(x)tbcteo    Dirriv-64et_e 	e,  pcbr_sos x  pcbr_sosk)
,addrs paddr + reqc     -=1e 	exifoaddr);pver_reg    }prrrr/pcTrydiabs  tsfyce rlited torwiewceis 32-b
  dirriv      *bar ao fisRca lfa*/ ofth  bconfigur      bcteodn      *bDcvb_w(x)tconflort wiewcoppet, pc,    ofth        *b*aralksa   esscdsReutsi	dein rstsIR_OFF cibr_sor_code & code_pts                         pd     
 1ULL << 31nr	 pcibr_dmamap_pciaddr_gexifoex_u =lhanotp;
  _vertiir_xex_unr	 pcibr_dmamap_pciaddr_gerotg &==exifoaddr -=xifoex_unr	 pcibr_dmamap_pciaddr_geer_rot 
 reqc     +orotg &; f	 DE__reqc     >e  pd    )=||_	 att)xifoaddr <=xifoex_u)=||_	 att)xifoporir!=lbci  xraph_vertiir_xpori)=||_	 att)er_rot >e  pd    ) vic		p oun_p =MA_=init\	r_gep#endif

    x ma_t, device:et_dftr_ge
\thant(enednecRur_icibr_ %vet_dftr_ge
\txternenednecRur_icibr_ %vet_dftr_ge
\tif (d_ hibr_n[0x%x..0x%x]et_dftr_ge
\txtern dma_t, devicei */
inen 0x%xvicdftr_ge
\txt(eerour_ieutsi	dedirriv32 *aralkvice 	er_gepa pcbr_sos x  pcbr_sosk)
r_gepaddrs paddr + reqc     -=1e 	ece =(ifoaddrdriver_reg	2 eate iceunlo *, addr =l_slop_verstd32_ex_unr	rrrr DE__ sl_ibr_ !=eist_p =32ucASE_UNSET  &&_		)t, pcr==l_slop_verstd32ut, pc))vic 		t d_ibr_ |=erotg &; f		p    printkoun_p =MA_=init\		h#endif

    x ma_t, device: [reer,]et_dftt
\thant(enednecRur_icibr_ %vet_dftt
\txternenednecRur_icibr_ %vet_dftt
\tif (d_ hibr_n[0x%x..0x%x]et_dftt
\txtern dma_t, devicei */
inen 0x%xvicdftt
\tmant deviaadirriv32 rotg &=0x%xvicdftt
\tiillpDingviaahansa    0x%xvice 	ee,  pcbr_sos x  pcbr_sosk)
e,addrs paddr + reqc     -=1e 	e	(ifoaddrs rotg &s hanoaddr);pver_reg		 */
int(hanoaddr);p3    }p_unlodge)!pandletryv_etdi->bcei {pcib======hant(p_slo,ot, pc,   }

  DEV	D32ucITS))vicg		han addr cioun32uDIE_CTpcASE;
		_slop_verstd32ut, pcr= t, pc re	_slop_verstd32uex_u =lhanoaddrof		t d_ibr_ |=erotg &; fPPO cMmke ess isozhavelan RRB_(hartwo).uss ibr_	dge)!(hanotp;
  _verttrb_fixe _& (1 << hant(p_slof      	unlohaveciiosr= bci  xraph_vertt);
_slotlhant(p_slo]GE 	IOECODE_haveciiosr< 2) _ptt	dge)_slop_verstd->bcep&   }

  DEV	PE_D) 	
	     inciiosr= 2;ptt	e>bsdf		     inciiosr= 1;ptt	dge)haveciiosr< minciios)_		t    e_t *, pciio_autociio_s {pcib======hant(p_slo,ominciiosr-rhaveciios _dft    }p_	} 		p oun_p =MA_=init\		p HWG_PERF_CHECKce	dge)xifoaddr !=e0x20000000)rver_reg		r_gep#endif

    x ma_t, device:et_dfttr_ge
\thant(enednecRur_icibr_ %vet_dfttr_ge
\txternenednecRur_icibr_ %vet_dfttr_ge
\tif (d_ hibr_n[0x%x..0x%x]et_dfttr_ge
\txtern dma_t, devicei */
inen 0x%xvicdfttr_ge
\tmant deviaadirriv32 rotg &=0x%xvicdfttr_ge
\tiillpDingviaahansa    0x%xvicdfttr_ge
\tnew t, pc: 0x%xvice 	eer_gepa pcbr_sos x  pcbr_sosk)

r_gepaddrs paddr + reqc     -=1e 	eece =(ifoaddrs rotg &s hanoaddr, (ufnu4_ug)vt, pcmmaver_reg		 */
int(hanoaddr);p3    }p_unloac}iupet, pcbconflort wiewcDcvb_w(x).us     *& 		p oun_p =MA_=init\	r_gep#endif

    x ma_t, device:et_dftr_ge
\thant(enednecRur_icibr_ %vet_dftr_ge
\txternenednecRur_icibr_ %vet_dftr_ge
\tif (d_ hibr_n[0x%x..0x%x]et_dftr_ge
\txtern dma_t, devicei */
inen 0x%xvicdftr_ge
\tUnterr
kol.eg Dcvb_w(x)tbcteo    Dirriv-32et_e 	er_gepa pcbr_sos x  pcbr_sosk)
r_gepaddrs paddr + reqc     -=1e 	ece =(ifoaddrdriver_reg	2E    }r 		p oun_p =MA_=init\r_gep#endif

    x ma_t, device:et_dfr_ge
\thant(enednecRur_icibr_ %vet_dfr_ge
\txternenednecRur_icibr_ %vet_dfr_ge
\tif (d_ hibr_n[0x%x..0x%x]et_dfr_ge
\txtern dma_t, devicei */
inen 0x%xvicdfr_ge
\tnoeacceptterr
o  rr   esso  rs.    constructterret_e 	r_gepa pcbr_sos x  pcbr_sosk)r_gepaddrs paddr + reqc     -=1e 	ce =(ifoaddrdriver_regrIOEC */
int0;e}amunap(devfse & alenlus dm_t dev, ma_t, delus _ar *buf, uint bp  pcbr_sosk)
aph_   line_t, de d ob_t, ,k)
aphoalenlus dm palenlus sen/o_h_s;
    vot, pc_t       sliio_info pcpciio_info_get(pconn_vhpl);
    pcern) { pcbr_sofbrio_ihanotp_soft = pcibr_soft_get(pcibr_vh(handle_soft pciio_info);
m	aci_ibr_soft_t       oooo   pcibr_vhdl = hwgraph_cox{ pcbr_sor= bci  xraph_vert{ pcGEinfo_get(p_slot rrpciio_info_get(p_slopciio_info);
    pcibr_soft_t       io_ihanotp_soft_slot rrpciio_slop_vh&bci  xraph_vertibr_lhant(p_slo]GEh_cox xtalk_in  = hwgraph_coxiiopori  r    alenlus dm              hant(palenlus c= 0nrf(pcalenlus dm              xt    alenlus  
 0nref(pcderrrrrrrrrrrrrrrrrrrrrrinplapc; ciads;
    voooooooooooooooodirriv64  
   s;
    vooooooooooooooooal
t, pcnref(pcdpcibr_dmamap_pciaddr_gexifoex_unrf(pcalenibr_dmamap_pciaddr_gxifoaddrofr_ges                       xifos   nref(pcs                         pd    ofr_ge pcibr_dmamap_pciaddr_gepan_ex_unrf(pcalenibr_dmamap_pciaddr_gpan_ibr_;srf(pcs;
    voooooooooooooooorelbcteo
 0nref(pcwanmerk;odn ss mrvot, pce & IOEC , pcr|= bci  xraph_vertima_t, pc  ice =inplapc 
 t, pcr&i &= (pINPLACEnEce =dirriv64r= t, pcr&i &= (p=MAeA64  i _ial
t, pcr= )t, pcr&i &= (p ~SLEEP) ? ALp ~SLEEP : 0  _i _idge)dirriv64) {    pd     
 1ullp<< 48;_	xifoex_u =l0bre sloex_u =l_slop_verstd64tex_unr	 DE_( sloex_u !=eist_p =4_ucASE_UNSET  &&_	 att)t, pcr==l_slop_verstd64tt, pc))vic_unloac}reer,
c  viallbex_u conn_br_	tp->bssdge), s);
tryv_etdi->bcei {pcib======hant(p_slo,ot, pc,   }

  DEV	D4_ucITS)r< 0)vic_unloac}Dingnlnfigur      conflort  & _unlogowedfa*/nE32 eate iceunlorelbcteo
   }

  DEV	D4_ucITS  e atthan ex_u =
	e, s);
t, pc_totd64(t, pc,  pcs_v_sof);

}s    } ->bss_ptxifoex_u =lhanotp;
  _vertiir_xex_unr	  pd     
 1ullp<< 31nr	 sloex_u =l_slop_verstd32_ex_unr	 DE_( sloex_u !=eist_p =32ucASE_UNSET  &&_	 att)t, pcr==l_slop_verstd32ut, pc))vic_unloac}reer,
c  viallbex_u conn_br_	tp->bssdge), s);
tryv_etdi->bcei {pcib======hant(p_slo,ot, pc,   }

  DEV	D32ucITS)r< 0)vic_unloac}Dingnlnfigur      conflort  & _unlogowedfa*/nE32 eate iceunlorelbcteo
   }

  DEV	D32ucITS  e atthan ex_u =ioun32uDIE_CTpcASE;
	2E    }r     xt    alenlus  =lxt    dma_t, delus _x  pcbr_sos 0sehilenlus sen//c	et, pcr&i=MAMAP_FLAGS)  i _idge)!xt    ilenlus )s	gowedfa*/nE_    alenlus dcursg__snit(xt    ilenlus s 0seNULLfbr i _idge)inplapc)vice *,t(palenlus c= xt    alenlus ;s    } ->bss_pt, st(palenlus c= alenlus dc* HtI(al
t, pcmma	ODE_!, st(palenlus )
	 attgowedfa*/nE    }r i _iwhi.
d(ALENLIST_SUCCESS ==_	 atalenlus dibr_xt    ilenlus s NULLs 0sen/c&(ifoaddrs &xifos   ,oal
t, pc))vicreacus c}fiespwn conXIO porirhiisRgoesRtoors ibr_IT)
XIO=PACKED((ifoaddr))viceh_coODE_(ifoaddr ==lXIO= ~BHERE vic		p oun_p =MA_=init\		h#endif

    x ma_t, device:et_dftt
\thant(enednecRur_icibr_ %vet_dftt
\txternenednecRur_icibr_ %vet_dftt
\tif (d_ hibr_n[0x%x..0x%x]et_dftt
\txtern dma_t, devicei */
inen 0x%xvice 	ee,  pcbr_sos x  pcbr_sosk)
e,addrs paddr + reqc     -=1e 	e	(ifoaddrmmaver_reg		 */
int0;p3    }p_unloxiioporir= XIO=PO = (ifoaddr);_	h_coxiioaddr =lXIO=ADDR (ifoaddr);_	}Ce>bsdfh_coxiioporir= bci  xraph_vertmxid   eacus c}If * gs Dingnloescback *anus,i	dc}rn/
inte  b &= MEM a   ess (v 	 *bwn con
  wouldtlr_v,    NULL 	 *bif * e naralksval osaew cg 	 *bo es   pciell p r ani &= MEMors ibr_IT)
xifoporir=
  pcs_v_sof_vertxid)vicedr_g *, addr =lt dev,ibr_d(ifot_t *,(handle_sofse(ifoaddrs xifos   fGEtIOECODE_E_ sl_ibr_ =
 (alenibr_dm)NULLf )_		gowedfa*/nE32 eate dge)dirriv64) {  dr_g len = xifoporir!=l0) re atthan_ibr_ =lt d_ex_u |)xifoaddr_		|)
(ufnu4_ug)vxifoporir<<  &=64_ATTR_Tap(_SHFT);s32 eate iceunlodpcibr_dmamap_pciaddr_gerotg &==exifoaddr -=xifoex_unr	unlodpcibr_dmamap_pciaddr_geer_rot 
 xifos    +orotg &; f	    dge)(xifo     >e  pd    )=||_		)xifoaddr <=xifoex_u)=||_		)xifoporir!=lbci  xraph_vertiir_xpori)=||_		)er_rot >e  pd    ) _		gowedfa*/nEre atthan_ibr_ =lt d_ex_u +)(xifoaddr -=xifoex_u)  e} f	/*bwritr * sioundsinga   essus c}iutt*an* e s_a ter-gall p lus .us ibr_IT)
inplapc)viceunlodge)ALENLIST_SUCCESS !=_		alenlus dreplapc(, st(palenlus s NULLsen//cl&hanoaddr, &xifos   ,oal
t, pc))_		gowedfa*/nE32 eate iceunlodge)ALENLIST_SUCCESS !=_		alenlus dant  d(, st(palenlus sen//chanoaddr, xifos   ,oal
t, pc))_		gowedfa*/nE32E    }r i _idge)relbcte)viceODE_dirriv64) {  dr_g slop_verstd64tt, pcr= t, pc re att_slop_verstd64tex_u =lhanoex_unr	2 eate iceunlo_slop_verstd32ut, pcr= t, pc reunlo_slop_verstd32uex_u =lhanoex_unr	2s    }prrrr= ne!inplapc)
	alenlus d
 nelxt    ilenlus )t);
   ac}Re.eg hherir.
	n  fcursg_
ofelegbalenlus ce hbei */
inen *ackid of pc*an* e cm st_or_code & IOECalenlus dcursg__snit(, st(palenlus s 0seNULLfbrunlore/
inthant(palenlus t);
 fa*/:;
   dge)relbcte)i	 *, pcit_t pcdi->bcei {pcib======hant(p_slo,orelbcte)  i _idge), st(palenlus c&& !inplapc)
	alenlus d
ectboyehant(palenlus  brIOEC */
int0;e}am_t, _t dev,dma  pd
rainlT(pcibema  pdmh  p)r      xt    dma  pd
rainl  p_ved
xternDGE}am_t, _t dev,dmaibr_d
rainlar *buf, uint bp  pcbr_sosk)
aph_nibr_dmahibr_,k)
aph_       bytec_t       sliio_info pcpciio_info_get(pconn_vhpl);
    pcern) { pcbr_sofbrio_ihanotp_soft = pcibr_soft_get(pcibr_vh(handle_soft pciio_info);
m	aci_ibr_soft_t       oooo   pcibr_vhdl = hwgraph_cox{ pcbr_sor= bci  xraph_vert{ pcGE     xt    dmaibr_d
rainlx  pcbr_sos ,addrs bytec_GE}am_t, _t dev,dmalus d
rainlar *buf, uint bp  pcbr_sosk)
aph_alenlus dm lus )
       sliio_info pcpciio_info_get(pconn_vhpl);
    pcern) { pcbr_sofbrio_ihanotp_soft = pcibr_soft_get(pcibr_vh(handle_soft pciio_info);
m	aci_ibr_soft_t       oooo   pcibr_vhdl = hwgraph_cox{ pcbr_sor= bci  xraph_vert{ pcGE     xt    dmalus d
rainlx  pcbr_sos lus  br}amunhiniGeg hherwo ha wheP&=bllba   ess (uttofahherg.ven singar a
e a fisRfuncRur_ival uppos
  ogrbeiut plby a clos
 frier_ of o  rs   pc
e asincetit reliesrr_i sizfactie evieis wo ha whea   ess (fieis ar eis fixe _atidnilegbalio_a t(v  imeeint* sicurring  mplrminga t(v of o  rs   pc   ibr pcibr_dm_t dev, ma  pdt dibr_dibr_sofcibema  pdmhpandleema  p)r       */
int(hanev,dma  p_ved
hanoaddr br}amun =



































































idnit bCONFIGURATION MANAGEMENT  ibrunap(devfse & _t, _t dev,provide xro hauplar *buf, uint bp dev)r  }amunap(devfse & _t, _t dev,provide xrhutdownlar *buf, uint bp dev)r  }amir.
 *, pcit_et(ar *buf, uint b  pc)
       sliio_info pcpciio_info_get(pconn_vhpl);
    pcern)  pc)GEinfo_get(p_slot rrpciio_info_get(p_slopciio_info);
    pcibr_soft_t       io_ihanotp_soft = pcibr_soft_get(pcibr_vh(handle_soft pciio_info);
m	aci_ibr_soft_t       oooo mask;e = pcibr_soft->bs_base;

  pcs_v_sof_vertex_unroooo mask;4;
    bridgereg_tctlreg  eg_ts;
    voooooooooooooooocfgctl[8];     s;
    volo waaaaaaaaaaast)aaaaderrrrrrrrrrrrrrrrrrrrrrf, nf  io_ihanotpfo);
h= pcibr_soft_get(pfo);h  io_ihanotpfo);
 = pcibr_soft_get(pfo);t)aaaaderrrrrrrrrrrrrrrrrrrrrrwin; _i _idge), s);
raph_vertibr_lhant(p_slo].habufost)vice *,t(p   p_vhhandleraph_vertibr_lhant(p_slo].host
s  p  	_get(pfo);_vhhandleraph_vertibr_lhant(p_slo].erstfo);s[0] Errrrtprrrr= nehant(pset(p< 4) {  (_vhdl);
 lf (mask == BRIDGE	nf_vhhandleraph_vertibr_lhant(p_slo].erstnfo);t)	_get(pfo);h_vhhandleraph_vertibr_lhant(p_slo].erstfo);st)	    (f =l0b fp< nf  ++f)ceunlodge)_get(pfo);h[f])_		cfgctl[f] =l/* wait untype0_cfgge->lhant(p_slo].f[f].l[oun_CFGpCOMMAND / 4]nErectlreg =l/* wait until nedtrolGE	/* wait until nedtrol =lctlreg |   }

  CTRL_RSTehant(pset(DGE	un XXXo  lay? ibr_/* wait until nedtrol =lctlregGE	un XXXo  lay? ibr)	    (f =l0b fp< nf  ++f)ceunlodge)(_get(pfo);_vhhandlefo);h[f]))_		    (win =l0b win < 6  ++win) 	
    dge)_get(pfo);->fntir_rw[tir].woex_u !=e0)_		t/* wait untype0_cfgge->lhant(p_slo].f[f].l[oun_CFGpcASE_ADDR win) / 4] =
	ee atthant(pfo);->fntir_rw[tir].woex_ut)	    (f =l0b fp< nf  ++f)ceunlodge)_get(pfo);h[f])_		/* wait untype0_cfgge->lhant(p_slo].f[f].l[oun_CFGpCOMMAND / 4] =lcfgctl[f]t)	_get(punlf (mask == BRI, sfbrg	   }
  0off(pc} 		pcibrSUPPO =_PRINTING_V_FORMAT\r_gep#endifKERN_WARNING   "%v:  *, pcit_et s; mplrming pl    set(p%det_e 	r_ge  pc==hant(p_slo);pver_reg       }
  -1br}amhant(per_ranv.
 *, pcer_ranv_et(ar *buf, uint bp  pcbr_sosk)
ahant(per_ranv._   lineer_sk)
ahant(per_ranv._  sthedper_)
       sliio_info pcpciio_info_get(pconn_vhpl);
    pcern) { pcbr_sofbrio_ihant(p_slot rrpciio_info_get(p_slopciio_info);
    pcibr_soft_t       io_ihanotp_soft = pcibr_soft_get(pcibr_vh(handle_soft pciio_info);
m	aci_ibr_soft_t       oooo mask;4;
    bridgereg_tnevreg  eg_ts;
    volo waaaaaaaaaaast);aaaath      *b    pci upportsihardwan ece iptwh; soisozca laliays      *barr>bssw    * e cm st_'s_  sthed er_rannessor_code & ef(pcs_vhdl);
 lf (mask == BRIDGEeg_tnevreg_vhhandleraph_vertibr_lhant(p_slo].erste->bce  unlodge)   lineer_ !=e  sthedper_)
	nevreg_|
   }

  DEV	SWAP_cITS  unloe>bsdfnevreg_&

   }

  DEV	SWAP_cITS  ;
   ac}NOTE-odgesozevt_ puttSWAPtbcte      *bo *an* e disterrd lus s soziill      *bhavel*anch>bssn* e logic
s rsor_code & codedge)handleraph_vertibr_lhant(p_slo].erste->bce !=e  vreg) {   mask;e = pcibr_soft->bs_base;

  pcs_v_sof_vertex_unrr_/* wait une->bcelhant(p_slo].reg_vhnevreg  e slotp;
  _vertibr_lhant(p_slo].erste->bce vhnevreg  e/* wait until Bridge PPO complete */
    pcibr_unlock(pcibr_    }prrrr_get(punlf (mask == BRI, sfbrg		p    printkoun_p =EV_DEnit\r_gep#endif

     Dcvb_w(%d): 0x%pet_e=hant(p_slo,o/* wait une->bcelhant(p_slo].regdriver_regrIOEC */
int  sthedper_br}amun  fisR(re)_etsceis GBRcann REALTIMEtbcteoann also keeps nt,ck of how
 *barny _etscan eeutstandeen.}Re.eg succoldsRonlybif * e tumberiof eutstandeen
e asetsc== 1   ibr r.
 *, pcp#eority_bctev_et(hanotp_soft =ask == BRI, 	ee, nt(p_slot rhant(p_slo, 	ee, nt(pp#eority_._   linep#eo)
      s;
    volo waaaaaaaaaaast)aaaaderrrrrrrrrrrrrrrrrrrrr*counger  oooo mask;4;
    bridgereg_trtbcteo
 0nroooo mask;4;
    bridgereg_tnevreg  eg_tderrrrrrrrrrrrrrrrrrrrrrrc =lPRIO_SUCCESS  ;
   ac}in dual-_slopnlnfigur     s,elegbhostaaesp he      *bgd torhavelsepara b singitsources,esan* ey      *bhavelsepara b ore themingeo    p#eoritytbcteor_code & ef(pccounger_vh&)handleraph_vertibr_lhant(p_slo].erstp#e_uctr)t);aaaath      *b    pci upportsio  rno    siof LOWaaespHIGH p#eority      *barbctr      reensgviaaa "REAL_TIME"tbcteint* siper-   lin      *b    pcierous to.a fio"GBR"tbctenedtrolseaccessc*an* e GBR      *breenrr_i sizxbow.a fisen*wotbcteoan e(re)_etc*agell por_code       *bXXX- BugeintRev Bb    pciSi:      *bSymptom: Prefetcher_ro hasioperaa wheincororctlyao fisRhant  e      *bduel*ancoroup t(v of legba   ess storapcieameint* siprefetcher      *bin nearnon-oros  imeeo  rited torisinu starann azoros- imee ne ie      *bputtintit's_plapc. Workarers.: UseRonlybaasinglebarbctr      reen      *bo  ound ue. GBRcann RRzca lstillpbeiunied lybut plper      *b   lin. NETLIST MERGE DONE, WILL BE FIXED IN REV Cor_code & _i _idge), s);
raph_verttev_tum !=e  }

  _A =_REV_B)
	rtbcteo|
   }

  DEV	RT  ;
   ac}NOTE-odgesozevt_ puttDEV	RT    DEV	GBRc(v      *b* e disterrd lus s soziillbhavel*antakn      *bcteintoeaccoung s rsor_code & ef(pcs_vhdl);
 lf (mask == BRIDGEeg_tnevreg_vhhandleraph_vertibr_lhant(p_slo].erste->bce  unlodge)   linep#eo ==  &==PRIO_HIGH)viceODE_(++*counger_v= 1)) iceunlodge)rtbcte)_		nevreg_|
 rtbcte reunloe>bsdf	rc =lPRIO_FAIL;

}s    } ->bssdge)   linep#eo ==  &==PRIO_LOW)viceODE_*counger_<=e0)_	rrrrrc =lPRIO_FAIL;

->bssdge)--*counger_v= 0)ceunlodge)rtbcte)_		nevreg_&

 rtbcte r    } codedge)handleraph_vertibr_lhant(p_slo].erste->bce !=e  vreg) {   mask;e = pcibr_soft->bs_base;

  pcs_v_sof_vertex_unrr_/* wait une->bcelhant(p_slo].reg_vhnevreg  e slotp;
  _vertibr_lhant(p_slo].erste->bce vhnevreg  e/* wait until Bridge PPO complete */
    pcibr_unlock(pcibr_    }prrrr_get(punlf (mask == BRI, sfbrgrrrrr*/
intrcbr}amhant(pp#eority_.
 *, pcp#eority__et(ar *buf, uint bp  pcbr_sosk)
afo_get(pp#eority_._   linep#eo)
       sliio_info pcpciio_info_get(pconn_vhpl);
    pcern) { pcbr_sofbrio_ihant(p_slot rrpciio_info_get(p_slopciio_info);
    pcibr_soft_t       io_ihanotp_soft = pcibr_soft_get(pcibr_vh(handle_soft pciio_info);
m	aci_ibr_soft_t        io_i(_t, )  *, pcp#eority_bctev_et(hanotp_sofe=hant(p_slo,o   linep#eo);grIOEC */
int   linep#eobr}amunhiniIngerfapcsRe hallow specios (e.n.}SGI) drivous kol.eg/cleaohini    pc-specific
e->bce t, pco  Mrny t, pcran emo_rei
  ohrough
dc} &=-generic
ingerfapcs;cwe * n'_hallow * em ogrbeidirrivly
 *barnipulBRId s rso  Onlybt, pcre eviioehiisRcibr_ seem prettyhini    pc-specific
ca lbo .eg hhrough tfisenspecios ingerfapcsa
e aWenperfa  emore t, pcrasceis noldiarisec,    hemove t, pcrand
e ac* HtI} &=-generic
ingerfapcsrasceis noldiariseca
e 
e aR*/
ins 0bo  fa*/ss , 1bo  succoss  ibr r.
 *, pc   linet, pc__et(ar *buf, uint bp  pcbr_sosk)
afosoft_get(p   linet, pc_tot, pc_t       sliio_info pcpciio_info_get(pconn_vhpl);
    pcern) { pcbr_sofbrio_ihant(p_slot rrpciio_info_get(p_slopciio_info);
    pcibr_soft_t       io_ihanotp_soft = pcibr_soft_get(pcibr_vh(handle_soft pciio_info);
m	aci_ibr_soft_t       soft mask;4;
    bridgereg_tg &==e0nroooo mask;4;
    bridgereg_tclro
 0nref(pc len = )t, pcr&i &=BR_DEVICE_FLAGS)_v= t, pcmma codedge)t, pcr&i &=BR_WRITE_GATHER)_	g &=|
   }

  DEV	PMU_WRGA_EN; codedge)t, pcr&i &=BR_NOWRITE_GATHER)_	clro|
   }

  DEV	PMU_WRGA_EN;  codedge)t, pcr&i &=BR_WRITE_GATHER)_	g &=|
   }

  DEV	sIR_WRGA_EN; codedge)t, pcr&i &=BR_NOWRITE_GATHER)_	clro|
   }

  DEV	sIR_WRGA_EN;  codedge)t, pcr&i &=BR_GE_DEV_P)r	g &=|
   }

  DEV	PE_D gcodedge)t, pcr&i &=BR_NOGE_DEV_P)r	clro|
   }

  DEV	PE_D g_codedge)t, pcr&i &=BR_GE_CISE)r	g &=|
   }

  DEV	PE_CISE gcodedge)t, pcr&i &=BR_NOGE_CISE)r	clro|
   }

  DEV	PE_CISE ggcodedge)t, pcr&i &=BR_BARRIER)_	g &=|
   }

  DEV	BARRIER gcodedge)t, pcr&i &=BR_NOBARRIER)_	clro|
   }

  DEV	BARRIER ggcodedge)t, pcr&i &=BR_64BIT)_	g &=|
   }

  DEV	sEV	SIZEbrcodedge)t, pcr&i &=BR_NO64BIT)_	clro|
   }

  DEV	sEV	SIZEbrrcodedge)g &=||tclr) {   mask;4;
    bridgereg_tnevreg  	s;
    volo waaaaaaaaaaast); (_vhdl);
 lf (mask == BRIDGE	nevreg_vhhandleraph_vertibr_lhant(p_slo].erste->bce  	nevreg_vh(nevreg_& ~clr) | g &; 	dge)handleraph_vertibr_lhant(p_slo].erste->bce !=e  vreg) {  oooo mask;e = pcibr_soft->bs_base;

  pcs_v_sof_vertex_unrr_oooo mask;t une->bcelhant(p_slo].reg_vhnevreg  esoft_get(pcibr_vertibr_lhant(p_slo].erste->bce vhnevreg  eoooo mask;t until Bridge PO complete */
    pcibr_unlock(pcibr_	t)	_get(punlf (mask == BRI, sfbr		p    printkoun_p =EV_DEnit\	p#endif

     Dcvb_w(%d): %Ret_e=hant(p_slo,o/* wait une->bcelhant(p_slo].reg/* wait une->bcelhant(p_slo].reg,o   linebcte)  ver_reg    }prrrr */
int(1)br}amhant(pp#ovide x = pcibr_ *, pcp#ovide  =t      _soft_tpio  pdciio__f *)  *, pcpio  pdciio_,     _soft_tpio  pdio_s_f *)  *, pcpio  pdio_s,     _soft_tpio  pdibr_df *)  *, pcpio  pdcbr_,k    _soft_tpio  pd
 nedf *)  *, pcpio  pd
 ne,k    _soft_tpio_t, devicedf *)  *, pcpio_t, device,k    _soft_tpiospapcdciio__f *)  *, pcpiospapcdciio_,k    _soft_tpiospapcdio_s_f *)  *, pcpiospapcdio_s,
k    _soft_t ma  pdciio__f *)  *, pc ma  pdciio_,k    _soft_t ma  pdio_s_f *)  *, pc ma  pdio_s,k    _soft_t ma  pdvicedf *)  *, pc ma  pdvice,k    _soft_t ma  pdlus df *)  *, pc ma  pdlus se    _soft_t ma  pd
 nedf *)  *, pc ma  pd
 nese    _soft_t ma_t, devicedf *)  *, pc ma_t, devicese    _soft_t ma_t, delus df *)  *, pc ma_t, delus se    _soft_t ma  pd
raindf *)  *, pc ma  pd
rainse    _soft_t maibr_d
raindf *)  *, pc maibr_d
rainse    _soft_t malus d
raindf *)  *, pc malus d
rain,
k    _soft_tendrdciio__f *)  *, pcendrdciio_,k    _soft_tendrdio_s_f *)  *, pcendrdio_s,k    _soft_tendrdnednecR_f *)  *, pcendrdnednecR,k    _soft_tendrddisnednecR_f *)  *, pcendrddisnednecR,k    _soft_tendrdnpu,);
 f *)  *, pcendrdnpu,);
,
k    _soft_tprovide xro haup_f *)  *, pcprovide xro haup,k    _soft_tprovide xrhutdown_f *)  *, pcprovide xrhutdown,k    _soft_tit_et_f *)  *, pcit_et,k    _soft_twritr_gall p_ridge_f *)  *, pcwritr_gall p_ridge,k    _soft_ter_ranv_et_f *)  *, pcer_ranv_et,k    _soft_tpreority__et_f *)  *, pcpreority__et,k    _soft_tnlnfig,);
 f *)  *, pcnlnfig,);
,k    _soft_tnlnfig,s;
 f *)  *, pcnlnfig,s;
,
k    _soft_terro(p   enterr f *) 0,k    _soft_terro(pexnt,ct f *) 0,k 		pcibrLATERe    _soft_t rivou_reg_cm s*ackdf *)  *, pc rivou_reg_cm s*ack,e    _soft_t rivou_unreg_cm s*ackdf *)  *, pc rivou_unreg_cm s*ack,s#e>bsfrrrr_soft_t rivou_reg_cm s*ackdf *) 0,e    _soft_t rivou_unreg_cm s*ackdf *) 0, ver_reg    _soft_t   lineunregus todf 	*)  *, pc   lineunregus tose    _soft_t ma_enterrd_f		*)  *, pc ma_enterrd,
};amir.
 *, pc ma_enterrdsar *buf, uint bp  pcbr_so)r  ibr_ *,iio_info pcpciio_info_get(pconn_vhpl);
    pcern) { pcbr_sofbrio_ihanotp_soft = pcibr_soft_get(pcibr_vh(handle_soft pciio_info);
m	aci_ibr_soft_t       	grIOEC */
intxt    dma_enterrds_